/*
 * Linux MegaRAID IDEal Software RAID Stack
 * Copyright (c) 2001 LSI Logic Corporation, All Rights Reserved.
 *-----------------------------------------------------------------------------
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *-----------------------------------------------------------------------------
 */

#ifndef __MEGAIDE_DEVEXT_H__
#define __MEGAIDE_DEVEXT_H__

#define PROCBUFSIZE		4096

typedef enum _IRCD_TYPE {
	NO_IRCD=0,
	IRCD_2X = 1,
	IRCD_3X = 2
}IRCD_TYPE;

typedef enum _CONTROLLER_TYPE {
	CMD_648 = 0,
	CMD_649 = 1,
	CMD_680 = 2,
	RCC_SERVER_WORKS_CSB5 = 3,
	SIL_3112 = 4,
	RCC_SERVER_WORKS_CSB6 = 5,
	INTEL_ICH5R = 6,
	INTEL_ICH6R = 7,
	INTEL_ICH7R = 8 
} CONTROLLER_TYPE;

/* Note... we have to change this for each controller added */
#define NUM_SUPPORTED_CONTROLLERS         7

/* SGL definition for DMA operations */
typedef struct _PHYSICAL_SGL_ENTRY {
	UINT  Address;
	UINT  Length:31;
	UINT  EndOfListFlag:1;
} PHYSICAL_SGL_ENTRY, *PPHYSICAL_SGL_ENTRY;

/* SGL definition for PIO operations */
typedef struct _LOGICAL_SGL_ENTRY {
	UINT  Address;
	UINT  Length:31;
	UINT  EndOfListFlag:1;
} LOGICAL_SGL_ENTRY, *PLOGICAL_SGL_ENTRY;

/* SGL definition for PIO operations */
typedef struct _PRINTABLE_SGL_ENTRY {
	UINT  Address;
	UINT  Length;
} PRINTABLE_SGL_ENTRY, *PPRINTABLE_SGL_ENTRY;

/* Scatter/Gather list definition */
typedef union _SGL_ENTRY {
	PHYSICAL_SGL_ENTRY   Physical;
	LOGICAL_SGL_ENTRY   Logical;
	PRINTABLE_SGL_ENTRY   Printable;
} SGL_ENTRY, *PSGL_ENTRY;

typedef struct _COMMAND_SPLIT_STATUS {
	union
	{
		struct
		{
			UCHAR  Drive0:1;
			UCHAR  Drive1:1;
			UCHAR  Drive2:1;
			UCHAR  Drive3:1;
			UCHAR  Reserved:4;
		}Bits;
		UCHAR Drives;
	}u;
}COMMAND_SPLIT_STATUS, *PCOMMAND_SPLIT_STATUS;

typedef struct _DMA_INFO {
	ULONG	ulPhyAddr;
	UINT	uiPhyAddrLen;
}
__attribute__ ((packed))
mega_dma_info_t;

typedef struct _FAST_IDE_SCB_EXTENSION {
	COMMAND_SPLIT_STATUS   CmdBuildStatus;
	UINT  NumberOfPendingCmd;
	UCHAR  Reserved[2];
	UCHAR  CmdStatus;
	/*  scsi cmd stuff copied here */
	UINT  target;
	UINT  lun;
	UINT  channel;
	UCHAR  scsi_cdb[16];
	UCHAR  cdb_len;
	VOID  *request_buffer;
#ifdef _ATAPI_SUPPORT_
   	UINT  request_bufflen;
	UCHAR direction;
#endif /*_ATAPI_SUPPORT_*/
	USHORT  sglem;
	VOID  *pstScsiCmd;
	mega_dma_info_t		dma_info;
	struct list_head  list;
#ifdef PERFORMANCE_NUMBER
	UINT time;
#endif
	UCHAR  PhysicalDrvCmd[1];
}
__attribute__ ((packed))
FAST_IDE_SCB_EXTENSION, *PFAST_IDE_SCB_EXTENSION;


/* Logical drive descriptor */
typedef struct _LOGICAL_DRIVE_INFO {
	MP_RAID_LEVEL   RaidLevel;
	UCHAR  LogicalDriveNumber;
	MP_LOGICAL_DRIVE_STATUS   Status;
	UINT  Sectors;		/* Total number of available sectors */
	UINT  StripeSize;	/* In 512-byte units */
	UCHAR  StripesPerRow;
	UCHAR  PhysicalDriveCount;
	UCHAR  PhysicalDriveTid[MAX_DRIVES_PER_CONTROLLER];
	UCHAR  SlotPhysicalDriveTid[MAX_DRIVES_PER_CONTROLLER];
	UINT  StartBlock;	/* start block on phy drive below */
	BOOLEAN  IsBkOperationStarted;
	BOOLEAN IsAbortReceived;
	UINT  EachArmLength;
	UINT  RebuildStartBlock;
	UINT  CheckConStartBlock;
	UINT  CompletedBlocks;	/* NO of completed blocks within the LD */
	UINT  uiInconsistencyCount;
	UINT  uiErrorCountDuringBackgroundTask;
	UCHAR  ucLastBackgroundTaskOperation;
	MP_ACTIVITY_STATUS  ucLastBackgroundTaskStatus;
	VOID  *GateKeeper;
	FAST_IDE_SCB_EXTENSION  *InternalScbExtension;
	UCHAR  ucControllerId;
	struct _ARRAY_DRIVE_INFO   *pArrayInfo;
	struct _LOGICAL_DRIVE_INFO   *NextLogDrive;
	struct _LOGICAL_DRIVE_INFO   *PrevLogDrive;
}
__attribute__ ((packed))
LOGICAL_DRIVE_INFO, *PLOGICAL_DRIVE_INFO;


/* Array drive descriptor */
typedef struct _ARRAY_DRIVE_INFO {
	UCHAR  ucArrayDriveNumber;
	UCHAR  ucLogicalDriveCount;
	LOGICAL_DRIVE_INFO  *HeadLogDrive;
	LOGICAL_DRIVE_INFO  *TailLogDrive;
}
__attribute__ ((packed))
ARRAY_DRIVE_INFO, *PARRAY_DRIVE_INFO;

#ifdef _SOFT_BBM_
typedef struct _SBBM_CHANGER_ {
	BOOLEAN			bIsValidSBBM;
	UINT			uiSBBMSector;
	UINT			uiSBBMBackupSector;
	UINT			State;
	VOID			*pucSBBMData;
	FAST_IDE_SCB_EXTENSION	*SbbmScb;
	SGL_ENTRY		sbbm_sgl;
	BOOLEAN			bNeedSBBMUpdate;
	BOOLEAN			bNewListToRead; /* to notify application */
} SBBM_CHANGER;
#endif	/*_SOFT_BBM_*/


/* Physical drive descriptor */
typedef struct _PHYSICAL_DRIVE_INFO {
	/* Fixed Physical Drive Info */
	UCHAR  PhysicalDriveId;	/* 0,1,2,3 for four physcial Drive */
	UCHAR  LogicalDriveId;	/* these are needed for old IRCD */
	UCHAR  LogicalIndex;	/* these are needed for old IRCD */
	UINT  BiosExposedSectors;
	UCHAR  ucMirrorDriveId;
	ARRAY_DRIVE_INFO  *pArrayInfo;	/* points to log data of this drv */
	UCHAR  ucConfigurationWasSwitched;
	UCHAR  error_log_needs_reading;
	USHORT  ErrorFound; /* count of total errors */
	/* to program down the drives if we get */
	/* 3 sequentail errors on a physical drive */
	/* this is the counter for this */
	USHORT  SequentialErrorCount;
	VOID  *ErrorLogSector;
	SGL_ENTRY  DiskErrorLogSgl;
	UINT  ErrorLogSectorIndex;
	USHORT  ErrorReported;
	UINT  IrcdSectorIndex;
	UINT  EnquirySectorIndex;
	UINT  DeviceFlags;
	BOOLEAN  bEnableRwCache;
	BOOLEAN  bEnableReadAheadCache;
	/* this is added to change read write cache on the fly */
	BOOLEAN  bCacheSettingChanged;
	BOOLEAN  bIsTurboRbldOn;	
	BOOLEAN  ucIsPowerConnected;
	MP_TRANSFER_MODE TransferMode;
	UINT  StartBlock;
	UINT  NumberOfBlocks;
	UINT  NumberOfBlocks48bit;
	MP_PHYSICAL_DRIVE_STATUS Status;
	BOOLEAN  bIsSpareDrive;
	IDENTIFY_DATA IdentifyData;
	UINT  MaximumBlockXfer;
	UCHAR  RetryCount;
	UINT  LastBlockRead;
	UINT  BadErrorSector;   /* location of last error found on drive */
	UCHAR  errorByte; /*type of error found on drive */
	BOOLEAN  IsFrozenQueue;
	UINT  CurrentQueueLength;
	VOID  *HoldQueue; /* Head Pointer */
	VOID  *TailQueue; /* Tail pointer */
	UCHAR  ucControllerId;
	BOOLEAN  bErrorLogUpdatePending;
	UINT last_check_time;
#ifdef _SOFT_BBM_
	UINT	auiBBRWBadLBAList[SBBM_MAX_HANDLE_BBRW_SECTORS];
	UINT	uiBBRWBadLBAListLen;
	SBBM_CHANGER   SBBMChanger;
#endif /*_SOFT_BBM_*/
}
__attribute__ ((packed))
PHYSICAL_DRIVE_INFO, *PPHYSICAL_DRIVE_INFO;

/* Channel descriptor */
typedef struct _CHANNEL {
	UCHAR  LastDriveFed;
	UCHAR  SwitchDrive;
	SGL_ENTRY  *pSglEntry;
	SGL_ENTRY  *SglBaseVirtualAddress;
	ULONG  SglBasePhysicalAddress;
	UINT  MaxSglEntries;
	IDE_REGISTERS_1  *BaseIoAddress1;
	IDE_REGISTERS_2  *BaseIoAddress2;
	BM_REGISTERS *BaseBmAddress;
	VOID  *ActiveQueue;
	UCHAR  ExpectingInterrupt;
	UCHAR  ucActiveCmd;
	UCHAR  ucActiveTargetId;
	UCHAR  SglElemCnt;
	SGL_ENTRY  DiskConfigurationSgl;
	UCHAR  *DiskConfiguration;
	SGL_ENTRY  DiskEnquirySgl;
	UCHAR  *DiskEnquiry;
	PASS_THRU_DATA  passthru_data;
	UCHAR  ChannelBusy;
	UCHAR  ucCurrentPMMode;	/* Serial ATA power Management */
	UINT time_jiffies;
	
	/* following elements are added for */
	/* ++SIL 3112 + Seagate Avalanche SATA (15*N)+1 data write issue */
	ULONG	ulSIL3112PatchAddress1;
	UINT	uiSIL3112PatchLength1;
	ULONG	ulSIL3112PatchAddress2;
	UINT	uiSIL3112PatchLength2;
	BOOLEAN	bLastBlockToBeDone;
	SGL_ENTRY aSGLBackUp[MAX_SGL_ENTRIES_PER_PHYSICAL_DRIVE];
	UINT	uiSGLCountBack;
	/* --SIL 3112 + Seagate Avalanche SATA (15*N)+1 data write issue */
}
__attribute__ ((packed))
CHANNEL, *PCHANNEL;

#ifdef DEBUG_IO_ACCESS
#define IO_ACCESS_READ_PCI		1
#define IO_ACCESS_WRITE_PCI		2
#define IO_ACCESS_READ_IO		3
#define IO_ACCESS_WRITE_IO		4
#define MAX_IO_ACCESS_LOG_ENTRIES	512
typedef struct{
    UCHAR	io_access;
    ULONG	io_register;
    UINT	io_value;
}IO_ACCESS_LOG;
#endif

/* Device extension */
typedef struct _mide_dev_ext_t {
	UCHAR  ucControllerId;
	UINT  uiVendorId; /* PCI PNP IDs */
	UINT  uiDeviceId;
	USHORT  usSubsysVendorId;
	USHORT  usSubsysDeviceId;
	UINT  uiUniqueId;
	UINT  uiProdId;
	USHORT  usSlot;
	UCHAR  ucPCIBus;
	UCHAR  ucPCIDev;
	UCHAR  ucPCIFun;
	UINT  uiIntLine; /* Base register locations */
	UINT  MaximumTransferLength;
	UINT  uiLogicalDriveCount;
	UINT  uiLastChannelFed;
	UINT  uiArrayDriveCount;
	ARRAY_DRIVE_INFO ArrayDrive[MAX_DRIVES_PER_CONTROLLER];
	LOGICAL_DRIVE_INFO LogicalDrive[MAX_LOGICAL_DRIVES];
	UCHAR  BootLogDrvID;
	UCHAR  BootLogDrvMapping[MAX_LOGICAL_DRIVES];
	PHYSICAL_DRIVE_INFO PhysicalDrive[MAX_DRIVES_PER_CONTROLLER];
	CHANNEL   Channel[MAX_CHANNELS_PER_CONTROLLER];
	ULONG  baseAddress5;
	UCHAR  *virtBaseAddress5;
	VOID  *ActiveGateKeeper;
	BOOLEAN  IsUsed;
	UCHAR  *real_gk_buf;
	UCHAR  *GateKeeperBuffer;
	ULONG  GateKeeperPhysicalBuffer;
	UINT  GateKeeperBufferLength;
	UCHAR  ConfigurationVirtualBuffer[IDE_SECTOR_SIZE];
	UCHAR  EnquiryVirtualBuffer[ IDE_SECTOR_SIZE ];
	UINT  NumberOfPendingCmd;
	UINT  NumberOfPendingScb;
	UCHAR  BusyTarget;
	COMMAND_SPLIT_STATUS drive_inserted;
	COMMAND_SPLIT_STATUS drive_removed;
	UCHAR  detect_try_count;
	VOID  *GateKeeper;
	FAST_IDE_SCB_EXTENSION *GateScbExtension;
	MP_IDE_VERSION stFirmwareVersion;
	MP_IDE_VERSION stIRCDVersion;
	IRCD_TYPE ucIRCDType;
	UCHAR invalid_conf_flag; 
	UCHAR   ucIRCDMajVer;
	UCHAR   ucIRCDMinVer;
	USHORT  uwCodeSize;
	UINT  uiFlag;
	UCHAR  time_stamp_update_flag; /* 1 - update needed, 0 - no update */
	UCHAR  ucErrorRecovery;
	UCHAR  aucChannelList[MAX_CHANNELS_PER_CONTROLLER];
	CHAR  achProcBuf[PROCBUFSIZE];
	INT   iProcIdx;
	VOID *adapter;
	MP_RAID_OPTIONS_EX Options;
	UINT rebuild_rate_temp;
	BOOLEAN  bFoundDeletedLD;
	CONTROLLER_SPEED	ControllerSpeed;
	CONTROLLER_TYPE		ucCardType;
	/* function pointers to support multiple adapters */
	BOOLEAN (* find_intr)(struct _mide_dev_ext_t *, UCHAR *);
	BOOLEAN (* disable_intr)(struct _mide_dev_ext_t *);
	BOOLEAN (* enable_intr)(struct _mide_dev_ext_t *);
	BOOLEAN (* is_80pin_cable)(struct _mide_dev_ext_t *, UCHAR);
	BOOLEAN (* prgrm_chip_or_drvs)(struct _mide_dev_ext_t *, UCHAR, UCHAR);
	BOOLEAN (* post_cmd_to_drive)(struct _mide_dev_ext_t *, UINT,
			UINT, UINT, UCHAR, ULONG);

	/* for back plane supprt */	
	VOID	*bp_data;
	VOID	(* release_bp)(struct _mide_dev_ext_t *);
	VOID 	(* cntrl_led_bp)(struct _mide_dev_ext_t *, UCHAR, UCHAR);
	
#ifdef DEBUG_IO_ACCESS
	UINT debug_io_head;
	IO_ACCESS_LOG debug_io[MAX_IO_ACCESS_LOG_ENTRIES];
#endif

#ifdef PERFORMANCE_NUMBER
	UINT pcmd_2_isr[2][30];
	UINT io_2_pcmd[2][30];
	UINT scbduration[30];
	UINT int_count[2];
	UINT io_count;
#endif

	INT  ICHxR_thread_pid;
	INT  ICHxR_thread_flag;
#ifdef _SOFT_BBM_
	BOOLEAN  bIsSBBMSupported; /*1-supported, 0-not supported*/
#endif /* _SOFT_BBM_ */
	BOOLEAN turbo_rbld_flag; 
#ifdef  ENABLE_ERROR_INJECTOR
	UINT  ErrorAfterCounter;
	PDErrorInjectorInfo  pdErrorInjectorInfo[MAX_DRIVES_PER_CONTROLLER];
	UINT  ErrorCounter[10];
	ErrorPatternData ErrorPattern[MAX_PATTERN];
#endif
	/* added for driver testing */
#ifdef DEBUG_ON	
	UINT test_count; 
#endif
}
__attribute__ ((packed))
HW_DEVICE_EXTENSION, *PHW_DEVICE_EXTENSION;
#endif /* __MEGAIDE_DEVEXT_H__ */

/* vim: set ts=8 sw=8 tw=78 ai si: */


