/*
 * Linux MegaRAID IDEal Software RAID Stack
 * Copyright (c) 2001 LSI Logic Corporation, All Rights Reserved.
 *-----------------------------------------------------------------------------
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *-----------------------------------------------------------------------------
 *  File Name	: megaide_scsi.h
 */
 
#ifndef __MEGAIDE_SCSI_H__
#define __MEGAIDE_SCSI_H__

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 #include <linux/smp_lock.h>
#endif

/* definitions for 2.6 kernels */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)

#define MRAID_SCSI_REGISTER		scsi_host_alloc
#define MRAID_SCSI_HOST_DEALLOC		scsi_host_put
#define MRAID_SCSI_SET_PCI_DEVICE(host, pdev) do {} while(0)
/* conversion from scsi command */
#define SCP2HOST(scp)			(scp)->device->host	/* host */
#define SCP2HOSTDATA(scp)		SCP2HOST(scp)->hostdata	/* soft state*/
#define SCP2CHANNEL(scp)		(scp)->device->channel	/* channel */
#define SCP2TARGET(scp)			(scp)->device->id	/* target */
#define SCP2LUN(scp)			(scp)->device->lun	/* LUN */
#define MRAID_CURRENT_TIME_IN_SEC	(CURRENT_TIME.tv_sec)
#define MRAID_SCSI_ADD_HOST(x,y)	scsi_add_host(x,y)
#define MRAID_SCSI_REMOVE_HOST(x)	scsi_remove_host(x)
#define MRAID_SCSI_SCAN_HOST(x)		scsi_scan_host(x)
#define MRAID_IRQRETURN_T 		irqreturn_t

#else

/* definitions for 2.4 kernels */
#define MRAID_SCSI_REGISTER		scsi_register
#define MRAID_SCSI_HOST_DEALLOC		scsi_unregister
#define MRAID_SCSI_SET_PCI_DEVICE(host, dev) \
	scsi_set_pci_device(host, dev);
#define MRAID_SCSI_ADD_HOST(x,y)	(0)
#define MRAID_SCSI_REMOVE_HOST(x)	do {} while(0)
#define MRAID_SCSI_SCAN_HOST(x)		do {} while(0)
/* conversion from scsi command */
#define SCP2HOST(scp)			(scp)->host		/* host*/
#define SCP2HOSTDATA(scp)		SCP2HOST(scp)->hostdata	/* soft state*/
#define SCP2CHANNEL(scp)		(scp)->channel		/* channel */
#define SCP2TARGET(scp)			(scp)->target		/* target */
#define SCP2LUN(scp)			(scp)->lun		/* LUN */
#define MRAID_IRQRETURN_T 		VOID
#define	IRQ_RETVAL(x)
#define MRAID_CURRENT_TIME_IN_SEC	(CURRENT_TIME)
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	#define MRAID_HOST_SET_LOCK(host, lock) scsi_assign_lock(host, lock)
#else
	#define MRAID_HOST_SET_LOCK(host, lock) lock = &io_request_lock;
#endif

typedef struct _KERN_PHYSICAL_DRIVE_INFO_ {
	dma_addr_t	h_err_log_sec;
#ifdef _SOFT_BBM_
	dma_addr_t	h_sbbm_sec;
#endif	/*_SOFT_BBM_*/
} __attribute__ ((packed)) KERN_PHYSICAL_DRIVE_INFO;

typedef struct _KERN_CHANNEL_ {
	dma_addr_t	h_sgl_entry;
	dma_addr_t	h_disc_conf;
	dma_addr_t	h_enquiry_conf;
} __attribute__ ((packed)) KERN_CHANNEL;

typedef struct _CARD_INFO {
	HW_DEVICE_EXTENSION	DeviceExtension;
	KERN_PHYSICAL_DRIVE_INFO PhysicalDrive[MAX_DRIVES_PER_CONTROLLER];
	KERN_CHANNEL		Channel[MAX_CHANNELS_PER_CONTROLLER];
	INT			irq;
	struct pci_dev		*pstDev;
	struct Scsi_Host	*pstScsiHost;
	wait_queue_head_t	wait_queue;
	struct completion	ICHxR_completion;
	spinlock_t		lock;
	spinlock_t		*host_lock;
	UINT			gate_keeper_buff_length;
	dma_addr_t		h_gate_keeper_buff;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	wait_queue_head_t	internal_wait;
#endif	
	struct list_head	scb_ext_free_list;
	/*for IOCTL memory allocation*/
	struct semaphore	nit_lock;
	Scsi_Cmnd		*nit_scsi_cmd;
	dma_addr_t		h_nit_scsi_cmd;
	CHAR			*host_scribble;
	dma_addr_t		h_host_scribble;
	VOID			*ioctl_data;
	dma_addr_t		h_ioctl_data;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	struct scsi_device	*nit_scsi_device;
	dma_addr_t		h_nit_scsi_device;
#endif

#ifdef CONFIG_PROC_FS
	struct proc_dir_entry	*adp_proc_entry;
	struct proc_dir_entry	*pd_dir_proc_entry;
	struct proc_dir_entry	*ld_dir_proc_entry;
#endif

} __attribute__ ((packed)) adapter_t;


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static INT megaide_detect(Scsi_Host_Template * hostTpl);
static INT megaide_release(struct Scsi_Host * host);
static INT megaide_biosparam(Disk * disk, kdev_t dev, INT * piGeom);
static INT megaide_reboot_notify(struct notifier_block *, unsigned long, VOID *);
static INT megaide_command(Scsi_Cmnd * scmd);
static VOID mide_internal_done(Scsi_Cmnd * scmd);
#else		    
static INT megaide_init(VOID);
static VOID megaide_exit(VOID);
static INT megaide_biosparam(struct scsi_device *sdev, 
		struct block_device *bdev, sector_t capacity, INT geom[]);
static VOID megaide_shutdown(struct device *dev);
#endif
static VOID __megaide_shutdown(adapter_t *adapter);
static INT __devinit megaide_probe_one(struct pci_dev *, 
		const struct pci_device_id *);
static VOID megaide_detach_one(struct pci_dev *pdev);
static const CHAR *megaide_info(struct Scsi_Host * host);
static INT megaide_queue(Scsi_Cmnd * scmd, VOID(*scsi_done)(Scsi_Cmnd *));
static INT megaide_abort(Scsi_Cmnd * scmd);
static INT megaide_reset(Scsi_Cmnd * scmd, UINT i_ResetFlags);
static MRAID_IRQRETURN_T megaide_isr(INT iIrq, VOID *vpDevp, struct pt_regs * pstRegs);
static INT eh_megaide_abort(Scsi_Cmnd *scmd);
static INT eh_megaide_reset_bus(Scsi_Cmnd *scmd);
static INT eh_megaide_reset_host(Scsi_Cmnd *scmd);
static INT eh_megaide_reset_dev(Scsi_Cmnd *scmd);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/* MegaIDE driver Host Template */
#define MEGAIDE_TEMPLATE					\
{								\
	module			: THIS_MODULE,			\
	name			: "MegaIDE",			\
	proc_name		: "megaide",			\
	proc_info		: megaide_proc_info,		\
	info			: megaide_info,			\
	queuecommand		: megaide_queue,		\
	bios_param		: megaide_biosparam,		\
	eh_abort_handler	: eh_megaide_abort,		\
	eh_device_reset_handler	: eh_megaide_reset_dev,		\
	eh_bus_reset_handler	: eh_megaide_reset_bus,		\
	eh_host_reset_handler	: eh_megaide_reset_host,	\
	can_queue		: M_IDE_MAX_CMD_PER_LUN,	\
	this_id			: M_IDE_HOST_INITIATOR_ID,	\
	sg_tablesize		: M_IDE_MAX_SGLIST,		\
	cmd_per_lun		: M_IDE_MAX_CMD_PER_LUN,	\
	present			: 0,				\
	unchecked_isa_dma	: 0,				\
	use_clustering		: ENABLE_CLUSTERING		\
}
#else

#define MEGAIDE_TEMPLATE 						\
{								\
	name			: "MegaIDE",			\
	proc_name		: "megaide",			\
	proc_info		: megaide_proc_info,		\
	detect			: megaide_detect,		\
	release			: megaide_release,		\
	info			: megaide_info,			\
	command			: megaide_command,		\
	queuecommand		: megaide_queue,		\
	bios_param		: megaide_biosparam,		\
	eh_abort_handler	: eh_megaide_abort,		\
	eh_device_reset_handler	: eh_megaide_reset_dev,		\
	eh_bus_reset_handler	: eh_megaide_reset_bus,		\
	eh_host_reset_handler	: eh_megaide_reset_host,	\
	can_queue		: M_IDE_MAX_CMD_PER_LUN,	\
	this_id			: M_IDE_HOST_INITIATOR_ID,	\
	sg_tablesize		: M_IDE_MAX_SGLIST,		\
	cmd_per_lun		: M_IDE_MAX_CMD_PER_LUN,	\
	present			: 0,				\
	unchecked_isa_dma	: 0,				\
	use_clustering		: ENABLE_CLUSTERING,		\
	use_new_eh_code		: 1				\
}
#endif


VOID megaide_complete_command(HW_DEVICE_EXTENSION *pdev,
		FAST_IDE_SCB_EXTENSION *scb, INT result);
VOID fill_sense_buffer(FAST_IDE_SCB_EXTENSION *scb, UCHAR error_byte,
		BOOLEAN blReadWriteError, UINT error_sector);
ULONG GetPhysicalAddressAndLength(VOID * request_buffer,
		USHORT sglem, UINT uiDataOffset, 
		UINT *puiContiguousLength);

#ifdef _ATAPI_SUPPORT_
/* these function are added to support ATAPI Non Raid devices */
VOID mega_place_sense_data_and_result(HW_DEVICE_EXTENSION *pdev,
		FAST_IDE_SCB_EXTENSION *scb, UINT chan, UCHAR status, INT result);
ULONG mega_get_sgl_len_and_addr(VOID *buffer, UINT sg_elem, UINT *plen);
VOID mega_scsi_to_atapi(FAST_IDE_SCB_EXTENSION *scb);
#endif /*_ATAPI_SUPPORT_*/

#endif /*__MEGAIDE_SCSI_H__*/

/* vim: set ts=8 sw=8 tw=78 ai si: */
