/*
 * Linux MegaRAID IDEal Software RAID Stack
 * Copyright (c) 2001 LSI Logic Corporation, All Rights Reserved.
 *-----------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *-----------------------------------------------------------------------------
 * FILE		:	megaide_scsi.c
 *
 * Authors	:	Manoj Jose	<Manoj.Jose@lsil.com>
 *			Atul Mukker	<Atul.Mukker@lsil.com>
 *
 * Description	:
 * This driver is supported by LSI Logic Corportation. This software is the
 * Open Source Code of MegaRAID IDEal Software RAID Stack. Closed Part of the
 * Source contains LSI's RAID intelligence and is provided as binary
 * image
 *
 * This driver Support CSB6/CMD649/CMD680/SIL3112/ICH5R/HR/ICH6R/ICH7R Chipsets
 */

/***IF THESE FLAGS ENABLED DO IT IN CLOSED SOURCE TOO***/
/*
#define DEBUG_IO_ACCESS 0xFFFF
#define ENABLE_ERROR_INJECTOR 0xFF
#define ENABLE_ERROR_VALIDATION 0xFF
*/
#define		_SOFT_BBM_
#define		_ATAPI_SUPPORT_
//#define	TEST_ICH6R_ALL	  

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/ctype.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/pci_ids.h>
#ifdef SUSE70_INSTALL
 #include <linux/ide.h>
#endif
#include <linux/ide.h>
#include <linux/proc_fs.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 #include <linux/blkdev.h>
#else
 #include <linux/blk.h>
#endif
#include <linux/wait.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 #include <linux/tqueue.h>
#endif
#include <linux/interrupt.h>
#include <linux/mm.h>
#include <asm/pgtable.h>
#include <linux/sched.h>
#include <linux/stat.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 7)
 #include <linux/malloc.h>
#else
 #include <linux/slab.h>
#endif
#include <linux/spinlock.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/dma.h>
#include <asm/uaccess.h>
#include <asm/system.h>
/* for shutdown Notification routines */
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/kernel.h>
/* to support 32 bit applications */
#ifdef CONFIG_COMPAT 
#include <asm/ioctl32.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 #include <linux/moduleparam.h>
 #include <scsi/scsi_host.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 #include "../sd.h"
 #include "../hosts.h"
#endif
#include "../scsi.h"
#include "./inc/megaide_datatype.h"
#include "./inc/megaide_debug.h"
#include "./inc/megaide_adapter.h"
#include "./inc/megaide_ata.h"
#include "./util/mp_hd_ioctls.h"
#include "./inc/megaide_devext.h"
#include "./inc/megaide_pci.h"
#include "./inc/megaide_kernel_lib.h"
#include "./inc/megaide_char.h"
#include "./inc/megaide_scsi.h"
#include "./inc/megaide_proc.h"
#include "./inc/megaide_macros.h"
#ifndef MEGAIDE_LIBRARY_READY
 #include "./lib/megaide_lib.c"
#endif

#define SHUTDOWN_SIGS	(sigmask(SIGKILL)|sigmask(SIGINT)|sigmask(SIGTERM))
#define OEM_PRODUCT_ID	" LD XY  IDERAID "
#define OEM_VENDOR_ID	"LSILOGIC"
/* Maximum allowed request size */
#ifndef KMALLOC_MAXSIZE
 #define KMALLOC_MAXSIZE		131072UL
#endif

/* Global variables */
adapter_t *gastCardInfo[MAX_CONTROLLERS] = {};
static INT giNumberOfCtrlrs = 0;
static Scsi_Host_Template megaide_template_g	= MEGAIDE_TEMPLATE;
static Scsi_Host_Template *megaide_template_gp	= &megaide_template_g;

INT debug_level =
	AR_DEVEL_DEBUG|AR_DEBUG_ATA;  


/* supported controller ids */
static struct pci_device_id megaide_pci_tbl_g[] = {
	{ PCI_VENDOR_ID_CMD, PCI_DEVICE_ID_CMD_648,
	  PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ PCI_VENDOR_ID_CMD, PCI_DEVICE_ID_CMD_649,
	  PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ PCI_VENDOR_ID_CMD, PCI_DEVICE_ID_DELL_680,
	  PCI_VENDOR_ID_DELL, PCI_DEVICE_ID_DELL_CERC_ATA_AS, 0, 0, 0 },
	{ PCI_VENDOR_ID_CMD, PCI_DEVICE_ID_CMD_680,
	  PCI_VENDOR_ID_LSI_LOGIC, PCI_DEVICE_ID_LSI_524, 0, 0, 0 },
	{ PCI_VENDOR_ID_CMD, PCI_DEVICE_ID_CMD_3112,
	  PCI_VENDOR_ID_LSI_LOGIC, 0x534, 0, 0, 0 },
	{ PCI_VENDOR_ID_SERVERWORKS, PCI_DEVICE_ID_SVWS_CSB6IDE,
	  PCI_VENDOR_ID_COMPAQ, PCI_DEVICE_ID_COMPAQ_CSB6, 0, 0, 0 },
	{ PCI_VENDOR_ID_SERVERWORKS, PCI_DEVICE_ID_SVWS_CSB6IDE,
	  PCI_VENDOR_ID_IBM, PCI_DEVICE_ID_IBM_CSB6, 0, 0, 0 },
	{ PCI_VENDOR_ID_SERVERWORKS, PCI_DEVICE_ID_SVWS_CSB6IDE,
	  PCI_VENDOR_ID_LSI_LOGIC, PCI_DEVICE_ID_SVWS_CSB6IDE, 0, 0, 0 },
	/* Intel CSB6 - mccaran */
	{ PCI_VENDOR_ID_SERVERWORKS, PCI_DEVICE_ID_SVWS_CSB6IDE,
	  PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_SVWS_CSB6IDE, 0, 0, 0 },
	
	/* Intel Hance Rapid - Sun Prairie */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  PCI_VENDOR_ID_INTEL, 0x3441, 0, 0, 0 },
	/* Intel Hance Rapid - Sun Prairie */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  PCI_VENDOR_ID_INTEL, 0x3443, 0, 0, 0 },
	/* Intel Hance Rapid - Grand Prairie */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  PCI_VENDOR_ID_INTEL, 0x3445, 0, 0, 0 },
	/* Intel Hance Rapid - Volcano Peak */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  PCI_VENDOR_ID_INTEL, 0x3451, 0, 0, 0 },
	/* Intel Hance Rapid - Eden Prairie */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  PCI_VENDOR_ID_INTEL, 0x3459, 0, 0, 0 },
	/* Intel Hance Rapid - Eden Prairie (SCSI depop)*/
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  PCI_VENDOR_ID_INTEL, 0x345B, 0, 0, 0 },
	/* Intel Hance Rapid - Rolling Prairie */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  PCI_VENDOR_ID_INTEL, 0x345D, 0, 0, 0 },
	/* Intel Hance Rapid - Rolling Prairie SCSI depop */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  PCI_VENDOR_ID_INTEL, 0x345F, 0, 0, 0 },
	/* Intel ICH5R - Volcano Peak 2 */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  PCI_VENDOR_ID_INTEL, 0x3467, 0, 0, 0 },
	  
	/* FSC Hance Rapid - RX100-S2 */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  0x1734, 0x1034, 0, 0, 0 },
	/* FSC Hance Rapid - Econel 40 */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  0x1734, 0x1040, 0, 0, 0 },
	/* FSC Hance Rapid - Econel 200 */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  0x1734, 0x1073, 0, 0, 0 },
	/* Lenovo HR/ESB6300 */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  0x17AA, 0x497C, 0, 0, 0 },
	/* Lenovo HR/ESB6300 -2*/
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  0x17AA, 0x4964, 0, 0, 0 },
	/* Asustek NCLV-D2*/
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_HR,
	  0x1043, 0x81A9, 0, 0, 0 },

	/* Intel ICH5R - Alief HP */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x3432, 0, 0, 0 },
	/* Intel ICH5R - Alief HPC (SCSI de-pop) */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x3434, 0, 0, 0 },
	/* Intel ICH5R - Alief non-HP */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x344D, 0, 0, 0 },
	/* Intel ICH5R - Brandon SCSI */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x343A, 0, 0, 0 },
	/* Intel ICH5R - Brandon-V */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x343C, 0, 0, 0 },
	/* Intel ICH5R - Brandon SATA */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x343E, 0, 0, 0 },
	/* Intel ICH5R - Jarrell */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x3436, 0, 0, 0 },
	/* Intel ICH5R - Jarrell SCSI de-pop */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x3438, 0, 0, 0 },
	/* Intel ICH5R - Jarrell 2 */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x3455, 0, 0, 0 },
	/* Intel ICH5R - Jarrell 2 SCSI de-pop */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x3457, 0, 0, 0 },
	/* Intel ICH5R - Brandon 2 */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x3461, 0, 0, 0 },
	/* Intel ICH5R - Brandon 2-V */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x3463, 0, 0, 0 },
	/* Intel ICH5R - Brandon SCSI depop */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  PCI_VENDOR_ID_INTEL, 0x3465, 0, 0, 0 },
	/* Intel ICH5R - Asustek NCL-DE*/
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5R,
	  0x1043, 0x81A8, 0, 0, 0 },

#ifdef TEST_ICH6R_ALL	  
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  0xFFFF, 0xFFFF, 0, 0, 0 },
#else	  
	/* Intel ICH6R - Buckner 5U */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  PCI_VENDOR_ID_INTEL, 0x344B, 0, 0, 0 },
	/* Intel ICH6R - Buckner 1U */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  PCI_VENDOR_ID_INTEL, 0x344F, 0, 0, 0 },
	/* Intel ICH6R - Bartman */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  PCI_VENDOR_ID_INTEL, 0x3453, 0, 0, 0 },

	/* FSC ICH6R - Econel 50 */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  0x1734, 0x105C, 0, 0, 0 },
	/* FSC ICH6R - RX100-S3 */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  0x1734, 0x107A, 0, 0, 0 },
	/* FSC ICH6R Asustec */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  0x1043, 0x8184, 0, 0, 0 },
	/* LSI ICH6R - For NEC */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  0x1458, 0xB003, 0, 0, 0 },
	/* SuperMicro ICH6R - P8SCT */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  0x15D9, 0x7480, 0, 0, 0 },
	/* SuperMicro ICH6R - P8SCT/I */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  0x15D9, 0x6280, 0, 0, 0 },
	/* Lenovo ICH6R */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  0x17AA, 0x1627, 0, 0, 0 },
	/* IESC ICH6R */
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6R,
	  0x1170, 0x000B, 0, 0, 0 },
#endif	/*TEST_ICH6R_ALL*/

	/* FSC ICH7R */ 
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH7R,
	  0x1734, 0x1085, 0, 0, 0 },
	/* FSC Asustec */ 
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH7R,
	  0x1043, 0x819E, 0, 0, 0 },
	/* Intel Nob Hill LX-RAID */ 
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH7R,
	  PCI_VENDOR_ID_INTEL, 0x3469, 0, 0, 0 },
	/* Intel Nob Hill LC-RAID */ 
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH7R,
	  PCI_VENDOR_ID_INTEL, 0x346B, 0, 0, 0 },
	/* LSI ICH7R test */ 
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH7R,
	  PCI_VENDOR_ID_INTEL, 0x27C0, 0, 0, 0 },

	{ 0, }
};

MODULE_DEVICE_TABLE(pci, megaide_pci_tbl_g);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
#ifdef MODULE
 MODULE_PARM(debug_level, "i");
#endif
#else
module_param(debug_level, int, 0);
#endif
MODULE_AUTHOR("LSI Logic");
MODULE_DESCRIPTION("LSI Logic megaide driver");
MODULE_LICENSE("Proprietary");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static struct pci_driver megaide_pci_driver_g = {
	.name		= "megaide",
	.id_table	= megaide_pci_tbl_g,
	.probe		= megaide_probe_one,
	.remove		= __devexit_p(megaide_detach_one),
	.driver		= {
		.shutdown = megaide_shutdown,
	}
};
#else
static struct notifier_block megaide_reboot_notifier = {
	notifier_call : megaide_reboot_notify
};
#endif

static UINT mraid_process_and_queue_req(HW_DEVICE_EXTENSION *, Scsi_Cmnd *);
static INT mide_scsi_free(HW_DEVICE_EXTENSION *pDE, VOID *data, UINT length);
static VOID *mide_scsi_malloc(HW_DEVICE_EXTENSION *pDE, UINT length);
static INT mide_init_cmd_memory_pool(HW_DEVICE_EXTENSION *pDE);
static INT mide_free_cmd_memory_pool(HW_DEVICE_EXTENSION *pDE);
static FAST_IDE_SCB_EXTENSION *mide_get_scb_from_pool(HW_DEVICE_EXTENSION *pDE);
static INT mide_release_scb_pool_memory(HW_DEVICE_EXTENSION *pDE,
	FAST_IDE_SCB_EXTENSION *scb);

#ifdef _ATAPI_SUPPORT_
static UINT mega_process_and_q_atapi_cmd(HW_DEVICE_EXTENSION *, Scsi_Cmnd *);
#endif /*_ATAPI_SUPPORT_*/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
/**
 * Description	:
 * This is used only with <2.6 linux kernels. This is detect entry point for
 * the mid-layer, driver scan the PCI bus for raid software stack supported
 * controllers
 *
 * @param host_template	: megaraid host template
 * @return	:
 *	number of adapter found
 **/
INT
megaide_detect(Scsi_Host_Template *host_template)
{
	struct pci_dev		*dev = NULL;
	struct pci_device_id	*id;
	UCHAR			class_dev;

	AR_DEBUG(AR_DBG_ENTRY, (KERN_DEBUG "megaide_detect called\n"));

	/* override the template for old-style interface */
	megaide_template_gp	= host_template;

	printk_Rlsi(KERN_INFO "megaide: " MEGAIDE_VERSION M_IDE_CRLFSTR);

	for (id = megaide_pci_tbl_g; id->vendor; id++)
	{
	while ((dev = pci_find_device(id->vendor, id->device, dev))) {

		pci_read_config_byte(dev, PCI_CLASS_DEVICE, &class_dev);

		if((class_dev == 4) ||
			(id->device == PCI_DEVICE_ID_CMD_3112)) {
			
/* no sub system id check for ich6r */
			if (((dev->subsystem_vendor & id->subvendor) == 
					dev->subsystem_vendor) &&
				((dev->subsystem_device & id->subdevice) == 
					 dev->subsystem_device)) {

				if (!giNumberOfCtrlrs) {
					RemoveBuildInIDESupport(id);
				}

				megaide_probe_one(dev, id);
				break;
			}
		}
	}
	}

	if (giNumberOfCtrlrs) {

		megaidedev_attach();

		/* Register the Shutdown Notification hook in kernel */
		if(register_reboot_notifier(&megaide_reboot_notifier)) {
			printk_Rlsi (KERN_INFO
				"megaide: reboot_notifier not registered!\n");
		}
	}

	return(giNumberOfCtrlrs);
}

/**
 * Description	:
 * This is used only with <2.6 linux kernels. We call the corresponding detach
 * routine for each adapeters.
 *
 * @param host	: host to be released
 * Return	: 
 *	zero if command succeeds
 **/
INT
megaide_release(struct Scsi_Host *host)
{
	adapter_t	*adapter = (adapter_t *) host->hostdata;

	AR_DEBUG(AR_DBG_ENTRY, (KERN_DEBUG "megaide _release Called\n"));

	megaide_detach_one(adapter->pstDev);

	if (!giNumberOfCtrlrs) {
	
		unregister_reboot_notifier(&megaide_reboot_notifier);

		/* Un Register character Driver interface */
		megaidedev_detach();
	}

	return 0;
}

/**
 * Description	: 
 * This is used only with <2.6 linux kernels and it return the disk geometry
 * for a particular disk
 *
 * Parameters
 * @param disk	: Disk geometry
 * @param dev	: Device node
 * @param piGeom: Returns geometry fields
 *
 * Return	: 
 *	zero if command succeeds
 **/
INT 
megaide_biosparam(Disk * disk, kdev_t dev, INT * piGeom)
{
	INT			heads;
	INT			sectors;
	INT			cylinders;
	HW_DEVICE_EXTENSION	*pdev;
	adapter_t		*adapter;

	AR_DEBUG(AR_DBG_ENTRY, (KERN_DEBUG "megaide: biosparam Called\n"));
	/* Get pointer to host config structure */
	adapter = (adapter_t *) disk->device->host->hostdata;
	pdev = (HW_DEVICE_EXTENSION *) &adapter->DeviceExtension;

	if (disk->capacity < 1032192) { /* size less than 504MB */
		heads = 16;
		sectors = 63;
		cylinders = (INT) disk->capacity / (heads * sectors);
	} else if (disk->capacity < 2064384) { /* size less than 1008MB */
		heads = 32;
		sectors = 63;
		cylinders = (INT) disk->capacity / (heads * sectors);
	} else if (disk->capacity < 4128768) { /* size less than 2016MB */
		heads = 64;
		sectors = 63;
		cylinders = (INT) disk->capacity / (heads * sectors);
	} else if (disk->capacity < 8257536) { /* size less than 4032MB */
		heads = 128;
		sectors = 63;
		cylinders = (INT) disk->capacity / (heads * sectors);
	} else { /* size greater than 4032MB */
		heads = 255;
		sectors = 63;
		cylinders = (INT) disk->capacity / (heads * sectors);
	}

	/* return result */
	piGeom[0] = heads;
	piGeom[1] = sectors;
	piGeom[2] = cylinders;

	return 0;
}

volatile static INT giInternalDoneFlag = 0;
volatile static INT giInternalDoneErrcode = 0;

/**
 * Description	: 
 * This is used only with <2.6 linux kernels. The command function issues a
 * command synchronously and waits for it to complete. Driver put these 
 * commands in our queue function and wait for completion.
 *
 * Parameters
 * @param scmd	: scsi command pointer
 *
 * Return	:
 *	Returns the result of scsi command
 **/
INT
megaide_command(Scsi_Cmnd * scmd)
{
	HW_DEVICE_EXTENSION	*pdev;
	adapter_t		*adapter = (adapter_t *) SCP2HOSTDATA(scmd);

	AR_DEBUG(AR_DBG_ENTRY, (KERN_DEBUG "megaide _command Called\n"));

	giInternalDoneFlag = 0;
	pdev = (HW_DEVICE_EXTENSION *) &adapter->DeviceExtension;

	/* this command should go to the head of the queue */
	pdev->uiFlag |= M_IDE_BLOCK_CMD;

	/* Queue command, and wait until it has completed */
	megaide_queue(scmd, mide_internal_done);

	/* Wait till command get over */
	while (!giInternalDoneFlag) {
		interruptible_sleep_on(&adapter->internal_wait);
	}
	
	return giInternalDoneErrcode;
}


VOID
mide_internal_done(Scsi_Cmnd * scmd)
{
	adapter_t	*adapter = (adapter_t *) SCP2HOSTDATA(scmd);
	
	giInternalDoneErrcode = scmd->result;
	giInternalDoneFlag++;
	wake_up(&adapter->internal_wait);
}

/**
 * Description	:
 * This routine will be called when the use has done a forced shutdown on the
 * system. Flush the disks cache.
 *
 * Parameters
 * @param this	:
 * @param code	: shutdown code
 * @param nused	: unused
 *
 * Return	:
 *	NOTIFY_DONE after finishing 
 **/
static INT
megaide_reboot_notify (struct notifier_block *this, unsigned long code,
		VOID *unused)
{
	INT	i;

	for (i=0; i<giNumberOfCtrlrs; i++)
	{
		__megaide_shutdown((adapter_t *)gastCardInfo[i]);
	}
	return NOTIFY_DONE;
}
#else

/**
 * Description	: 
 * This function is used only with >=2.6 linux kernels and it returne the 
 * disk geometry for a particular disk
 *
 * Return	: 
 *	zero if command succeeds
 **/
static INT
megaide_biosparam(struct scsi_device *sdev, struct block_device *bdev,
		sector_t capacity, INT geom[])
{
	INT		heads;
	INT		sectors;
	INT		cylinders;

	AR_DEBUG(AR_DBG_ENTRY, (KERN_DEBUG "megaide: biosparam Called\n"));

	if (capacity < 1032192) { /* size less than 504MB */
		heads = 16;
		sectors = 63;
		cylinders = (INT) capacity / (heads * sectors);
	} else if (capacity < 2064384) { /* size less than 1008MB */
		heads = 32;
		sectors = 63;
		cylinders = (INT) capacity / (heads * sectors);
	} else if (capacity < 4128768) { /* size less than 2016MB */
		heads = 64;
		sectors = 63;
		cylinders = (INT) capacity / (heads * sectors);
	} else if (capacity < 8257536) { /* size less than 4032MB */
		heads = 128;
		sectors = 63;
		cylinders = (INT) capacity / (heads * sectors);
	} else { /* size greater than 4032MB */
		heads = 255;
		sectors = 63;
		cylinders = (INT) capacity / (heads * sectors);
	}

	/* return result */
	geom[0] = heads;
	geom[1] = sectors;
	geom[2] = cylinders;
	return 0;
}

/**
 * Description	: 
 * This is driver module entry point in >=2.6 kernel. This is called 
 * during driver load.
 *
 * Return	: 
 *	zero if command succeeds
 **/
static INT __init 
megaide_init(VOID)
{
	INT	error;
	
	printk_Rlsi(KERN_INFO "megaide: " MEGAIDE_VERSION M_IDE_CRLFSTR);

	error = pci_module_init(&megaide_pci_driver_g);
	if (error) {
		printk_Rlsi ("megaide: driver initialization error...\n");
		return error;
	}
	
	megaidedev_attach();
	return 0;
}

/**
 * Description	: 
 * This is driver module exit point in >=2.6 kernel. This is called 
 * during driver unload.
 *
 * Return	: 
 *	zero if command succeeds
 **/
static VOID __exit 
megaide_exit(VOID)
{
	megaidedev_detach();
	pci_unregister_driver(&megaide_pci_driver_g);
}

/**
 * Description	: 
 * This is driver shut down routine in >=2.6 kernel. This function make use 
 * of generic shut down rotine for both <2.6 and >=2.6 kernels.
 *
 * Return	: None
 **/
static VOID
megaide_shutdown(struct device *dev)
{
	struct Scsi_Host *host = pci_get_drvdata(to_pci_dev(dev));
	
	__megaide_shutdown((adapter_t *) host->hostdata);
}
#endif

/**
 * Description	: 
 * Generic shut down rotine for both <2.6 and >=2.6 kernels.
 *
 * Return	: None
 **/
static VOID
__megaide_shutdown(adapter_t *adapter)
{
	HW_DEVICE_EXTENSION	*pdev = &adapter->DeviceExtension;

	free_irq (pdev->uiIntLine, adapter);

	/* kill ICHxR polling thread */
	if (IS_ICHxR_CONTROLLER(pdev)) {
		mega_kill_ICHxR_poll_thread(pdev);
	}
		
	DisableInterrupts(pdev);
	FlushCache(pdev);
	EnableInterrupts(pdev);
}


/**
 * Description	: 
 * This routine should be called whenever a new adapter is detected by the
 * PCI hotplug susbsytem.
 *
 * Parameters
 * @param dev	: handle to this controller's PCI configuration space
 * @param id	: pci device id of the class of controllers
 *
 * Return	: 
 *	zero if command succeeds
 **/
 
static INT __devinit
megaide_probe_one(struct pci_dev *dev, const struct pci_device_id *id)
{
	struct Scsi_Host	*host = NULL;
	HW_DEVICE_EXTENSION	*pdev = NULL;
	adapter_t		*adapter = NULL;
	uint32_t		unique_id;
	INT			i;
	
	/* Make sure this adapter is not already setup */
	unique_id	= dev->bus->number << 8 | dev->devfn;
	for (i = 0; i < giNumberOfCtrlrs; i++) {
		if (gastCardInfo[i]->DeviceExtension.uiUniqueId == unique_id) {
			
			/* already probed device! */
			AR_DEBUG(AR_DEBUG_WARNING,
				("megaide: reject probe request for dev: "));
  
			AR_DEBUG(AR_DEBUG_WARNING,
				("%#4.04x:%#4.04x:%#4.04x:%#4.04x:",
				dev->vendor, dev->device,
				dev->subsystem_vendor,
				dev->subsystem_device));

			AR_DEBUG(AR_DEBUG_WARNING,
				("bus %d:slot %d:func %d\n",
				dev->bus->number, PCI_SLOT(dev->devfn),
				PCI_FUNC(dev->devfn)));
			return -ENODEV;
		}
	}

	AR_DEBUG(AR_DBG_INIT, (KERN_DEBUG
		"megaide: probe new device %#4.04x:%#4.04x:%#4.04x:%#4.04x: ",
		dev->vendor, dev->device, dev->subsystem_vendor,
		dev->subsystem_device));

	AR_DEBUG(AR_DBG_INIT, (KERN_DEBUG "bus %d:slot %d:func %d\n",
		dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn)));

	dev->dma_mask = 0xFFFFFFFF;

	/* Initialize SCSI Host structure */
	host = MRAID_SCSI_REGISTER(megaide_template_gp, sizeof(adapter_t));
	if (!host) {
		printk_Rlsi(KERN_WARNING "megaide: scsi_register failed\n");
		return(-ENODEV);
	}
	
	/* to make sure the SCSI Mid Layer IOCTL Works */
	MRAID_SCSI_SET_PCI_DEVICE(host, dev);

#if LINUX_VERSION_CODE == KERNEL_VERSION(2, 4, 9)
	host->can_dma_32	= 0;
	host->single_sg_ok	= 1;
#endif
	adapter	= (adapter_t *) host->hostdata;
	memset_Rlsi(adapter, 0, sizeof(adapter_t));

	pdev		= (HW_DEVICE_EXTENSION *)&adapter->DeviceExtension;
	pdev->adapter	= (VOID *)adapter;
	
	gastCardInfo[giNumberOfCtrlrs] = adapter;

#ifdef PERFORMANCE_NUMBER
	pdev->int_count[0]	= 0;
	pdev->int_count[1]	= 0;
	pdev->io_count		= 0;
#endif
	/* initialize adapter type, speed, and function pointers */
	if (mide_init_adp_info_and_pointers(pdev, dev->device) != AR_TRUE) {
		goto go_scsi_unregister;
	}
	
	/* initialize all arrays/drives as being not-present */
	for (i = 0; i < MAX_LOGICAL_DRIVES; i++)
	{
		pdev->LogicalDrive[i].ucControllerId = giNumberOfCtrlrs;
	}

	for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++)
	{
		pdev->PhysicalDrive[i].ucControllerId = giNumberOfCtrlrs;
	}

	if (mide_init_cmd_memory_pool (pdev)==FALSE){
		goto go_scsi_unregister;
	}

	/* memory alloc failure is handled properly */
	if (FALSE == mide_allocate_device_ext_mem (pdev, dev, giNumberOfCtrlrs)) {
		goto go_release_pool_mem;
	}

	/* Semaphore for IOCTL */
	init_MUTEX (&adapter->nit_lock);

	/* in 2.4 kernel lock is not used */
	spin_lock_init(&adapter->lock);
	adapter->host_lock = &adapter->lock;

	/* 
	 * in 2.4 kernel host_lock is reset to io request lock
	 * in 2.6 kernel it is assigned as midlayer host lock
	 */
	MRAID_HOST_SET_LOCK(host, adapter->host_lock);

	/* Need to initialize Device Extension resource structure */
	adapter->pstScsiHost	= host;
	adapter->irq		= dev->irq;
	host->irq		= dev->irq;
	host->io_port		= pci_resource_start(dev, 0);
	host->n_io_port		= 16;
	host->unique_id		= (dev->bus->number << 8) | dev->devfn;
	host->max_channel	= M_IDE_MAX_CHANNELS;
	host->max_id		= M_IDE_MAX_LOGICAL_DRIVES;
	host->max_lun		= M_IDE_MAX_LOGICAL_DRIVES;
	host->cmd_per_lun	= M_IDE_MAX_CMD_PER_LUN;
	host->max_sectors	= MAX_SECTORS_PER_IDE_TRANSFER;
	/* Store card Information */
	pdev->ucPCIBus		= dev->bus->number;
	pdev->ucPCIDev		= dev->devfn >> 3;
	pdev->ucPCIFun		= dev->devfn & 7;
	pdev->usSlot		= PCI_SLOT(dev->devfn);
	pdev->uiUniqueId	= (dev->bus->number << 8) | dev->devfn;
	pdev->uiProdId		= id->device;
	pdev->ucControllerId	= giNumberOfCtrlrs;
	pdev->uiVendorId	= dev->vendor;
	pdev->uiDeviceId	= dev->device;
	pdev->usSubsysVendorId	= dev->subsystem_vendor;
	pdev->usSubsysDeviceId	= dev->subsystem_device;
	adapter->pstDev		= dev;

	/* set default values for our IOCTL options */
	pdev->Options.uiRebuildRate	= 10;
	pdev->Options.uiIdentical	= TRUE;
	pdev->Options.uiAutoRebuild	= FALSE;
	pdev->Options.uiCheckArray	= 0;
	pdev->Options.uiStopRebuild	= 0;
	pdev->Options.uiErrorThreshold	= 0;
	pdev->Options.uiFailDrive	= 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	init_waitqueue_head(&adapter->internal_wait);
#endif	

	/* controller's base address in device ext */
	if (FALSE == mide_init_base_address (pdev, dev)) {
		goto go_cleanup_mem;
	}

	if (AR_TRUE != ClaimOurIOPorts(pdev)) {
		printk_Rlsi(KERN_ERR "megaide: err Claiming ioports\n");
		goto go_cleanup_mem;
	}

	DisableInterrupts(pdev);

	if ((dev->device == PCI_DEVICE_ID_SERVERWORKS_CSB5IDE) ||
		(dev->device == PCI_DEVICE_ID_SVWS_CSB6IDE)) {
		
		UCHAR		swks_irq = dev->irq;

		/* 
		 * To Fix CSB6 Problem, native IDE Driver sets IRQ to 0
		 * We are reading directly from Configuration space
		 */
		if (dev->device == PCI_DEVICE_ID_SVWS_CSB6IDE)
			pci_read_config_byte(dev, 0x3C, &swks_irq);
		
		if (swks_irq != dev->irq) 
			dev->irq = swks_irq;
		
		adapter->pstScsiHost->irq = dev->irq;
	}

	if (request_irq(dev->irq, megaide_isr, SA_SHIRQ, "megaide", adapter)) {
		printk_Rlsi(KERN_ERR "megaide: Couldn't register IRQ %d\n",
			dev->irq);
		goto go_release_ioports;
	}

	pdev->uiIntLine	= dev->irq;

	mide_get_option_rom_info(pdev);

	if (AR_ERROR == HyperDiskScanForDevices(pdev)) {
		printk_Rlsi(KERN_WARNING "megaide: Scan Devices failed\n");
		goto go_err_release;
	}

	EnableInterrupts(pdev);

	if (!pdev->ucIRCDType) {
		/* 
		 * This is the case when No device detected with proper
		 * IRCD Setting to the default IRCD
		 */
		pdev->ucIRCDType = IRCD_3X;
		pdev->invalid_conf_flag = 1;
	}

	/* start polling thread as soon as the drive loads */
	if (IS_ICHxR_CONTROLLER(pdev)) {
		mega_start_ICHxR_poll_thread(pdev);
	}

#ifdef _ATAPI_SUPPORT_
	for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++)
	{
		if(IS_ATAPI_DRIVE(pdev, i)) {
			host->max_sectors	= 124;
			break;
		}
	}
#endif
	/* Adjust SGL's buffer pointer for the channels */
	AssignSglPtrsForChannels(pdev);
	
	/* Discover back plane if present */
	mega_discover_bp(pdev);
	
	AR_DEBUG(AR_DBG_INIT,(KERN_DEBUG "megaide: scsi_register ok\n"));
	
	/* found controller successfully */
	giNumberOfCtrlrs++;
	
	/* setup adapter handle in PCI soft state */
	pci_set_drvdata(dev, host);

	if (MRAID_SCSI_ADD_HOST(host, &dev->dev)) {
		printk_Rlsi (KERN_WARNING
			"megaide: MRAID_SCSI_ADD_HOST failed.\n");
		goto go_err_release;
	}

	MRAID_SCSI_SCAN_HOST(host);

#ifdef CONFIG_PROC_FS
	megaideproc_attach(adapter);
#endif
	/* success */
	return(0);

/* if error happens */
go_err_release:
	pdev->release_bp(pdev);
	free_irq(host->irq, adapter);

go_release_ioports:
	ReleaseOurIOPorts(pdev);
	/* we disabled the interrupt before, now hit an error 
	 * so enable it back to the same state
	 */
	pdev->enable_intr(pdev);

go_cleanup_mem:
	CleanUpAdapter(pdev);

go_release_pool_mem:
	mide_free_cmd_memory_pool(pdev);

go_scsi_unregister:
	MRAID_SCSI_HOST_DEALLOC(host);

	printk_Rlsi(KERN_WARNING "megaide: leaving device !!!\n");
	
	return -ENODEV;
}

static VOID
megaide_detach_one(struct pci_dev *dev)
{
	struct Scsi_Host	*host = pci_get_drvdata(dev);
	adapter_t		*adapter = (adapter_t *) host->hostdata;
	HW_DEVICE_EXTENSION	*pdev;
		
	AR_DEBUG(AR_DBG_ENTRY, (KERN_DEBUG "megaide detach one\n"));

	/* Make sure we don't break for multiple detach(?) for a controller */
	if (!adapter) {
		printk_Rlsi(KERN_CRIT
		"megaide: Invalid detach on %#4.04x:%#4.04x:%#4.04x:%#4.04x\n",
			dev->vendor, dev->device, dev->subsystem_vendor,
			dev->subsystem_device);

		return;
	} else {
		printk_Rlsi(KERN_NOTICE
		"megaide: detaching device %#4.04x:%#4.04x:%#4.04x:%#4.04x\n",
			dev->vendor, dev->device, dev->subsystem_vendor,
			dev->subsystem_device);
	}
	
	pdev	= (HW_DEVICE_EXTENSION *) &adapter->DeviceExtension;
	
	MRAID_SCSI_REMOVE_HOST(host);
	
	free_irq(adapter->irq, adapter);
	
	DisableInterrupts(pdev);
	
	/* kill ICHxR polling thread */
	if (IS_ICHxR_CONTROLLER(pdev)) {
		mega_kill_ICHxR_poll_thread(pdev);
	}
	
	FlushCache(pdev);
	
	EnableInterrupts(pdev);
	
	pdev->release_bp(pdev);
	
	ReleaseOurIOPorts(pdev);
	
	/* deallocate the momory allocated in each adapter */
	CleanUpAdapter(pdev);
	
#ifdef CONFIG_PROC_FS
	megaideproc_detach(adapter);
#endif
	mide_free_cmd_memory_pool (pdev);
	
	pci_set_drvdata (dev, NULL);
	
	giNumberOfCtrlrs --;

	/* Unregister the driver */
	MRAID_SCSI_HOST_DEALLOC(adapter->pstScsiHost);

	return;
}

/**
 * Description	:
 * Driver entry point to the scsi mid-layer, Get information about the
 * controller.
 *
 * Parameters
 * @param host	: host pointer
 *
 * Return Value
 *	Pointer to a static character array containing a brief description of
 *	the driver
 **/

const CHAR *
megaide_info(struct Scsi_Host * host)
{
	static CHAR		acBuffer[512];
	HW_DEVICE_EXTENSION	*pdev;
	adapter_t		*adapter;

	AR_DEBUG(AR_DBG_ENTRY, (KERN_DEBUG "megaide_info Called\n"));

	adapter	= (adapter_t *) host->hostdata;
	pdev	= (HW_DEVICE_EXTENSION *) &adapter->DeviceExtension;
	
	if ((pdev->stFirmwareVersion.usMajorVer == 0) &&
		(pdev->stFirmwareVersion.usMinorVer == 0) &&
		(pdev->stFirmwareVersion.uiBuild == 0)) {
		
		/* Bios disabled */
		sprintf(acBuffer,
"LSI Logic MegaIDE RAID BIOS Disabled, %d targs %d chans %d luns",
		host->max_id,
		host->max_channel,
		host->max_lun);
	} else {
		
		sprintf(acBuffer,
"LSI Logic MegaIDE RAID BIOS Version %d.%d.%8.8d, %d targs %d chans %d luns",
			pdev->stFirmwareVersion.usMajorVer,
			pdev->stFirmwareVersion.usMinorVer,
			pdev->stFirmwareVersion.uiBuild,
			host->max_id,
			host->max_channel,
			host->max_lun);
	}
	
	return acBuffer;
}

/**
 * Description	:
 * This function issues a command and does not wait for it to finish.
 * When the command is finished, the done() function is called with the
 * Scsi_Cmnd structure pointer as a parameter. 
 *
 * Parameters
 * @param scmd		: Pointer to the Scsi_Cmnd structure,
 * @param scsi_done	: callback function funtion pointer
 *
 * Return	:
 *	0- if command succeeds
 **/
INT
megaide_queue(Scsi_Cmnd * scmd, VOID(*scsi_done)(Scsi_Cmnd *))
{
	adapter_t		*adapter = (adapter_t *) SCP2HOSTDATA(scmd);

	scmd->scsi_done = scsi_done;

#ifdef _ATAPI_SUPPORT_
	if (SCP2CHANNEL(scmd)) {
		return (mega_process_and_q_atapi_cmd(
			(HW_DEVICE_EXTENSION *) &adapter->DeviceExtension,
			scmd));
	}
#endif /*_ATAPI_SUPPORT_*/
	
	return (mraid_process_and_queue_req(
		(HW_DEVICE_EXTENSION *) &adapter->DeviceExtension, scmd));
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
UINT 
mraid_validate_request(Scsi_Cmnd *scmd)
{
	struct	scatterlist *sg = (struct scatterlist *) scmd->request_buffer;
	UINT	i, total_sz = 0;

	if (scmd->use_sg) {
		for(i = 0; i < scmd->use_sg; i++)
		{
			if (!sg[i].page) {
				AR_DEBUG(AR_DEBUG_ERRORS, 
					("megaide: error wrong sgl entry----\n"));
				return FALSE;
			}
			total_sz += sg[i].length;
		}	
	} else {
		if (scmd->request_buffer) {
			AR_DEBUG(AR_DEBUG_ERRORS, ("megaide: error wrong sgl----\n"));
			return FALSE;
		}
	}

	if (total_sz > (MAX_SECTORS_PER_IDE_TRANSFER * IDE_SECTOR_SIZE)) {
		AR_DEBUG(AR_DEBUG_ERRORS, ("megaide: error wrong sgl size----\n"));
		return FALSE;
	}

	return TRUE;
}
#endif

UINT
mraid_process_and_queue_req(HW_DEVICE_EXTENSION *pdev, Scsi_Cmnd *scmd)
{
	LOGICAL_DRIVE_INFO	*ldInfo;
	FAST_IDE_SCB_EXTENSION	*scb;

	ldInfo	= GetLDInfo(pdev, SCP2CHANNEL(scmd), SCP2TARGET(scmd),
			SCP2LUN(scmd));

	if (!ldInfo) {
		/* No such device */
		scmd->result = (DID_BAD_TARGET << 16); 
		scmd->scsi_done(scmd);
		return 0;
	}

	if ((ldInfo->Status == LDS_OffLine)) {
		check_for_config_update(pdev);
		scmd->result = (DID_BAD_TARGET << 16); 
		printk_Rlsi(KERN_ERR "megaide: log drv offline!!\n");
		scmd->scsi_done(scmd);
		return 0;
	}

	check_for_config_update(pdev);
	
	scmd->host_scribble = 0;

	switch (scmd->cmnd[0]) {
	case VERIFY :
	case READ_6:
	case READ_10:
	case WRITE_6:
	case WRITE_10:
	case WRITE_VERIFY:
	{
		 /* No such device */
		if (ldInfo->Status == LDS_Deleted) {
			scmd->result = (DID_BAD_TARGET << 16);
			scmd->scsi_done(scmd);
			return 0;
		}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
		if (mraid_validate_request(scmd) == FALSE) {
			scmd->result = (DID_ERROR << 16);
			scmd->scsi_done (scmd);
			return 1;
		}
#endif

		/* alloc host_scribble */
		scmd->host_scribble = 
			mide_scsi_malloc(pdev, get_scb_extension_size());
			
		scb = (FAST_IDE_SCB_EXTENSION *) scmd->host_scribble;

		if (scb == NULL) {
			printk_Rlsi(KERN_WARNING
				"megaide: host_scribble mem alloc failed \n");
			
			scmd->result = (DID_ERROR << 16);
			scmd->scsi_done (scmd);
			return 1;
		}

		memset_Rlsi(scb, 0, get_scb_extension_size());

		/* os specific scsi packet stuff copied to srb extension */
		scb->pstScsiCmd	= (VOID *)scmd;
		scb->target	= SCP2TARGET(scmd);
		scb->lun	= SCP2LUN(scmd);
		scb->channel	= SCP2CHANNEL(scmd);
#ifdef PERFORMANCE_NUMBER
		pdev->io_count++;
		scb->time	= jiffies;
#endif
		
		memcpy_Rlsi(scb->scsi_cdb, scmd->cmnd, MAX_COMMAND_SIZE);
		
		scb->cdb_len		= scmd->cmd_len;
		scb->request_buffer	= (VOID *)scmd->buffer;
		scb->sglem		= scmd->use_sg;

		if(!megaide_atapi_startio(pdev,scb)) {
			
			mide_scsi_free(pdev, scmd->host_scribble,
				get_scb_extension_size());
			
			scmd->result		= (DID_BAD_TARGET << 16);
			scmd->host_scribble	= 0;
			scmd->scsi_done(scmd);
		}
		break;
	}

	case READ_CAPACITY:
	{
		UINT			bytesPerBlock = IDE_SECTOR_SIZE;
		PREAD_CAPACITY_DATA	capacity;
		UINT			lsize;

		memset_Rlsi(scmd->buffer, 0, scmd->request_bufflen);

		capacity	= (PREAD_CAPACITY_DATA) scmd->buffer;
		/* reporting logical block size */
		lsize		= ldInfo->Sectors - 1;
		capacity->BytesPerBlock = BIG_ENDIAN_UINT(bytesPerBlock);
		capacity->LogicalBlockAddress = BIG_ENDIAN_UINT(lsize);
		scmd->result	= (DID_OK << 16);
		scmd->scsi_done(scmd);
		break;
	}

	case TEST_UNIT_READY:
		scmd->result = (DID_OK << 16);
		scmd->scsi_done(scmd);
		break;

	case MODE_SENSE :
	{
		CHAR *sense;

		sense = scmd->request_buffer;
		sense[0] = 3;	/* Mode data length */
		sense[1] = 0;	/* Medium type - default */
		sense[2] = 0;	/* Device-specific param, bit 8: 0/1 = write enabled/protected */
		sense[3] = 0;	/* Block descriptor length */

		scmd->result = (DID_OK << 16);
		scmd->scsi_done(scmd);
		break;
	}

	case INQUIRY:
	{
		INQUIRYDATA		*pInq;

		memset_Rlsi(scmd->buffer, 0, scmd->request_bufflen);

		pInq			= (INQUIRYDATA *) scmd->buffer;
		pInq->Versions	= 2;
		pInq->ResponseDataFormat	= 2;
		((UCHAR *) pInq) [7]		= 0x32;
		pInq->AdditionalLength	= 0xFA;

		/* Fill Actual Vendor ID depending on SubsystemVendorID */
		memmove_Rlsi(pInq->ProductId, OEM_PRODUCT_ID, 16);

		/* Fill Actual Product ID depending on SubsystemVendorID */
		memmove_Rlsi(pInq->VendorId, OEM_VENDOR_ID, 8);
		pInq->ProductId[4] = ' ';/* higher digit */
		/* lower digit */
		pInq->ProductId[5] = (UCHAR)ldInfo->LogicalDriveNumber + '0';
		scmd->result = (DID_OK << 16);
		scmd->scsi_done(scmd);
		break;
	}

	case REQUEST_SENSE :
	{
		memset_Rlsi(scmd->buffer, 0, scmd->request_bufflen);

		((UCHAR *) scmd->buffer) [0] = 0x70;
		((UCHAR *) scmd->buffer) [7] = 0x18;
		((UCHAR *) scmd->buffer) [8] = 0xFF;
		((UCHAR *) scmd->buffer) [9] = 0xFF;
		((UCHAR *) scmd->buffer) [10] = 0xFF;
		((UCHAR *) scmd->buffer) [11] = 0xFF;
		((UCHAR *) scmd->buffer) [19] = 0xFF;
		((UCHAR *) scmd->buffer) [20] = 0xFF;
		((UCHAR *) scmd->buffer) [21] = 0xFF;
		((UCHAR *) scmd->buffer) [22] = 0xFF;
		((UCHAR *) scmd->buffer) [23] = 0xFF;

		scmd->result = (DID_OK << 16);
		scmd->scsi_done(scmd);
		break;
	}

	case FORMAT_UNIT :
	scmd->result = (DID_BAD_TARGET << 16);
	scmd->scsi_done(scmd);
	break;

	case MEGAIDE_IOCTL_NEW:
	case MEGAIDE_IOCTL :
	{
		/* alloc host_scribble */
		scmd->host_scribble =
			mide_scsi_malloc(pdev, get_scb_extension_size());

		scb = (FAST_IDE_SCB_EXTENSION *) scmd->host_scribble;

		if (scb == NULL) {
			printk_Rlsi(KERN_WARNING
					"megaide: unable to allocate host_scribble\n");
			scmd->result = (DID_ERROR << 16);
			scmd->scsi_done (scmd);
			/* non-zero status, could not find memory */
			return 1;
		}

		memset_Rlsi(scb, 0, get_scb_extension_size());
		/* os specific scsi packet stuff copied to srb extension */
		scb->pstScsiCmd	= (VOID *)scmd;
		scb->target	= SCP2TARGET(scmd);
		scb->lun	= SCP2LUN(scmd);
		scb->channel	= SCP2CHANNEL(scmd);

		memcpy_Rlsi(scb->scsi_cdb, scmd->cmnd, MAX_COMMAND_SIZE);

		scb->cdb_len		= scmd->cmd_len;
		scb->request_buffer	= (VOID *)scmd->buffer;
		scb->sglem		= scmd->use_sg;

		switch (HandleIOCTLCall(pdev, scb)) {

			case IOCTL_OK:
				megaide_complete_command(pdev, scb, (DID_OK << 16));
				break;

			case IOCTL_ERROR:
				megaide_complete_command(pdev, scb,
						(DID_BAD_TARGET << 16));
				AR_DEBUG(AR_DEBUG_IOCTL, (KERN_DEBUG
							"megaide: IOCTL err\n"));
				break;

			case IOCTL_QUEUE:
				AR_DEBUG(AR_DEBUG_IOCTL,
						(KERN_DEBUG "megaide: IOCTL queued\n"));
				break;

			default:
				break;
		}

		break;
	}

	default:
	scmd->result = (DID_BAD_TARGET << 16);
	scmd->scsi_done(scmd);
	break;
	}

	return 0;
}

#ifdef _ATAPI_SUPPORT_
	static CHAR 
mega_get_data_direction(UCHAR sc_data_direction)
{
	UCHAR direction = READ_TRANSFER;

	switch (sc_data_direction) {
		case SCSI_DATA_WRITE:
			direction = WRITE_TRANSFER;
			break;
		case SCSI_DATA_READ:
		case SCSI_DATA_NONE:
			break;
		case SCSI_DATA_UNKNOWN:
		default:
			printk_Rlsi ("megaide: unknowd data direction=0x%x!!\n", 
					sc_data_direction);
			break;
	}
	return (direction);
}

/**
 * mega_process_and_q_atapi_cmd: process atapi commands
 *
 * Here ATAPI commands are processed separately. All channel one devices
 * are reserved for atapi devices. So we can support ATAPI devices on all
 * ports in the controller
 *
 * Parameters
 * @pdev	: Device extension pointer
 * @scmd	: SCSI command pointer
 * @return	:
 *	0	: if the command success or processed
 *	1.	: if Driver is busy
 **/
static UINT
mega_process_and_q_atapi_cmd(HW_DEVICE_EXTENSION *pdev, Scsi_Cmnd *scmd)
{
	FAST_IDE_SCB_EXTENSION *scb;

	if ((SCP2CHANNEL(scmd) != 1) ||
		(SCP2TARGET(scmd) >= MAX_DRIVES_PER_CONTROLLER)) {
		scmd->result = (DID_BAD_TARGET << 16);
		scmd->scsi_done(scmd);
		return 0;
	}

	if (!IS_ATAPI_DRIVE(pdev, SCP2TARGET(scmd))) {
		scmd->result = (DID_BAD_TARGET << 16);
		scmd->scsi_done(scmd);
		return 0;
	}

	// alloc host_scribble
	scmd->host_scribble = mide_scsi_malloc(pdev, get_scb_extension_size());

	scb = (FAST_IDE_SCB_EXTENSION *) scmd->host_scribble;

	if (scb == NULL) {
		printk_Rlsi(KERN_WARNING
			"megaide: host_scribble mem alloc failed \n");
		scmd->result = (DID_ERROR << 16);
		scmd->scsi_done (scmd);
		/*non-zero status, could not find memory*/
		return 1;
	}

	memset_Rlsi(scb, 0, get_scb_extension_size());
	scb->direction = 
		mega_get_data_direction(scmd->sc_data_direction);
	
	/*os specific scsi packet stuff copied to srb extension*/
	scb->pstScsiCmd = (VOID *)scmd;
	scb->target	= SCP2TARGET(scmd);
	scb->lun	= SCP2LUN(scmd);
	scb->channel	= SCP2CHANNEL(scmd);
#ifdef PERFORMANCE_NUMBER
	pdev->io_count++;
	scb->time = jiffies;
#endif
	memcpy_Rlsi(scb->scsi_cdb, scmd->cmnd, MAX_COMMAND_SIZE);

	scb->cdb_len = scmd->cmd_len;
	scb->request_buffer = (VOID *)scmd->buffer;
	scb->sglem = scmd->use_sg;
	scb->request_bufflen = scmd->request_bufflen;

	if(! megaide_atapi_startio(pdev,scb)) {
		
		mide_scsi_free(pdev, scmd->host_scribble,
			get_scb_extension_size());

		scmd->result = (DID_BAD_TARGET << 16);
		scmd->host_scribble = 0;
		scmd->scsi_done(scmd);
	}

	return 0;
}
#endif /* _ATAPI_SUPPORT_ */

VOID
megaide_complete_command(HW_DEVICE_EXTENSION *pdev,
	FAST_IDE_SCB_EXTENSION *scb, INT result)
{
	Scsi_Cmnd	*scmd;
	
	scmd	= (Scsi_Cmnd *) scb->pstScsiCmd;

	if(! scmd) { return; }

	if(scmd->host_scribble) {
		/* IOCTL memory is allocated by driver,
		* not by scsi, so handled separately
		*/
		if (scmd->state != MIDE_CHAR_IOCTL_CALLED) {
			mide_scsi_free(pdev, scmd->host_scribble,
				get_scb_extension_size());
		}
		scmd->host_scribble = 0;
	}
	
	scmd->result = result;
	scmd->scsi_done(scmd);
}

/**
 * Description	:
 * This function is called by SCSI midlayer to abort a scsi command. New error
 * handling entry points also uses this.
 *
 * Parameters
 * @param scmd	: Pointer to the Scsi_Cmnd structure
 *
 * Return	:
 *	SCSI_ABORT_SNOOZE	- We are not doing anything
 *	SCSI_ABORT_SUCCESS	- Abort command success
 *	SCSI_ABORT_PENDING	- Called abort, waiting for interrupt
 *	SCSI_ABORT_BUSY		- Unable to Abort
 *	SCSI_ABORT_ERROR	- Something went wrong!!
 **/
INT megaide_abort(Scsi_Cmnd * scmd)
{
	HW_DEVICE_EXTENSION	*pdev;
	adapter_t		*adapter = (adapter_t *) SCP2HOSTDATA(scmd);
	PLOGICAL_DRIVE_INFO	ldInfo;
	FAST_IDE_SCB_EXTENSION	*scb = NULL;

	pdev = (HW_DEVICE_EXTENSION *) &adapter->DeviceExtension;

	AR_DEBUG(AR_DBG_ENTRY, (KERN_DEBUG "megaide _abort Called\n"));

	if ((scmd->cmnd[0] == MEGAIDE_IOCTL_NEW) ||
		(scmd->cmnd[0] == MEGAIDE_IOCTL)) {
		
		scb	= (FAST_IDE_SCB_EXTENSION *) scmd->host_scribble;
		
		megaide_abort_ioctl(pdev, scb);
		return SCSI_ABORT_SUCCESS;
	}

	ldInfo = GetLDInfo(pdev, SCP2CHANNEL(scmd), SCP2TARGET(scmd),
		SCP2LUN(scmd));

	if (ldInfo) {
		if (ldInfo->Status == LDS_Deleted) {
			/* No such device */
			scmd->result = (DID_BAD_TARGET << 16);
			scmd->scsi_done(scmd);
			return SCSI_ABORT_SUCCESS;
		}

		if (ldInfo->Status == LDS_OffLine) {
			printk_Rlsi(KERN_ERR "megaide: log drv offline!!\n");
			scmd->result = (DID_BAD_TARGET << 16); 
			scmd->scsi_done(scmd);
			return SCSI_ABORT_SUCCESS;
		}
	}

	printk_Rlsi(KERN_INFO "megaide: snooze and wait for reset \n");
	return(SCSI_ABORT_SNOOZE);
}


/**
 * Description	:
 * Reset a previous SCSI request. This function is called by SCSI midlayer only
 *
 * Parameters
 * @param scmd		: Pointer to the Scsi_Cmnd structure
 * @param i_ResetFlags	:
 *
 * Return Value
 * SCSI_RESET_SNOOZE	- Not doing anything,Cmd may finish
 * SCSI_RESET_PUNT	- Not doing anything, Given up, midlayer can proceed
 * SCSI_RESET_SUCCESS	- Reset command success
 * SCSI_RESET_PENDING	- Called reset, waiting for interrupt
 * SCSI_RESET_WAKEUP	- Did reset,don't expect an inter to signal DID_RESET
 * SCSI_RESET_NOT_RUNNING - Command is not active in Low level code
 * SCSI_RESET_ERROR	- Something went wrong!
 */
INT 
megaide_reset(Scsi_Cmnd *scmd, UINT i_ResetFlags)
{
	PLOGICAL_DRIVE_INFO	ldInfo;
	adapter_t		*adapter = (adapter_t *) SCP2HOSTDATA(scmd);

	ldInfo = GetLDInfo(&adapter->DeviceExtension, 
		SCP2CHANNEL(scmd), SCP2TARGET(scmd), SCP2LUN(scmd));
		
	if (!ldInfo || ldInfo->Status == LDS_OffLine) {
		return SCSI_RESET_SUCCESS;
	}

	if (FastIDEResetBus(&adapter->DeviceExtension) != TRUE)
		return(SCSI_RESET_ERROR);
	
	return(SCSI_RESET_SUCCESS);
}

/** 
 * Description	:
 * Interrupt service routine. This routine should register during request IRQ.
 *
 * Parameters
 * @param iIrq		: IRQ number
 * @param vpDdevp	: Device specific data passed by the kernel
 * @param pstRegs	: Stucture defines the way the registers
 *			 are stored on the stack during system call
 * Return Value
 * In <2.6 kernel
 *	None
 * In >=2.6 kernel
 *	IRQ_RETVAL(1) if interrupt handled
 *	IRQ_RETVAL(0) if interrupt not handled
 */

MRAID_IRQRETURN_T
megaide_isr(INT iIrq, VOID *vpDevp, struct pt_regs * pstRegs)
{
	unsigned long	flags;
	INT		handled = AR_FALSE;
	adapter_t	*adapter = (adapter_t *) vpDevp;

	AR_DEBUG(AR_DEBUG_INTERUPT, (KERN_DEBUG "megaide _isr Called\n"));

	spin_lock_irqsave(adapter->host_lock, flags);

	handled = FastIDEInterrupt(
		(HW_DEVICE_EXTENSION *)&adapter->DeviceExtension);

	spin_unlock_irqrestore(adapter->host_lock, flags);
	
	if (handled == AR_TRUE) {
		return IRQ_RETVAL(1);
	} else {
		return IRQ_RETVAL(0);
	}
	
}

/**
 * new error handler entry points 
 **/
INT
eh_megaide_abort(Scsi_Cmnd *scmd)
{
	UINT ret_val = 0;

	AR_DEBUG (AR_DBG_ENTRY,(KERN_DEBUG "eh_megaide_abort() Called\n"));
	
	ret_val = megaide_abort(scmd);

	return ((SCSI_ABORT_SUCCESS == ret_val)? SUCCESS: FAILED);
}

INT 
eh_megaide_reset_bus(Scsi_Cmnd *scmd)
{
	INT		ret_val = 0;

	AR_DEBUG (AR_DBG_ENTRY,(KERN_DEBUG "eh_megaide_reset_bus() Called\n"));
	
	ret_val= megaide_reset(scmd, 0);

	return ((SCSI_RESET_SUCCESS == ret_val)? SUCCESS: FAILED);
}

INT 
eh_megaide_reset_host(Scsi_Cmnd *scmd)
{
	AR_DEBUG(AR_DBG_ENTRY,(KERN_DEBUG "eh_megaide_reset_host() Called\n"));
	return eh_megaide_reset_bus(scmd);
}

INT eh_megaide_reset_dev(Scsi_Cmnd *scmd)
{
	AR_DEBUG(AR_DBG_ENTRY,(KERN_DEBUG "eh_megaide_reset_dev() Called\n"));
	return eh_megaide_reset_bus(scmd);
}

INT
mide_scsi_free (HW_DEVICE_EXTENSION *pdev, VOID *data, UINT length)
{
	mide_release_scb_pool_memory (pdev, (FAST_IDE_SCB_EXTENSION *) data);
	return TRUE;
}

VOID *
mide_scsi_malloc (HW_DEVICE_EXTENSION *pdev, UINT length)
{
	return ((VOID *)mide_get_scb_from_pool (pdev));
}

INT
mide_init_cmd_memory_pool(HW_DEVICE_EXTENSION *pdev)
{
	INT			i = 0;
	FAST_IDE_SCB_EXTENSION	*pScbExt;
	adapter_t		*adapter = (adapter_t *) pdev->adapter;
	
	INIT_LIST_HEAD (&adapter->scb_ext_free_list);
	
	for (i=0; i<M_IDE_TOTAL_COMMANDS; i++)
	{
		pScbExt = kmalloc (get_scb_extension_size(), GFP_KERNEL);
		
		list_add((struct list_head *)&pScbExt->list,
			&adapter->scb_ext_free_list);
	}
	return TRUE;
}

FAST_IDE_SCB_EXTENSION *
mide_get_scb_from_pool(HW_DEVICE_EXTENSION *pdev)
{
	adapter_t		*adapter = (adapter_t *) pdev->adapter;
	struct list_head	*head = &adapter->scb_ext_free_list;
	FAST_IDE_SCB_EXTENSION	*pScbExt = NULL;

	if (!list_empty(head)) {
	
		pScbExt = list_entry(head->next, FAST_IDE_SCB_EXTENSION, list);
		
		list_del_init(head->next);
	}
	return (pScbExt);
}

INT
mide_release_scb_pool_memory (HW_DEVICE_EXTENSION *pdev, 
	FAST_IDE_SCB_EXTENSION *pScbExt)
{
	adapter_t	*adapter = (adapter_t *) pdev->adapter;
	
	list_add ((struct list_head *)&pScbExt->list,
		&adapter->scb_ext_free_list);
	return TRUE;
}

INT
mide_free_cmd_memory_pool(HW_DEVICE_EXTENSION *pdev)
{
	struct list_head	*pos;
	struct list_head	*next;
	adapter_t		*adapter = (adapter_t *) pdev->adapter;
	FAST_IDE_SCB_EXTENSION	*pScbExt = NULL;

	list_for_each_safe(pos, next, &adapter->scb_ext_free_list) {
		
		pScbExt = list_entry(pos, FAST_IDE_SCB_EXTENSION, list);
		
		list_del_init((struct list_head *)&pScbExt->list);
		
		kfree (pScbExt);
	}
	return TRUE;
}

VOID
fill_sense_buffer(FAST_IDE_SCB_EXTENSION *scb , UCHAR errByte,
	BOOLEAN rwError , UINT errSector)
{
	Scsi_Cmnd	*scmd;

	scmd = (Scsi_Cmnd *) scb->pstScsiCmd;

	scmd->sense_buffer[0] = 0x70;
	scmd->sense_buffer[2] = MEDIUM_ERROR;
}


ULONG
GetPhysicalAddressAndLength(VOID *buffer, USHORT sglem,
	UINT offset, UINT *puiContiguousLength)
{
	ULONG			phyAddr;
	struct scatterlist	*sgl;
	UINT			curLen;
	UINT			sglInd;

	sgl			= (struct scatterlist *) buffer;
	*puiContiguousLength	= 0;
	phyAddr			= 0;

	for (sglInd=0, curLen = 0; sglInd < sglem; sglInd++)
	{
		if ((curLen + sgl[sglInd].length) > offset) {
			
			/* This is the SGL Element */
			*puiContiguousLength = sg_dma_len(
				&sgl[sglInd])-(offset - curLen);
			
			phyAddr = sg_dma_address(
				&sgl[sglInd])+(offset - curLen);
			break;
		}

		curLen += sg_dma_len(& sgl[sglInd]);
	}
	return phyAddr;
}

#ifdef _ATAPI_SUPPORT_
/**
 * mega_get_sgl_len_and_addr:
 * This function is used to get the virtual address of a scatter gather 
 * element and its length. 
 *
 * This function is required to provide drive independent shim driver source
 *
 * Parameters
 * @pdev	: SCSI midlayer given scatter gather list pointer
 * @sg_elem	: Required SG element number
 * @plen	: Data length request pointer
 * @return	:
 *	Physical address of the scatter gather element is returned.
 **/
ULONG
mega_get_sgl_len_and_addr(VOID *buffer, UINT sg_elem, UINT *plen)
{
	struct scatterlist	*sgl;

	sgl = (struct scatterlist *) buffer;

	/* pci_map_sg is already applied on scatter gather list so
	 * we can get dma address and lenth using following API's
	 */
	*plen = sg_dma_len(&sgl[sg_elem]);
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	return((ULONG)(page_address(sgl[sg_elem].page) + 
		sgl[sg_elem].offset));
#else
	if (sgl[sg_elem].address) {
		return((ULONG)(sgl[sg_elem].address));
	} else {
		return((ULONG)(page_address(sgl[sg_elem].page) + 
			sgl[sg_elem].offset));
	}
#endif
}


/**
 * mega_place_sense_data_and_result: place sense data and result in Scsi
 * command result and sense buffer.
 *
 * Here sense buffer is prepared with given status information and result
 * of the command is placed back in the SCSI command for SCSI midlayer to
 * proceed with further actions.
 * This function is required to provide drive independent shim driver source
 *
 * Parameters
 * @pdev	: SCSI midlayer given scatter gather list pointer
 * @scb		: MegaIDE SCSI command block pointer
 * @chan	: Channel number
 * @status	: Status of the command
 * @result	: result of the command
 * @return	:
 *	void
 **/
VOID
mega_place_sense_data_and_result(HW_DEVICE_EXTENSION *pdev,
	FAST_IDE_SCB_EXTENSION *scb, UINT chan, UCHAR status, INT result)
{
	IDE_REGISTERS_1		*base1 = pdev->Channel[chan].BaseIoAddress1;
	BM_REGISTERS		*baseBm= pdev->Channel[chan].BaseBmAddress;
	UCHAR			statusByte;
	Scsi_Cmnd		*scmd = scb->pstScsiCmd;
	char			cmd[12];

	result |= (DRIVER_SENSE << 24);

	//initialize sense buffer with zero
	memset_Rlsi (scmd->sense_buffer, 0, sizeof(scmd->sense_buffer));

	ScsiPortWritePortUchar(((ULONG)base1+1), 0x00);
	// fill rest of the task file registers
	ScsiPortWritePortUchar((ULONG)&base1->SectorCount, 0x00);
	ScsiPortWritePortUchar((ULONG)&base1->SectorNumber, 0x00);
	ScsiPortWritePortUchar((ULONG)&base1->CylinderHigh, 0x00);
	ScsiPortWritePortUchar((ULONG)&base1->CylinderLow, 0x12);

	// fire packet command
	ScsiPortWritePortUchar((ULONG)&base1->Command, 0xA0);

	WAIT_FOR_DRQ(base1, statusByte);

	if ( statusByte & IDE_STATUS_ERROR )
	{
		UCHAR ucStatusByte1, ucStatusByte2;
		ucStatusByte1 = ScsiPortReadPortUchar((ULONG)base1 + 1);
		ucStatusByte2 = ScsiPortReadPortUchar((ULONG)base1 + 2);
		scmd->sense_buffer[0] = 0x70;
		scmd->sense_buffer[2] = (status >> 4);
		scmd->result = result;
		goto complete;
	}

	memset_Rlsi(cmd, 0, 12);
	cmd[0]=0x03;
	cmd[4]=0x12;

	AR_WRITE_BUFFER(base1, cmd, 6);

	WAIT_FOR_DRQ(base1, statusByte);
	if (statusByte & IDE_STATUS_ERROR) {
		UCHAR ucStatusByte1, ucStatusByte2;

		ucStatusByte1 = ScsiPortReadPortUchar((ULONG)base1 + 1);
		ucStatusByte2 = ScsiPortReadPortUchar((ULONG)base1 + 2);
		scmd->sense_buffer[0] = 0x70;
		scmd->sense_buffer[2] = (status >> 4);
		scmd->result = result;
		goto complete;
	}

	AR_READ_BUFFER(base1, scmd->sense_buffer, (0x12/2));
	scmd->result = result;

complete:
	WAIT_ON_BUSY(base1, statusByte);
	GET_BASE_STATUS(base1, statusByte);
	ClearPCIInterupt(pdev, chan);

	if (IS_ICHxR_CONTROLLER(pdev)) {
		CLEAR_BM_INT(baseBm, status);
	}
}

/**
 * mega_scsi_to_atapi: Convert SCSI command to equivalent ATAPI commands
 *
 * This function is used to convert SCSI commands to equivalent ATAPI commands
 * it is used to give support to unsupported SCSI commands
 *
 * Parameters
 * @scb		: MegaIDE SCSI command block pointer
 * @return	:
 *	void
 **/
void
mega_scsi_to_atapi(FAST_IDE_SCB_EXTENSION *scb)
{
	switch (scb->scsi_cdb[0])
	{
		/* If we change these commands we can not 
		 * close / open tray!! So avoiding it, let the OS figure out 
		 * what to do, any way we are sending proper sense data back
		 */
		case MODE_SENSE :
		case MODE_SELECT :
			break;

		case FORMAT_UNIT:
			scb->scsi_cdb[0] = ATAPI_FORMAT_UNIT;
			break;
		default:
			printk_Rlsi("megaide: atapi cmd=0x%x not supported\n",
				scb->scsi_cdb[0]);
		break;
	}
}
#endif /*_ATAPI_SUPPORT_*/

#include "megaide_pci.c"
#include "megaide_ata.c"
#include "megaide_char.c"
#include "megaide_proc.c"
#include "megaide_kernel_lib.c"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_init(megaide_init);
module_exit(megaide_exit);
#else
static Scsi_Host_Template driver_template = MEGAIDE_TEMPLATE;
#include "../scsi_module.c"
#endif
	
/* vim: set ts=8 sw=8 tw=78 ai si: */

