/*
 * Linux MegaRAID IDEal Software RAID Stack
 * Copyright (c) 2001 LSI Logic Corporation, All Rights Reserved.
 *-----------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *-----------------------------------------------------------------------------
 * File Name	: megaide_ata.c
 */

BOOLEAN
SetTransferMode(HW_DEVICE_EXTENSION *pdev, UCHAR tid,
	MP_TRANSFER_MODE TransferMode)
{
	IDE_REGISTERS_1		*base1;
	BM_REGISTERS		*baseBm;
	UCHAR			status = 0;
	UCHAR			transfer_mode;

	/* this change is applicable only to 3x ircd */
	if ((pdev->ucIRCDType == IRCD_3X) && 
		(PCI_DEVICE_ID_CMD_3112 != GET_DEVICE_ID(pdev))) {
	
		/* to program drive as in enquiry Data*/
		UCHAR	tMode = GetInquiryTransferMode(pdev, tid);

		if (tMode >= MP_UdmaMode0) {
			TransferMode  =	MEGA_FIND_MIN(tMode, 
				pdev->PhysicalDrive[tid].TransferMode);
		}
	}

	/* This is the Fix for Maxtor drives not working 
	 * if DMA write is more than 62 secotrs.. This is a drive issue 
	 * and it is present with UDMA 6 mode. we are programming back to
	 * UDMA5
	 */
	if (PCI_DEVICE_ID_CMD_3112 == GET_DEVICE_ID(pdev)) {
		/* this change is applicable only Maxtor drives */
		if (!strncmp(pdev->PhysicalDrive[tid].IdentifyData.ModelNumber,
					"aMtxro", 6)) {
			if (MP_UdmaMode6 == TransferMode) {
				TransferMode = MP_UdmaMode5;
			}
		}
	}
	
	switch (TransferMode) {

	case MP_DmaMode0:
		transfer_mode = STM_MULTIWORD_DMA(0);
		break;

	case MP_DmaMode1:
		transfer_mode = STM_MULTIWORD_DMA(1);
		break;

	case MP_DmaMode2:
		transfer_mode = STM_MULTIWORD_DMA(2);
		break;

	case MP_UdmaMode0:
		transfer_mode = STM_UDMA(0);
		break;

	case MP_UdmaMode1:
		transfer_mode = STM_UDMA(1);
		break;

	case MP_UdmaMode2:
		transfer_mode = STM_UDMA(2);
		break;

	case MP_UdmaMode3:
		transfer_mode = STM_UDMA(3);
		break;

	case MP_UdmaMode4:
		transfer_mode = STM_UDMA(4);
		break;

	case MP_UdmaMode5:
		transfer_mode = STM_UDMA(5);
		break;

	case MP_UdmaMode6:
		transfer_mode = STM_UDMA(6);
		break;

	default:
		/* CURRENTLY we are not programming drives to PIO mode*/
		return FALSE;
		break;
	}

	base1 = (IDE_REGISTERS_1 *)pdev->Channel[tid >> 1].BaseIoAddress1;
	baseBm = (BM_REGISTERS *)pdev->Channel[tid >> 1].BaseBmAddress;
	
	/* Program Drive transfer mode */
	SELECT_DEVICE(base1, tid);
	ScsiPortWritePortUchar((ULONG)((UCHAR *)base1 + 1),
		FEATURE_SET_TRANSFER_MODE);
	ScsiPortWritePortUchar((ULONG)&(base1->SectorCount), transfer_mode);
	ScsiPortWritePortUchar((ULONG)&(base1->Command),
		IDE_COMMAND_SET_FEATURES);
	WAIT_ON_BASE_BUSY(base1, status);

	if ((status & IDE_STATUS_ERROR) || (status & IDE_STATUS_BUSY)) {
		printk_Rlsi(KERN_WARNING 
			"megaide: setfeatures TID-%d failed status=0x%x!!\n", 
			tid, status);
		return FALSE;
	}

	status = ScsiPortReadPortUchar((ULONG)&(baseBm->Status.AsUchar));
	status |= 0x60;
	ScsiPortWritePortUchar((ULONG)&(baseBm->Status.AsUchar), status);
	status = ScsiPortReadPortUchar((ULONG)&(baseBm->Status.AsUchar));

	/* Program PCI config space */
	pdev->prgrm_chip_or_drvs(pdev, tid, (UCHAR)TransferMode);

	pdev->PhysicalDrive[tid].TransferMode = TransferMode;
	return TRUE;
}


VOID
ClearPCIInterupt(HW_DEVICE_EXTENSION *pdev, UCHAR ucChannel)
{
	UCHAR		ucInterruptValue = 0;
	
	if ((GET_DEVICE_ID(pdev) == PCI_DEVICE_ID_CMD_648) ||
		(GET_DEVICE_ID(pdev) == PCI_DEVICE_ID_CMD_649)) {
		
		ucInterruptValue = (ucChannel ? 
			CHAN2_INTERRUPT : CHAN1_INTERRUPT) | 0x01;
		
		ScsiPortWritePortUchar(
			(ULONG)&(pdev->Channel[0].BaseBmAddress->Reserved),
			 ucInterruptValue);
	}
}


BOOLEAN
FoundInterrupt(HW_DEVICE_EXTENSION *pdev, UCHAR *pChan)
{
	BOOLEAN		retVal;

	retVal = pdev->find_intr(pdev, pChan);

	if (!retVal) {
		retVal = mide_handle_athena_pm_hp(pdev);
	}
	return retVal;
}


/* for all chipset support*/
BOOLEAN
EnableInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	IDE_REGISTERS_1	*base1;
	IDE_REGISTERS_2	*base2;
	BM_REGISTERS	*BMReg = NULL;
	UINT		uiDrive, chan;
	UCHAR		status;
	UCHAR		tid;

	/* clear drive interrupts*/
	for (chan = 0; chan < MAX_CHANNELS_PER_CONTROLLER; chan++)
	{
		base1 = (IDE_REGISTERS_1 *)pdev->Channel[chan].BaseIoAddress1;
		base2 = (IDE_REGISTERS_2 *)pdev->Channel[chan].BaseIoAddress2;

		for (uiDrive = 0; uiDrive < MAX_DRIVES_PER_CHANNEL; uiDrive++)
		{
			tid = (chan << 1) + uiDrive;
			if (!IS_IDE_DRIVE(pdev, tid)) {
				continue;
			}

			SELECT_DEVICE(base1, uiDrive);
			GET_BASE_STATUS(base1, status);
		}
	}

	/* clear drive interrupt*/
	for (chan = 0; chan < MAX_CHANNELS_PER_CONTROLLER; chan++)
	{
		BMReg = pdev->Channel[chan].BaseBmAddress;
		CLEAR_BM_INT(BMReg, status);
	}

	BMReg = pdev->Channel[0].BaseBmAddress;

	/* 
	 * Enable the Intr notification so that further interrupts are got.
	 * This is done because, there is no interrupt handler at the time
	 * before the actual registration of the Int. handler..
	 */

	return pdev->enable_intr(pdev);
}


/* This function dissables the controller interupt */
BOOLEAN
DisableInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	return pdev->disable_intr(pdev);
}


BOOLEAN
Is80PinCable(HW_DEVICE_EXTENSION *pdev, UCHAR chan)
{
	return pdev->is_80pin_cable(pdev, chan);
}

/* *
 ** CMD648 and 649 Chipset Related Functions
 **/
BOOLEAN
Found64xInterrupt(HW_DEVICE_EXTENSION *pdev, UCHAR *pChan)
{
	BM_REGISTERS	*baseBm;
	UCHAR		status;
	UINT		chan;
	UINT		uiIntChannelValue = 0;

	/* 
	 *	This interrupt status will show the
	 *	interrupt on two channels...(ie., per PCI Space
	 */
	for (chan = 0; chan < MAX_CHANNELS_PER_CONTROLLER; chan += 2) {
		baseBm = pdev->Channel[chan].BaseBmAddress;

		status = ScsiPortReadPortUchar((ULONG)&(baseBm->Reserved));

		if (status & CHAN1_INTERRUPT) {
			uiIntChannelValue |= (1 << chan);
		}

		if (status & CHAN2_INTERRUPT) {
			uiIntChannelValue |= (1 << (chan + 1));
		}
	}

	*pChan = uiIntChannelValue;

	if (uiIntChannelValue) {
		return TRUE;
	} else {
		return FALSE;
	}

}


static UCHAR auc64xUdmaModeValues[2][9] = {
	{0x0, 0x0, 0x0, 0x31, 0x21, 0x11, 0x25, 0x15, 0x05},	/* Master*/
	{0x0, 0x0, 0x0, 0xc2, 0x82, 0x42, 0x8a, 0x4a, 0x0a}	/* Slave*/
	};
	/* MDMA0, MDMA1, MDMA2, UDMA0, UMDA1, UDMA2, UDMA3, UDMA4, UDMA5*/
BOOLEAN
ProgramDriveFor64xChip(HW_DEVICE_EXTENSION *pdev,
	UCHAR tid, UCHAR TransferMode)
{
	UCHAR		ucOrMask, ucAndMask, ucValue, devNo;
	BM_REGISTERS	*bmBase;

	if (TransferMode > MP_UdmaMode5) {
		return FALSE;
	}

	devNo = (tid & 0x01);
	ucOrMask = auc64xUdmaModeValues[devNo][TransferMode - MP_DmaMode0];

	if (devNo) {
		ucAndMask = 0x35; 	/* if slave do not touch master bits*/
	} else {
		ucAndMask = 0xca; 	/* if master do not touch slave bits*/
	}

	bmBase = pdev->Channel[tid >> 1].BaseBmAddress;
	ucValue = ScsiPortReadPortUchar((ULONG)((UCHAR *)bmBase + 3));
	
	ucValue &= ucAndMask;
	ucValue |= ucOrMask;
	ScsiPortWritePortUchar((ULONG)((UCHAR *)bmBase + 3), ucValue);
	return(TRUE);
}


BOOLEAN
Enable64xInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	BM_REGISTERS *BMReg = NULL;
	UCHAR opcimcr;

	BMReg = pdev->Channel[0].BaseBmAddress;

	opcimcr = ScsiPortReadPortUchar((ULONG)((UCHAR *)BMReg + 1));
	opcimcr &= 0xCF;
	opcimcr |= 0x01;
	ScsiPortWritePortUchar((ULONG)((UCHAR *)BMReg + 1), opcimcr);
	return AR_TRUE;
}


BOOLEAN
Disable64xInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	BM_REGISTERS *BMReg = NULL;
	UCHAR opcimcr;

	BMReg = pdev->Channel[0].BaseBmAddress;

	/* 
	 * Enable Interrupt notification so that further interrupts are got.
	 * This is done because, there is no interrupt handler at the time
	 * before the actual registration of the Int. handler..
	 */
	opcimcr = ScsiPortReadPortUchar((ULONG)((UCHAR *)BMReg + 1));
	opcimcr |= 0x30;
	ScsiPortWritePortUchar((ULONG)((UCHAR *)BMReg + 1), opcimcr);

	return AR_TRUE;
}


BOOLEAN
Is64x80PinCable(HW_DEVICE_EXTENSION *pdev, UCHAR chan)
{
	UCHAR		ucIs80PinCable;
	
	pci_read_config_byte_Rlsi(pdev, 0x79, &ucIs80PinCable);
	if (chan) { /* secondary channel*/
		ucIs80PinCable = (ucIs80PinCable & 0x02) >> 1;
	} else { /* primary channel*/
		ucIs80PinCable = (ucIs80PinCable & 0x01);
	}
	return(ucIs80PinCable);
}


BOOLEAN
IsPostCommandToDriveSuccess64x(HW_DEVICE_EXTENSION *pdev,
		UINT tid, UINT uiStartBlock, UINT uiBlockCount,
		UCHAR ucCmd, ULONG ulSglPhysicalAddress)
{
	CHANNEL		*pChannel;
	IDE_REGISTERS_1	*base1;
	IDE_REGISTERS_2	*base2;
	BM_REGISTERS	*baseBm;
	UCHAR		ucStatus = 0;
	UCHAR	intVal = ((tid >> 1) ? CHAN2_INTERRUPT : CHAN1_INTERRUPT)|0x1;

	pChannel= &(pdev->Channel[(tid >> 1)]);
	base1	= (IDE_REGISTERS_1 *)pChannel->BaseIoAddress1;
	base2	= (IDE_REGISTERS_2 *)pChannel->BaseIoAddress2;
	baseBm	= pChannel->BaseBmAddress;

	PostCommandToDrive(pdev, tid, uiStartBlock,
		uiBlockCount, ucCmd, ulSglPhysicalAddress);
	
	POLL_FOR_INTERRUPT(pdev, intVal);
	POLL_FOR_DMA_INTERRUPT(baseBm)
	
	ScsiPortWritePortUchar((ULONG)&(baseBm->Command.AsUchar),
		STOP_TRANSFER);

	GET_BASE_STATUS(base1, ucStatus);

	/* Clear interrupt*/
	ClearPCIInterupt(pdev, tid >> 1);

	/* if no error continue;*/
	if (ucStatus & IDE_STATUS_BUSY) {
		return FALSE;
	}

	/* if no error continue*/
	if (ucStatus & IDE_STATUS_ERROR) {
		return FALSE;
	}
	
	return TRUE;
}


BOOLEAN
Found68xInterrupt(HW_DEVICE_EXTENSION *pdev, UCHAR *pChan)
{
	ULONG	baseBm;
	UCHAR	status;
	UINT	chan = 0;
	UINT	uiIntChannelValue = 0;

	baseBm = (ULONG)&(pdev->Channel[0].BaseBmAddress->Command.AsUchar);

	if(pdev->Channel[0].ExpectingInterrupt) {
		status = ScsiPortReadPortUchar((baseBm + 0x02));
		if ((status & 0x4) == 0x4) {
			uiIntChannelValue |= (1 << chan);
		}
	}

	if(pdev->Channel[1].ExpectingInterrupt) {
		status = ScsiPortReadPortUchar((baseBm + 0x0A));
		if ((status & 0x4) == 0x4) {
			uiIntChannelValue |= (1 << (chan + 1));
		}
	}

	*pChan = uiIntChannelValue;

	if (uiIntChannelValue) {
		return TRUE;
	} else {
		return FALSE;
	}
}


/* Refer Sil0680 Data Sheet for explanation of these values */
static UCHAR auc68xUdmaMode[7] = {0x0F,0x0B,0x07,0x05,0x03,0x02,0x01 };
static USHORT auc68xDmaMode[3] = { 0x2208,0x10C2,0x10C1};
static USHORT auc68xPIOMode[5] = { 0x328A,0x2283,0x1104,0x10C3,0x10C1};
static USHORT auc68xTFTiming[5]= { 0x328A,0x2283,0x1281,0x10C3,0x10C1};
BOOLEAN
ProgramDriveFor68xChip(HW_DEVICE_EXTENSION *pdev, UCHAR tid, UCHAR ucMode)
{
	UCHAR		ucReg = 0;
	UCHAR		tf_mode, pio_mode, mdma_mode, udma_mode;
	UCHAR		ucReg80 = 0, ucReg84 = 0;
	MP_DRIVE_ID	drive_id = (MP_DRIVE_ID) tid;

	/* program IDE CLK Select to 133 MHz*/
	pci_read_config_byte_Rlsi(pdev,SYS_CONF_STAT_CMD_REG680+2, &ucReg);
	ucReg = (ucMode >= MP_UdmaMode0) ?(ucReg | 0x10) :(ucReg & 0xEF);
	pci_write_config_byte_Rlsi(pdev,SYS_CONF_STAT_CMD_REG680+2, ucReg);

	/* PIO2 = MDMA0 PIO3 = MDMA1 PIO4 = MDMA2(and UDMAx)*/
	/* program TF Timing*/

	tf_mode = (ucMode < MP_UdmaMode0) ?
		(ucMode - MP_DmaMode0):(MP_DmaMode2 - MP_DmaMode0);
	tf_mode += 2;

	/* program PIO Timing*/
	pio_mode = (ucMode < MP_UdmaMode0) ?
		(ucMode - MP_DmaMode0) :(MP_DmaMode2 - MP_DmaMode0);
	pio_mode += 2;

	/* program mdma timing*/
	mdma_mode = (ucMode < MP_UdmaMode0) ?
		(ucMode - MP_DmaMode0) :(MP_DmaMode2 - MP_DmaMode0);

	/* program dma timing*/
	udma_mode = (ucMode >= MP_UdmaMode0) ?
		(ucMode - MP_UdmaMode0) : 6;

	/* set TF Timing*/
	if ((tid > 1) == 0) {
		pci_write_config_byte_Rlsi(pdev, IDE0_TFT_CONF_STAT_REG680+2,
			auc68xTFTiming[tf_mode]);
	} else {
		pci_write_config_byte_Rlsi(pdev, IDE1_TFT_CONF_STAT_REG680+2,
			auc68xTFTiming[tf_mode]);
	}

	/* get Transfer Mode registers values*/
	/* bits 1:0 - Device0 { 2=mdma, 3=udma }*/
	/* bits 5:4 - Device1 {2=mdma, 3=udma }*/
	pci_read_config_byte_Rlsi(pdev, DATA_TRFR_MODE_IDE0_REG680, &ucReg80);
	pci_read_config_byte_Rlsi(pdev, DATA_TRFR_MODE_IDE1_REG680, &ucReg84);

	ucReg80 &= ((tid & 1) ? 0xCF : 0xFC);
	ucReg84 &= ((tid & 1) ? 0xCF : 0xFC);

	if (ucMode >= MP_UdmaMode0) {
		ucReg80 |= ((tid & 1) ? 0x30 : 0x03);
		ucReg84 |= ((tid & 1) ? 0x30 : 0x03);
	} else if (ucMode >= MP_DmaMode0) {
		ucReg80 |= ((tid & 1) ? 0x20 : 0x02);
		ucReg84 |= ((tid & 1) ? 0x20 : 0x02);
	}

	switch (drive_id) {
	case PrimaryMaster :
		/* set PIO Timing*/
		pci_write_config_word_Rlsi(pdev, IDE0_PIO_TIMING_REG680,
			auc68xPIOMode[pio_mode]);

		/* set mdma timing*/
		pci_write_config_word_Rlsi(pdev, IDE0_DMA_TIMING_REG680,
			auc68xDmaMode[mdma_mode]);

		/* set udma timing*/
		pci_write_config_byte_Rlsi(pdev, IDE0_UDMA_TIMING_REG680,
			auc68xUdmaMode[udma_mode]);

		/* set tran mode*/
		pci_write_config_byte_Rlsi(pdev, DATA_TRFR_MODE_IDE0_REG680,
			ucReg80);

		break;


	case PrimarySlave :
		/* set PIO Timing*/
		pci_write_config_word_Rlsi(pdev, IDE0_PIO_TIMING_REG680+2,
			auc68xPIOMode[pio_mode]);

		/* set mdma timing*/
		pci_write_config_word_Rlsi(pdev, IDE0_DMA_TIMING_REG680+2,
			auc68xDmaMode[mdma_mode]);

		/* set udma timing*/
		pci_write_config_byte_Rlsi(pdev, IDE0_UDMA_TIMING_REG680+2,
			auc68xUdmaMode[udma_mode]);

		/* set tran mode*/
		pci_write_config_byte_Rlsi(pdev, DATA_TRFR_MODE_IDE0_REG680,
			ucReg80);

		break;


	case SecondaryMaster :
		/* set PIO Timing*/
		pci_write_config_word_Rlsi(pdev, IDE1_PIO_TIMING_REG680,
			auc68xPIOMode[pio_mode]);
		/* set mdma timing*/
		pci_write_config_word_Rlsi(pdev, IDE1_DMA_TIMING_REG680,
			auc68xDmaMode[mdma_mode]);
		/* set udma timing*/
		pci_write_config_byte_Rlsi(pdev, IDE1_UDMA_TIMING_REG680,
			auc68xUdmaMode[udma_mode]);
		/* set tran mode*/
		pci_write_config_byte_Rlsi(pdev, DATA_TRFR_MODE_IDE1_REG680,
			ucReg84);
		break;

	case SecondarySlave :
		/* set PIO Timing*/
		pci_write_config_word_Rlsi(pdev, IDE1_PIO_TIMING_REG680+2,
			auc68xPIOMode[pio_mode]);
		/* set mdma timing*/
		pci_write_config_word_Rlsi(pdev, IDE1_DMA_TIMING_REG680+2,
			auc68xDmaMode[mdma_mode]);
		/* set udma timing*/
		pci_write_config_byte_Rlsi(pdev, IDE1_UDMA_TIMING_REG680+2,
			auc68xUdmaMode[udma_mode]);
		/* set tran mode*/
		pci_write_config_byte_Rlsi(pdev, DATA_TRFR_MODE_IDE1_REG680,
			ucReg84);
		break;

	default :
		break;
	}
	
	return(TRUE);
}


BOOLEAN
Enable68xInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		opcimcr;

	pci_read_config_byte_Rlsi(pdev, SYS_CONF_STAT_CMD_REG680+2, &opcimcr);
	opcimcr &= 0x3F; /* 0011 111*/
	pci_write_config_byte_Rlsi(pdev, SYS_CONF_STAT_CMD_REG680+2, opcimcr);

	return AR_TRUE;
}


BOOLEAN
Disable68xInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		opcimcr;

	pci_read_config_byte_Rlsi(pdev, SYS_CONF_STAT_CMD_REG680+2, &opcimcr);
	opcimcr |= 0xC0; /* 1100 0000*/
	pci_write_config_byte_Rlsi(pdev, SYS_CONF_STAT_CMD_REG680+2, opcimcr);

	return AR_TRUE;
}


BOOLEAN
Is68x80PinCable(HW_DEVICE_EXTENSION *pdev, UCHAR chan)
{
	UCHAR		ucIs80PinCable;

	if (chan) {	/* secondary channel*/
		pci_read_config_byte_Rlsi(pdev, 0xB0, &ucIs80PinCable);
	} else {	/* primary channel*/
		pci_read_config_byte_Rlsi(pdev, 0xA0, &ucIs80PinCable);
	}
	
	return(ucIs80PinCable & 0x01);
}

BOOLEAN
IsPostCommandToDriveSuccess68x(HW_DEVICE_EXTENSION *pdev,
	UINT tid, UINT uiStartBlock, UINT uiBlockCount, UCHAR ucCmd,
	ULONG ulSglPhysicalAddress)
{
	CHANNEL		*pChannel;
	IDE_REGISTERS_1	*base1;
	IDE_REGISTERS_2	*base2;
	BM_REGISTERS	*baseBm;
	UCHAR		ucStatus = 0;

	pChannel = &(pdev->Channel[(tid >> 1)]);
	base1 = (IDE_REGISTERS_1 *)pChannel->BaseIoAddress1;
	base2 = (IDE_REGISTERS_2 *)pChannel->BaseIoAddress2;
	baseBm = pChannel->BaseBmAddress;

	PostCommandToDrive(pdev, tid, uiStartBlock, uiBlockCount,
		ucCmd, ulSglPhysicalAddress);

	POLL_FOR_INTERRUPT_68X(pdev, tid);
	POLL_FOR_DMA_INTERRUPT(baseBm)

	ScsiPortWritePortUchar((ULONG)&(baseBm->Command.AsUchar),
		STOP_TRANSFER);

	GET_BASE_STATUS(base1, ucStatus);

	/* Clear interrupt*/
	ClearPCIInterupt(pdev, tid >> 1);

	if (ucStatus & IDE_STATUS_BUSY) {
		printk_Rlsi("megaide: 680 intr busy bit set tid=%d\n", tid);
		return FALSE;
	}

	/* if no error continue;*/
	if (ucStatus & IDE_STATUS_ERROR){
		UCHAR		err;
		
		printk_Rlsi("megaide: 680 intr err bit set tid=%d\n", tid);
		err = ScsiPortReadPortUchar((ULONG)((UCHAR *)base1 + 1));
		printk_Rlsi("megaide: error byte = 0x%x\n", err);
		return FALSE;
	}

	return TRUE;
}

BOOLEAN
IsPostCommandToDriveSuccessCSB6(HW_DEVICE_EXTENSION *pdev,
	UINT tid, UINT uiStartBlock, UINT uiBlockCount,
	UCHAR ucCmd, ULONG ulSglPhysicalAddress)
{
	CHANNEL		*pChannel;
	IDE_REGISTERS_1	*base1;
	IDE_REGISTERS_2	*base2;
	BM_REGISTERS	*baseBm;
	UCHAR		ucStatus = 0;

	pChannel= &(pdev->Channel[(tid >> 1)]);
	base1	= (IDE_REGISTERS_1 *)pChannel->BaseIoAddress1;
	base2	= (IDE_REGISTERS_2 *)pChannel->BaseIoAddress2;
	baseBm	= pChannel->BaseBmAddress;

	PostCommandToDrive(pdev, tid, uiStartBlock, uiBlockCount,
		ucCmd, ulSglPhysicalAddress);

	POLL_FOR_INTERRUPT_CSB5(pdev, tid);
	POLL_FOR_DMA_INTERRUPT(baseBm)

	ScsiPortWritePortUchar((ULONG)&(baseBm->Command.AsUchar),
		STOP_TRANSFER);

	GET_BASE_STATUS(base1, ucStatus);


	if (ucStatus & IDE_STATUS_BUSY) {
		return FALSE;
	}

	/* if no error continue;*/
	if (ucStatus & IDE_STATUS_ERROR) {
		return FALSE;
	}

	return TRUE;
}


BOOLEAN
FoundCSB6Interrupt(HW_DEVICE_EXTENSION *pdev, UCHAR *pChan)
{
	UCHAR			status;
	UINT			chan = 0;
	UINT			uiIntChannelValue = 0;
	PBM_REGISTERS		baseBm;
	PIDE_REGISTERS_2	baseIoAddress2;

	for (chan = 0; chan < 2; chan++)
	{
	baseBm		= pdev->Channel[chan].BaseBmAddress;
	baseIoAddress2	= pdev->Channel[chan].BaseIoAddress2;
	
#ifdef _ATAPI_SUPPORT_ 
	if ((pdev->Channel[chan].ExpectingInterrupt == IDE_DMA_INTERRUPT)||
		(pdev->Channel[chan].ExpectingInterrupt ==
		ATAPI_DMA_INTERRUPT)) {
#else		
	if (pdev->Channel[chan].ExpectingInterrupt == IDE_DMA_INTERRUPT) {
#endif /*_ATAPI_SUPPORT_*/
		status = ScsiPortReadPortUchar((ULONG)&(baseBm->Status));
		if ((status & BM_STATUS_INTERRUPT) || 
			(status & BM_STATUS_ERROR)) {
			uiIntChannelValue |= (1 << chan);
			continue;
		}
	}

#ifdef _ATAPI_SUPPORT_ 
	if ((pdev->Channel[chan].ExpectingInterrupt == IDE_PIO_INTERRUPT)||
		(pdev->Channel[chan].ExpectingInterrupt ==
		ATAPI_PIO_INTERRUPT)) {
#else		
	if (pdev->Channel[chan].ExpectingInterrupt == IDE_PIO_INTERRUPT) {
#endif /*_ATAPI_SUPPORT_*/
		status = ScsiPortReadPortUchar((ULONG)&(baseBm->Status));
		if (!(status & IDE_STATUS_BUSY)) {
			uiIntChannelValue |= (1 << chan);
			continue;
		}
	}
	} /* for loop*/

	*pChan = uiIntChannelValue;

	if (uiIntChannelValue) {
		return TRUE;
	} else {
		return FALSE;
	}
}


BOOLEAN
EnableCSB6Interrupts(HW_DEVICE_EXTENSION *pdev)
{
	/* Enable interrupts on both channels */
	ScsiPortWritePortUchar((ULONG)&
		pdev->Channel[0].BaseIoAddress2->AlternateStatus,
		IDE_DC_REENABLE_CONTROLLER);
	ScsiPortWritePortUchar((ULONG)&
		pdev->Channel[1].BaseIoAddress2->AlternateStatus,
		IDE_DC_REENABLE_CONTROLLER);
	return AR_TRUE;
}


BOOLEAN
DisableCSB6Interrupts(HW_DEVICE_EXTENSION *pdev)
{
	/* disable interrupts on both channels */
	ScsiPortWritePortUchar((ULONG)&
		pdev->Channel[0].BaseIoAddress2->AlternateStatus, 
		IDE_DC_DISABLE_INTERRUPTS);
	ScsiPortWritePortUchar((ULONG)&
		pdev->Channel[1].BaseIoAddress2->AlternateStatus,
		IDE_DC_DISABLE_INTERRUPTS);
	return AR_TRUE;
}


BOOLEAN
IsCSB680PinCable(HW_DEVICE_EXTENSION *pdev, UCHAR chan)
{
	/* always expect a 80 pin cable*/
	return 0x01;
}


BOOLEAN
ProgramDriveForCSB6Chip(HW_DEVICE_EXTENSION *pdev, UCHAR tid, UCHAR ucMode)
{
	return TRUE;
}


/* *
 ** Silicon Image Athena Related Functions
 **/
BOOLEAN
EnableAthenaInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		opcimcr;

	pci_read_config_byte_Rlsi(pdev, SYS_CONF_STAT_CMD_REG680+2, &opcimcr);
	opcimcr &= 0x3F; /* 0011 111*/
	pci_write_config_byte_Rlsi(pdev, SYS_CONF_STAT_CMD_REG680+2, opcimcr);

	AthenaEnablePMInterrupts(pdev);
	AthenaEnableHPInterrupt(pdev);
	return AR_TRUE;
}


BOOLEAN
DisableAthenaInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		opcimcr;

	pci_read_config_byte_Rlsi(pdev, SYS_CONF_STAT_CMD_REG680+2, &opcimcr);
	opcimcr |= 0xC0; /* 1100 0000*/
	pci_write_config_byte_Rlsi(pdev, SYS_CONF_STAT_CMD_REG680+2, opcimcr);

	AthenaDisablePMInterrupts(pdev);
	AthenaDisableHPInterrupt(pdev);
	return AR_TRUE;

}


/* *
 * FoundICH5RInterrupt: Found ICH5R interrupt
 * This function checks whether any ICH5R channel raised any interrupt.
 * This check is done only if Driver expecting an interrupt from that
 * channel.
 *
 * Function parameters
 * @pdev	:	ICH5R device extension structure pointer
 * @p_int_return_val:	Interrupt detect status
 *		return 0x01 if interrupt detect in primary channel
 *		return 0x02 if interrupt detect in Secondary channel
 *		return 0x03 if interrupt detect in both channel
 * @return;
 *	TRUE if interrupt detected
 *	FALSE if no interrupt detected
 **/
BOOLEAN
FoundICH5RInterrupt(HW_DEVICE_EXTENSION *pdev, UCHAR *p_int_return_val)
{
	ULONG		baseBm;

	*p_int_return_val = 0;
	baseBm = (ULONG) &(pdev->Channel[0].BaseBmAddress->Command.AsUchar);

	if(pdev->Channel[0].ExpectingInterrupt) {
		if ((ScsiPortReadPortUchar((baseBm + 0x02)) & 0x4) == 0x4) {
			*p_int_return_val |= 0x01; /* found chan 1 interrupt */
		}
	}

	if(pdev->Channel[1].ExpectingInterrupt) {
		if ((ScsiPortReadPortUchar((baseBm + 0x0A)) & 0x4) == 0x4) {
			*p_int_return_val |= 0x02; /* found chan 2 interrupt */
		}
	}

	return ((*p_int_return_val) ? TRUE: FALSE);
}


/* *
 * ProgramDriveForICH5RChip: Program drives connected to ICH5R controller
 * This function supposed to program a drive connected to thecontroller.
 * It is used if bios is not programming the drives and in case of newly.
 *
 * Function parameters
 * @pdev	:	ICH5R device extension structure pointer
 * @tid		:	Target ID
 *		00h for primary master
 *		02h for secondary master
 * @ucMode	: Trasnsfer mode. it is defined
 *		MP_TRANSFER_MODE definition gives the values and meaning
 * @return;
 *	TRUE if succeed
 *	FALSE if failed
 **/
BOOLEAN
ProgramDriveForICH5RChip(HW_DEVICE_EXTENSION *pdev, UCHAR tid, UCHAR ucMode)
{
	return AR_TRUE;
}


/* *
 * EnableICH5RInterrupts: Enable ICH5R Interrupt
 * This function enables the ICH5R interrupts by programming pci config space.
 *
 * Function parameters
 * @pdev	:	ICH5R device extension structure pointer
 *
 * @return;
 *	TRUE if succeed
 *	FALSE if failed
 **/
BOOLEAN
EnableICH5RInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		reg_val;

	/* reading interrupt control register */
	pci_read_config_byte_Rlsi(pdev, 0x05, &reg_val);

	/* Enabling interrupt by writing 0 to 3rd bit */
	reg_val ^= 0x4;

	/* writing interrupt control register */
	pci_write_config_byte_Rlsi(pdev, 0x05, reg_val);

	ScsiPortStallExecution(1000);

	return AR_TRUE;
}


/* *
 * DisableICH5RInterrupts: Disable ICH5R Interrupt
 * This function Disables the ICH5R interrupts by programming pci config space.
 *
 * Function parameters
 * @pdev	:	ICH5R device extension structure pointer
 *
 * @return;
 *	TRUE if succeed
 *	FALSE if failed
 **/
BOOLEAN
DisableICH5RInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		reg_val;

	/* reading interrupt control register */
	pci_read_config_byte_Rlsi(pdev, 0x05, &reg_val);

	/* disabling interrupt by writing 1 to 3rd bit */
	reg_val |= 0x4;

	/* writing interrupt control register */
	pci_write_config_byte_Rlsi(pdev, 0x05, reg_val);

	ScsiPortStallExecution(1000);

	return AR_TRUE;
}


/* *
 * IsICH5R80PinCable: Is This device using 80 pin cable?
 * This function will always return TRUE because this return value is used
 * for other ATA specific programming.
 *
 * Function parameters
 * @pdev	:	ICH5R device extension structure pointer
 * @ucChannelId	:	Channel ID 00h-Primary, 01h-secondary
 * @return;
 *	TRUE always
 **/
BOOLEAN
IsICH5R80PinCable(HW_DEVICE_EXTENSION *pdev, UCHAR ucChannelId)
{
	return AR_TRUE;
}


/* *
 * IsPostCommandToDriveSuccessICH5R: Post a Read Or Write Command to ICH5R Drive
 * This function is used to issue a read/write command to the drives
 * connected to ICH5R controller. This function is used to write/read
 * Errorlog, IRCD, Enquiry etc
 *
 * Function parameters
 * @pdev	:	ICH5R device extension structure pointer
 * @tid		:	Target ID 00h - primary master, 02h - secondary master
 * @uiStartBlock:	Start Logica Block address
 * @uiBlockCount:	Number of Blocks
 * @ucCmd	:	ATA read/write command READ_6/WRITE_6
 * @ulSglPhysicalAddress: Physical address of the SGList used for the command
 * @return;
 *	TRUE always
 **/
BOOLEAN
IsPostCommandToDriveSuccessICH5R(HW_DEVICE_EXTENSION *pdev, UINT tid,
	UINT uiStartBlock, UINT uiBlockCount,
	UCHAR ucCmd, ULONG ulSglPhysicalAddress)
{
	CHANNEL		*pChannel;
	IDE_REGISTERS_1	*base1;
	IDE_REGISTERS_2	*base2;
	BM_REGISTERS	*baseBm;
	UCHAR		ucStatus = 0;
	UCHAR		ucStatus2;

	pChannel= &(pdev->Channel[(tid >> 1)]);
	base1	= (IDE_REGISTERS_1 *)pChannel->BaseIoAddress1;
	base2	= (IDE_REGISTERS_2 *)pChannel->BaseIoAddress2;
	baseBm	= pChannel->BaseBmAddress;

	PostCommandToDrive(pdev, tid, uiStartBlock, uiBlockCount,
		ucCmd, ulSglPhysicalAddress);

	POLL_FOR_DMA_INTERRUPT(baseBm)

	ScsiPortWritePortUchar((ULONG)&(baseBm->Command.AsUchar),
		STOP_TRANSFER);

	GET_BASE_STATUS(base1, ucStatus);

	CLEAR_BM_INT(baseBm, ucStatus2);

	if (ucStatus & IDE_STATUS_BUSY) {
		printk_Rlsi("megaide: ICH5R intr busy bit set tid=%d\n", tid);
		return FALSE;
	}

	/* if no error continue;*/
	if (ucStatus & IDE_STATUS_ERROR) {
		UCHAR		err;
		printk_Rlsi("megaide: ICH5R intr err bit set tid=%d\n", tid);
		err = ScsiPortReadPortUchar((ULONG)((UCHAR *)base1 + 1));
		printk_Rlsi("megaide: error byte = 0x%x\n", err);
		return FALSE;
	}

	return TRUE;
}


/* *
 * ICH5R_reset_all_channels: reset both ICH5R channels
 * This function enable and dissable both channels in ICH5R to make sure 
 * Port control status register is updated with drive presence status. This 
 * is done by dissabling and enabling both channels.
 *
 * Function parameters
 * @pdev	:	ICH5R device extension structure pointer
 * @return	:	VOID
 **/
VOID
ICH5R_reset_all_channels(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		status;
	INT		i;

	/* reading port control and status register */
	pci_read_config_byte_Rlsi(pdev, 0x92, &status);

	/* disabling both channels */
	status ^= 0x03;
	pci_write_config_byte_Rlsi(pdev, 0x92, status);
	
	/* as per intel 400 to 500 milli second delay is required between
	 * disabling and enabling events of ports
	 */
	for (i = 0; i < 500; i++) {
		ScsiPortStallExecution(1000);
	}

	/* enabling both channels */
	status |= 0x03;
	pci_write_config_byte_Rlsi(pdev, 0x92, status);
	
	for (i = 0; i < 500; i++) {
		ScsiPortStallExecution(1000);
	}
}

/**
 * mega_ICH5R_poll_thread: ICH5R polling thread
 * This is a kernel thread that detects drive insertion in ICH5R. Thread starts
 * as soon as the driver loads and it will stop only when driver unloads or
 * when system shut down
 *
 * Polling time interval of this thread is 2 Sec
 *
 * Thread will detect only the drive insertion on a channel where drive was
 * not connected previously and it sets the respective flags for the same.
 * MegaIDE driver's drive insertion removal module will process this event
 * later.
 *
 * Access of the generic data structure in this thread should be SMP proof to
 * make sure a smooth operation in an SMP system.
 *
 * Parameters
 * @pdev	: Driver data structure pointer
 * @return	: VOID
 **/
VOID 
mega_ICH5R_poll_thread(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR			mask = 0;
	UCHAR			status;
	unsigned long		flags;
	struct Scsi_Host	*host;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	wait_queue_head_t	wait_queue_ICHxR;
#endif
	UCHAR			dissable_flag = 0;
	adapter_t		*adp = (adapter_t *)pdev->adapter;

	host = adp->pstScsiHost;

	/* Check if thread already active/running?*/
	if (pdev->ICHxR_thread_flag)
		return;


#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	lock_kernel();
#endif	
	daemonize("megaide");
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	unlock_kernel();
#endif
#else
	/*
	 * We only listen to signals if the HA was loaded as a module.
	 * If the HA was compiled into the kernel, then we don't listen
	 * to any signals.
	 */
	if( host->loaded_as_module ) {
		siginitsetinv(&current->blocked, SHUTDOWN_SIGS);
	} else {
		siginitsetinv(&current->blocked, 0);
	}

	/* Flush resources */
	daemonize();
	reparent_to_init();

	/* Set the name of this process. */
	sprintf(current->comm, "megaide");
#endif

	pdev->ICHxR_thread_flag = 1;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	init_waitqueue_head(&wait_queue_ICHxR);
#endif
	while(1)
	{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
		schedule();

		/* waiting for 2 sec, polling time is 2 Sec */
		interruptible_sleep_on_timeout(&wait_queue_ICHxR, (HZ*2));
#else		
		msleep(2000);
#endif 
		/* check if thread termination received? */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
		if (!pdev->ICHxR_thread_flag) {
			break;
		}
#else
		if( host->loaded_as_module ) {
			if(signal_pending(current))
				break;
		}
#endif
		/*
		 * Check if both drives are present?. If yes terminate thread
		 * Disable/enable channel with out drive.
		 */

		mask = 0;

		/* checking if primary master present already */
		if (!pdev->PhysicalDrive[0].ucIsPowerConnected)
			mask |= 0x01;

		/* checking if secondary master present already */
		if (!pdev->PhysicalDrive[2].ucIsPowerConnected)
			mask |= 0x02;

		if (!mask || pdev->drive_inserted.u.Drives) {
			/*if LD is not present, detect from here itself */
			if (!pdev->uiLogicalDriveCount && 
				pdev->drive_inserted.u.Drives) {	
				spin_lock_irqsave(adp->host_lock, flags);
				mega_check_drive_insertion_removal(pdev);
				spin_unlock_irqrestore(adp->host_lock, flags);
			}
			continue;
		}

		spin_lock_irqsave(adp->host_lock, flags);
		
		/* reading port control and status register */
		pci_read_config_byte_Rlsi(pdev, 0x92, &status);
		if (dissable_flag == 0) {
			status ^= mask;
			pci_write_config_byte_Rlsi(pdev, 0x92, status);
			dissable_flag = 1;
			AR_DEBUG(AR_HOT_PLUG, (KERN_DEBUG 
				"megaide: dissabled chan mask= 0x%x\n",
				mask));
			spin_unlock_irqrestore(adp->host_lock, flags);
			continue;
		} else if (dissable_flag == 1) {
			status |= mask;
			pci_write_config_byte_Rlsi(pdev, 0x92, status);
			dissable_flag = 2;
			AR_DEBUG(AR_HOT_PLUG, (KERN_DEBUG 
				"megaide: enabled chan mask= 0x%x\n",
				mask));
			spin_unlock_irqrestore(adp->host_lock, flags);
			continue;
		} else {
			dissable_flag = 0;
			AR_DEBUG(AR_HOT_PLUG, (KERN_DEBUG 
				"megaide: detecting status=0x%x\n",
				status));
		}

		/* Check only for the drive which is not present. */	 
		/* checking the presence of primary master */
		if (status & 0x10) {
			if (!pdev->PhysicalDrive[0].ucIsPowerConnected) {
				pdev->drive_inserted.u.Drives |= 1;
				AR_DEBUG(AR_HOT_PLUG, (KERN_DEBUG 
					"megaide: drv0 inserted! stat=0x%x\n",
					status));
			}
		}

		/* checking the presence of secondary master */
		if (status & 0x20) {
			if (!pdev->PhysicalDrive[2].ucIsPowerConnected) {
				pdev->drive_inserted.u.Drives |= (1 <<2);
				AR_DEBUG(AR_HOT_PLUG, (KERN_DEBUG 
					"megaide: drv2 inserted! stat=0x%x\n",
					status));
			}
		} 

		spin_unlock_irqrestore(adp->host_lock, flags);
	}

	/*
	 * If anyone is waiting for me to let them know I am on way out!
	 */
	pdev->ICHxR_thread_flag = 0;
	complete_and_exit(&((adapter_t *)pdev->adapter)->ICHxR_completion, 0);
}

/* ICH6R related functions */
/**
 * FoundICH6RInterrupt: Found ICH6R interrupt
 * This function checks whether any ICH6R channel raised any interrupt.
 * This check is done only if Driver expecting an interrupt from that
 * channel.
 *
 * Function parameters
 * @pdev	:	ICH6R device extension structure pointer
 * @p_int_return_val:	Interrupt detect status
 *		return 0x01 if interrupt detect in primary channel
 *		return 0x02 if interrupt detect in Secondary channel
 *		return 0x03 if interrupt detect in both channel
 * @return;
 *	TRUE if interrupt detected
 *	FALSE if no interrupt detected
 **/
BOOLEAN
FoundICH6RInterrupt(HW_DEVICE_EXTENSION *pdev, UCHAR *pChan)
{
	ULONG	baseBm =
		(ULONG)&(pdev->Channel[0].BaseBmAddress->Command.AsUchar);

	if(pdev->Channel[0].ExpectingInterrupt) {
		if (((ScsiPortReadPortUchar(baseBm + 0x02)) & 0x4) == 0x4) {
			*pChan |= 0X01;
		}
	}

	if(pdev->Channel[1].ExpectingInterrupt) {
		if (((ScsiPortReadPortUchar(baseBm + 0x0A)) & 0x4) == 0x4) {
			*pChan |= 0X02;
		}
	}

	if (*pChan) {
		return TRUE;
	} else {
		return FALSE;
	}
}


/**
 * EnableICH6RInterrupts: Enable ICH6R Interrupt
 * This function enables the ICH6R interrupts by programming pci config space.
 *
 * Function parameters
 * @pdev	:	ICH6R device extension structure pointer
 *
 * @return;
 *	TRUE if succeed
 *	FALSE if failed
 **/
BOOLEAN
EnableICH6RInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		reg_val;

	/* reading interrupt control register */
	pci_read_config_byte_Rlsi(pdev, 0x05, &reg_val);

	/* Enabling interrupt by writing 0 to 3rd bit */
	reg_val ^= 0x4;

	/* writing interrupt control register */
	pci_write_config_byte_Rlsi(pdev, 0x05, reg_val);

	ScsiPortStallExecution(1000);

	return AR_TRUE;
}

/**
 * DisableICH6RInterrupts: Disable ICH6R Interrupt
 * This function Disables the ICH6R interrupts by programming pci config space.
 *
 * Function parameters
 * @pdev	:	ICH6R device extension structure pointer
 *
 * @return;
 *	TRUE if succeed
 *	FALSE if failed
 **/
BOOLEAN
DisableICH6RInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		reg_val;

	/* reading interrupt control register */
	pci_read_config_byte_Rlsi(pdev, 0x05, &reg_val);

	/* disabling interrupt by writing 1 to 3rd bit */
	reg_val |= 0x4;

	/* writing interrupt control register */
	pci_write_config_byte_Rlsi(pdev, 0x05, reg_val);

	ScsiPortStallExecution(1000);

	return AR_TRUE;
}


/**
 * IsICH6R80PinCable: Is This device using 80 pin cable?
 * This function will always return TRUE because this return value is used
 * for other ATA specific programming.
 *
 * Function parameters
 * @pdev	:	ICH6R device extension structure pointer
 * @ucChannelId	:	Channel ID 00h-Primary, 01h-secondary
 * @return;
 *	TRUE always
 **/
BOOLEAN
IsICH6R80PinCable(HW_DEVICE_EXTENSION *pdev, UCHAR ucChannelId)
{
	return AR_TRUE;
}

/**
 * ProgramDriveForICH6RChip: Program drives connected to ICH6R controller
 * This function supposed to program a drive connected to thecontroller.
 * It is used if bios is not programming the drives and in case of newly.
 *
 * Function parameters
 * @pdev	:	ICH6R device extension structure pointer
 * @tid		:	Target ID
 *		00h for primary master
 *		02h for secondary master
 * @ucMode	: Trasnsfer mode. it is defined
 *		MP_TRANSFER_MODE definition gives the values and meaning
 * @return;
 *	TRUE if succeed
 *	FALSE if failed
 **/
BOOLEAN
ProgramDriveForICH6RChip(HW_DEVICE_EXTENSION *pdev, UCHAR tid, UCHAR ucMode)
{
	AR_DEBUG(AR_DBG_ICH6R, 
		("ProgramDriveForICH6RChip not implemented\n"));
	return AR_TRUE;
}


/**
 * IsPostCommandToDriveSuccessICH6R: Post a Read Or Write Command to ICH6R Drive
 * This function is used to issue a read/write command to the drives
 * connected to ICH6R controller. This function is used to write/read
 * Errorlog, IRCD, Enquiry etc
 *
 * Function parameters
 * @pdev	:	ICH6R device extension structure pointer
 * @tid		:	Target ID 00h - primary master, 02h - secondary master
 * @uiStartBlock:	Start Logica Block address
 * @uiBlockCount:	Number of Blocks
 * @ucCmd	:	ATA read/write command READ_6/WRITE_6
 * @ulSglPhysicalAddress: Physical address of the SGList used for the command
 * @return;
 *	TRUE always
 **/
BOOLEAN
IsPostCommandToDriveSuccessICH6R(HW_DEVICE_EXTENSION *pdev, UINT tid,
	UINT uiStartBlock, UINT uiBlockCount,
	UCHAR ucCmd, ULONG ulSglPhysicalAddress)
{
	CHANNEL		*pChannel;
	IDE_REGISTERS_1	*base1;
	IDE_REGISTERS_2	*base2;
	BM_REGISTERS	*baseBm;
	UCHAR		ucStatus = 0;
	UCHAR		ucStatus2;

	pChannel= &(pdev->Channel[(tid >> 1)]);
	base1	= (IDE_REGISTERS_1 *)pChannel->BaseIoAddress1;
	base2	= (IDE_REGISTERS_2 *)pChannel->BaseIoAddress2;
	baseBm	= pChannel->BaseBmAddress;

	PostCommandToDrive(pdev, tid, uiStartBlock, uiBlockCount,
		ucCmd, ulSglPhysicalAddress);

	POLL_FOR_DMA_INTERRUPT(baseBm)

	ScsiPortWritePortUchar((ULONG)&(baseBm->Command.AsUchar),
		STOP_TRANSFER);

	GET_BASE_STATUS(base1, ucStatus);

	CLEAR_BM_INT(baseBm, ucStatus2);

	if (ucStatus & IDE_STATUS_BUSY) {
		printk_Rlsi(KERN_WARNING 
			"ICH6R intr busy bit set tid=%d\n", tid);
		return FALSE;
	}

	/* if no error continue;*/
	if (ucStatus & IDE_STATUS_ERROR) {
		UCHAR		err;
		printk_Rlsi(KERN_WARNING 
			"ICH6R intr err bit set tid=%d\n", tid);
		err = ScsiPortReadPortUchar((ULONG)((UCHAR *)base1 + 1));
		printk_Rlsi(KERN_WARNING "error byte = 0x%x\n", err);
		return FALSE;
	}

	return TRUE;
}


/**
 * mega_ICH6R_poll_thread: ICH6R polling thread
 * This is a kernel thread that detects drive insertion in ICHxR. Thread starts
 * as soon as the driver loads and it will stop only when driver unloads or
 * when system shut down
 *
 * Polling time interval of this thread is 2 Sec
 *
 * Thread will detect only the drive insertion on a channel where drive was
 * not connected previously and it sets the respective flags for the same.
 * MegaIDE driver's drive insertion removal module will process this event
 * later.
 *
 * Access of the generic data structure in this thread should be SMP proof to
 * make sure a smooth operation in an SMP system.
 *
 * Parameters
 * @pdev	: Driver data structure pointer
 * @return	: VOID
 **/

#define	DEVICE_DETECTED_PHY_COMM_OK		0x03

VOID 
mega_ICH6R_poll_thread(HW_DEVICE_EXTENSION *pdev)
{
	unsigned long		flags;
	struct Scsi_Host	*host;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	wait_queue_head_t	wait_queue_ICHxR;
#endif	
	adapter_t		*adp = (adapter_t *)pdev->adapter;
	/* ich6r drive detection */
	ULONG	ulOffsetArray[4] = 
		{0x128/*PM*/, 0x228/*PS*/,
		 0x1a8/*SM*/, 0x2a8/*SS*/};
	UCHAR	tid;
	UINT	uiSATAStatus; 	

	host = adp->pstScsiHost;

	/* Check if thread already active/running?*/
	if (pdev->ICHxR_thread_flag)
		return;


#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	lock_kernel();
#endif	
	daemonize("megaide");
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	unlock_kernel();
#endif	
#else
	/*
	 * We only listen to signals if the HA was loaded as a module.
	 * If the HA was compiled into the kernel, then we don't listen
	 * to any signals.
	 */
	if( host->loaded_as_module ) {
		siginitsetinv(&current->blocked, SHUTDOWN_SIGS);
	} else {
		siginitsetinv(&current->blocked, 0);
	}

	/* Flush resources */
	daemonize();
	reparent_to_init();

	/* Set the name of this process. */
	sprintf(current->comm, "megaide");
#endif

	pdev->ICHxR_thread_flag = 1;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	init_waitqueue_head(&wait_queue_ICHxR);
#endif

	while(1)
	{

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
		schedule();

		/* waiting for 2 sec, polling time is 2 Sec */
		interruptible_sleep_on_timeout(&wait_queue_ICHxR, (HZ*2));
#else		
		msleep(2000);
#endif 

		/* check if thread termination received? */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
		if (!pdev->ICHxR_thread_flag) {
			break;
		}
#else
		if( host->loaded_as_module ) {
			if(signal_pending(current))
				break;
		}
#endif
		 
		if (pdev->drive_inserted.u.Drives ||
			pdev->drive_removed.u.Drives) {
			/*if LD present, detect from here itself */
			if (!pdev->uiLogicalDriveCount) {	
				spin_lock_irqsave(adp->host_lock, flags);
				mega_check_drive_insertion_removal(pdev);
				spin_unlock_irqrestore(adp->host_lock, flags);
			}
			continue;
		}

		spin_lock_irqsave(adp->host_lock, flags);
		
		for(tid = 0; tid < 4; tid++)
		{
			uiSATAStatus = ScsiPortReadRegisterUlong(
				(ULONG)(pdev->virtBaseAddress5 + 
					ulOffsetArray[tid]));	

			if (((uiSATAStatus & 0x0f) == 
				DEVICE_DETECTED_PHY_COMM_OK) &&
			(!pdev->PhysicalDrive[tid].ucIsPowerConnected)) {
				AR_DEBUG(AR_HOT_PLUG, (KERN_DEBUG 
					"megaide: ich6r drv[%d] inserted\n", 
					tid));
				pdev->drive_inserted.u.Drives |= (1 << tid);
			}

			if (((uiSATAStatus & 0x0f) != 
				DEVICE_DETECTED_PHY_COMM_OK) &&
				(pdev->PhysicalDrive[tid].ucIsPowerConnected)) {
				AR_DEBUG(AR_HOT_PLUG, (KERN_DEBUG 
					"megaide: ich6r drv[%d] removed\n",
					tid));
				pdev->drive_removed.u.Drives |= (1 << tid);
			}
		}

		spin_unlock_irqrestore(adp->host_lock, flags);
	}
	
	/*
	 * If anyone is waiting for me to let them know I am on way out!
	 */
	pdev->ICHxR_thread_flag = 0;
	complete_and_exit(&((adapter_t *)pdev->adapter)->ICHxR_completion, 0);
}

/*
 *----------------------------------------------------------------------------
 * ICH7R related functions 
 *----------------------------------------------------------------------------
 */
/**
 * FoundICH7RInterrupt: Found ICH7R interrupt
 * This function checks whether any ICH7R channel raised any interrupt.
 * This check is done only if Driver expecting an interrupt from that
 * channel.
 *
 * Function parameters
 * @pdev	:	ICH7R device extension structure pointer
 * @p_int_return_val:	Interrupt detect status
 *		return 0x01 if interrupt detect in primary channel
 *		return 0x02 if interrupt detect in Secondary channel
 *		return 0x03 if interrupt detect in both channel
 * @return;
 *	TRUE if interrupt detected
 *	FALSE if no interrupt detected
 **/
BOOLEAN
FoundICH7RInterrupt(HW_DEVICE_EXTENSION *pdev, UCHAR *pChan)
{
	ULONG	baseBm =
		(ULONG)&(pdev->Channel[0].BaseBmAddress->Command.AsUchar);

	if(pdev->Channel[0].ExpectingInterrupt) {
		if (((ScsiPortReadPortUchar(baseBm + 0x02)) & 0x4) == 0x4) {
			*pChan |= 0X01;
		}
	}

	if(pdev->Channel[1].ExpectingInterrupt) {
		if (((ScsiPortReadPortUchar(baseBm + 0x0A)) & 0x4) == 0x4) {
			*pChan |= 0X02;
		}
	}

	if (*pChan) {
		return TRUE;
	} else {
		return FALSE;
	}
}


/**
 * EnableICH7RInterrupts: Enable ICH7R Interrupt
 * This function enables the ICH7R interrupts by programming pci config space.
 *
 * Function parameters
 * @pdev	:	ICH7R device extension structure pointer
 *
 * @return;
 *	TRUE if succeed
 *	FALSE if failed
 **/
BOOLEAN
EnableICH7RInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		reg_val;

	/* reading interrupt control register */
	pci_read_config_byte_Rlsi(pdev, 0x05, &reg_val);

	/* Enabling interrupt by writing 0 to 3rd bit */
	reg_val ^= 0x4;

	/* writing interrupt control register */
	pci_write_config_byte_Rlsi(pdev, 0x05, reg_val);

	ScsiPortStallExecution(1000);

	return AR_TRUE;
}

/**
 * DisableICH7RInterrupts: Disable ICH7R Interrupt
 * This function Disables the ICH7R interrupts by programming pci config space.
 *
 * Function parameters
 * @pdev	:	ICH7R device extension structure pointer
 *
 * @return;
 *	TRUE if succeed
 *	FALSE if failed
 **/
BOOLEAN
DisableICH7RInterrupts(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		reg_val;

	/* reading interrupt control register */
	pci_read_config_byte_Rlsi(pdev, 0x05, &reg_val);

	/* disabling interrupt by writing 1 to 3rd bit */
	reg_val |= 0x4;

	/* writing interrupt control register */
	pci_write_config_byte_Rlsi(pdev, 0x05, reg_val);

	ScsiPortStallExecution(1000);

	return AR_TRUE;
}


/**
 * IsICH7R80PinCable: Is This device using 80 pin cable?
 * This function will always return TRUE because this return value is used
 * for other ATA specific programming.
 *
 * Function parameters
 * @pdev	:	ICH7R device extension structure pointer
 * @ucChannelId	:	Channel ID 00h-Primary, 01h-secondary
 * @return;
 *	TRUE always
 **/
BOOLEAN
IsICH7R80PinCable(HW_DEVICE_EXTENSION *pdev, UCHAR ucChannelId)
{
	return AR_TRUE;
}

/**
 * ProgramDriveForICH7RChip: Program drives connected to ICH7R controller
 * This function supposed to program a drive connected to thecontroller.
 * It is used if bios is not programming the drives and in case of newly.
 *
 * Function parameters
 * @pdev	:	ICH7R device extension structure pointer
 * @tid		:	Target ID
 *		00h for primary master
 *		02h for secondary master
 * @ucMode	: Trasnsfer mode. it is defined
 *		MP_TRANSFER_MODE definition gives the values and meaning
 * @return;
 *	TRUE if succeed
 *	FALSE if failed
 **/
BOOLEAN
ProgramDriveForICH7RChip(HW_DEVICE_EXTENSION *pdev, UCHAR tid, UCHAR ucMode)
{
	AR_DEBUG(AR_DBG_ICH7R, 
		("ProgramDriveForICH7RChip not implemented\n"));
	return AR_TRUE;
}


/**
 * IsPostCommandToDriveSuccessICH7R: Post a Read Or Write Command to ICH7R Drive
 * This function is used to issue a read/write command to the drives
 * connected to ICH7R controller. This function is used to write/read
 * Errorlog, IRCD, Enquiry etc
 *
 * Function parameters
 * @pdev	:	ICH7R device extension structure pointer
 * @tid		:	Target ID 00h - primary master, 02h - secondary master
 * @uiStartBlock:	Start Logica Block address
 * @uiBlockCount:	Number of Blocks
 * @ucCmd	:	ATA read/write command READ_6/WRITE_6
 * @ulSglPhysicalAddress: Physical address of the SGList used for the command
 * @return;
 *	TRUE always
 **/
BOOLEAN
IsPostCommandToDriveSuccessICH7R(HW_DEVICE_EXTENSION *pdev, UINT tid,
	UINT uiStartBlock, UINT uiBlockCount,
	UCHAR ucCmd, ULONG ulSglPhysicalAddress)
{
	CHANNEL		*pChannel;
	IDE_REGISTERS_1	*base1;
	IDE_REGISTERS_2	*base2;
	BM_REGISTERS	*baseBm;
	UCHAR		ucStatus = 0;
	UCHAR		ucStatus2;

	pChannel= &(pdev->Channel[(tid >> 1)]);
	base1	= (IDE_REGISTERS_1 *)pChannel->BaseIoAddress1;
	base2	= (IDE_REGISTERS_2 *)pChannel->BaseIoAddress2;
	baseBm	= pChannel->BaseBmAddress;

	PostCommandToDrive(pdev, tid, uiStartBlock, uiBlockCount,
		ucCmd, ulSglPhysicalAddress);

	POLL_FOR_DMA_INTERRUPT(baseBm)

	ScsiPortWritePortUchar((ULONG)&(baseBm->Command.AsUchar),
		STOP_TRANSFER);

	GET_BASE_STATUS(base1, ucStatus);

	CLEAR_BM_INT(baseBm, ucStatus2);

	if (ucStatus & IDE_STATUS_BUSY) {
		printk_Rlsi(KERN_WARNING 
			"ICH7R intr busy bit set tid=%d\n", tid);
		return FALSE;
	}

	/* if no error continue;*/
	if (ucStatus & IDE_STATUS_ERROR) {
		UCHAR		err;
		printk_Rlsi(KERN_WARNING 
			"ICH7R intr err bit set tid=%d\n", tid);
		err = ScsiPortReadPortUchar((ULONG)((UCHAR *)base1 + 1));
		printk_Rlsi(KERN_WARNING "error byte = 0x%x\n", err);
		return FALSE;
	}

	return TRUE;
}


/**
 * mega_ICH7R_poll_thread: ICH7R polling thread
 * This is a kernel thread that detects drive insertion in ICHxR. Thread starts
 * as soon as the driver loads and it will stop only when driver unloads or
 * when system shut down
 *
 * Polling time interval of this thread is 2 Sec
 *
 * Thread will detect only the drive insertion on a channel where drive was
 * not connected previously and it sets the respective flags for the same.
 * MegaIDE driver's drive insertion removal module will process this event
 * later.
 *
 * Access of the generic data structure in this thread should be SMP proof to
 * make sure a smooth operation in an SMP system.
 *
 * Parameters
 * @pdev	: Driver data structure pointer
 * @return	: VOID
 **/

#define	DEVICE_DETECTED_PHY_COMM_OK		0x03
VOID 
mega_ICH7R_poll_thread(HW_DEVICE_EXTENSION *pdev)
{
	unsigned long		flags;
	struct Scsi_Host	*host;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	wait_queue_head_t	wait_queue_ICHxR;
#endif	
	adapter_t		*adp = (adapter_t *)pdev->adapter;
	/* ich6r drive detection */
	ULONG	ulOffsetArray[4] = 
		{0x128/*PM*/, 0x228/*PS*/,
		 0x1a8/*SM*/, 0x2a8/*SS*/};
	UCHAR	tid;
	UINT	uiSATAStatus; 	

	host = adp->pstScsiHost;

	/* Check if thread already active/running?*/
	if (pdev->ICHxR_thread_flag)
		return;


#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	lock_kernel();
#endif	
	daemonize("megaide");
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	unlock_kernel();
#endif	
#else
	/*
	 * We only listen to signals if the HA was loaded as a module.
	 * If the HA was compiled into the kernel, then we don't listen
	 * to any signals.
	 */
	if( host->loaded_as_module ) {
		siginitsetinv(&current->blocked, SHUTDOWN_SIGS);
	} else {
		siginitsetinv(&current->blocked, 0);
	}

	/* Flush resources */
	daemonize();
	reparent_to_init();

	/* Set the name of this process. */
	sprintf(current->comm, "megaide");
#endif

	pdev->ICHxR_thread_flag = 1;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	init_waitqueue_head(&wait_queue_ICHxR);
#endif

	while(1)
	{

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
		schedule();

		/* waiting for 2 sec, polling time is 2 Sec */
		interruptible_sleep_on_timeout(&wait_queue_ICHxR, (HZ*2));
#else		
		msleep(2000);
#endif 

		/* check if thread termination received? */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
		if (!pdev->ICHxR_thread_flag) {
			break;
		}
#else
		if( host->loaded_as_module ) {
			if(signal_pending(current))
				break;
		}
#endif
		 
		if (pdev->drive_inserted.u.Drives ||
			pdev->drive_removed.u.Drives) {
			/*if LD present, detect from here itself */
			if (!pdev->uiLogicalDriveCount) {	
				spin_lock_irqsave(adp->host_lock, flags);
				mega_check_drive_insertion_removal(pdev);
				spin_unlock_irqrestore(adp->host_lock, flags);
			}
			continue;
		}

		spin_lock_irqsave(adp->host_lock, flags);
		
		for(tid = 0; tid < 4; tid++)
		{
			uiSATAStatus = ScsiPortReadRegisterUlong(
				(ULONG)(pdev->virtBaseAddress5 + 
					ulOffsetArray[tid]));	

			if (((uiSATAStatus & 0x0f) == 
				DEVICE_DETECTED_PHY_COMM_OK) &&
			(!pdev->PhysicalDrive[tid].ucIsPowerConnected)) {
				AR_DEBUG(AR_HOT_PLUG, (KERN_DEBUG 
					"megaide: ich6r drv[%d] inserted\n", 
					tid));
				pdev->drive_inserted.u.Drives |= (1 << tid);
			}

			if (((uiSATAStatus & 0x0f) != 
				DEVICE_DETECTED_PHY_COMM_OK) &&
				(pdev->PhysicalDrive[tid].ucIsPowerConnected)) {
				AR_DEBUG(AR_HOT_PLUG, (KERN_DEBUG 
					"megaide: ich6r drv[%d] removed\n",
					tid));
				pdev->drive_removed.u.Drives |= (1 << tid);
			}
		}

		spin_unlock_irqrestore(adp->host_lock, flags);
	}
	
	/*
	 * If anyone is waiting for me to let them know I am on way out!
	 */
	pdev->ICHxR_thread_flag = 0;
	complete_and_exit(&((adapter_t *)pdev->adapter)->ICHxR_completion, 0);
}

/* Initialize controller type and controller speed */
BOOLEAN
mide_init_adp_info_and_pointers(HW_DEVICE_EXTENSION *pdev, USHORT tid)
{
	switch (tid) {

	case PCI_DEVICE_ID_CMD_649:
		pdev->ControllerSpeed	= Udma100;
		pdev->ucCardType	= CMD_649;
		pdev->find_intr		= Found64xInterrupt;
		pdev->disable_intr	= Disable64xInterrupts;
		pdev->enable_intr	= Enable64xInterrupts;
		pdev->is_80pin_cable	= Is64x80PinCable;
		pdev->prgrm_chip_or_drvs= ProgramDriveFor64xChip;
		pdev->post_cmd_to_drive	= IsPostCommandToDriveSuccess64x;
		break;

	case PCI_DEVICE_ID_CMD_680:
	case PCI_DEVICE_ID_DELL_680:
		pdev->ControllerSpeed	= Udma133;
		pdev->ucCardType	= CMD_680;
		pdev->find_intr		= Found68xInterrupt;
		pdev->disable_intr	= Disable68xInterrupts;
		pdev->enable_intr	= Enable68xInterrupts;
		pdev->is_80pin_cable	= Is68x80PinCable;
		pdev->prgrm_chip_or_drvs= ProgramDriveFor68xChip;
		pdev->post_cmd_to_drive	= IsPostCommandToDriveSuccess68x;
		break;
	
	/* some functions are same as cmd 680 controller */
	case PCI_DEVICE_ID_CMD_3112:
		pdev->ControllerSpeed	= Udma133;
		pdev->ucCardType	= SIL_3112;
		pdev->find_intr		= Found68xInterrupt;
		pdev->disable_intr	= DisableAthenaInterrupts;
		pdev->enable_intr	= EnableAthenaInterrupts;
		pdev->is_80pin_cable	= Is68x80PinCable;
		pdev->prgrm_chip_or_drvs= ProgramDriveFor68xChip;
		pdev->post_cmd_to_drive	= IsPostCommandToDriveSuccess68x;
		break;

	case PCI_DEVICE_ID_SVWS_CSB6IDE:
		pdev->ControllerSpeed	= Udma100;
		pdev->ucCardType	= RCC_SERVER_WORKS_CSB6;
		pdev->find_intr		= FoundCSB6Interrupt;
		pdev->disable_intr	= DisableCSB6Interrupts;
		pdev->enable_intr	= EnableCSB6Interrupts;
		pdev->is_80pin_cable	= IsCSB680PinCable;
		pdev->prgrm_chip_or_drvs= ProgramDriveForCSB6Chip;
		pdev->post_cmd_to_drive	= IsPostCommandToDriveSuccessCSB6;
		
		/* enable turbo rebuild if IBM CSB6 system */
		if (((pdev->usSubsysVendorId == PCI_VENDOR_ID_IBM) &&
			(pdev->usSubsysDeviceId == PCI_DEVICE_ID_IBM_CSB6)) ||
			((pdev->usSubsysVendorId == PCI_VENDOR_ID_INTEL) &&
			(pdev->usSubsysDeviceId==PCI_DEVICE_ID_SVWS_CSB6IDE))) {
			pdev->turbo_rbld_flag = 1; 
		}
		break;

	case PCI_DEVICE_ID_INTEL_ICH5R:
	case PCI_DEVICE_ID_INTEL_HR:
		pdev->ControllerSpeed	= Udma100;
		pdev->ucCardType	= INTEL_ICH5R;
		pdev->find_intr		= FoundICH5RInterrupt;
		pdev->disable_intr	= DisableICH5RInterrupts;
		pdev->enable_intr	= EnableICH5RInterrupts;
		pdev->is_80pin_cable	= IsICH5R80PinCable;
		pdev->prgrm_chip_or_drvs= ProgramDriveForICH5RChip;
		pdev->post_cmd_to_drive	= IsPostCommandToDriveSuccessICH5R;
		break;

	case PCI_DEVICE_ID_INTEL_ICH6R:
		pdev->ControllerSpeed	= Udma133;
		pdev->ucCardType	= INTEL_ICH6R;
		pdev->find_intr		= FoundICH6RInterrupt;
		pdev->disable_intr	= DisableICH6RInterrupts;
		pdev->enable_intr	= EnableICH6RInterrupts;
		pdev->is_80pin_cable	= IsICH6R80PinCable;
		pdev->prgrm_chip_or_drvs= ProgramDriveForICH6RChip;
		pdev->post_cmd_to_drive	= IsPostCommandToDriveSuccessICH6R;
#ifdef _SOFT_BBM_
		pdev->bIsSBBMSupported	= 1;
#endif /* _SOFT_BBM_ */
		break;
		
	case PCI_DEVICE_ID_INTEL_ICH7R:
		pdev->ControllerSpeed	= Udma133;
		pdev->ucCardType	= INTEL_ICH7R;
		pdev->find_intr		= FoundICH7RInterrupt;
		pdev->disable_intr	= DisableICH7RInterrupts;
		pdev->enable_intr	= EnableICH7RInterrupts;
		pdev->is_80pin_cable	= IsICH7R80PinCable;
		pdev->prgrm_chip_or_drvs= ProgramDriveForICH7RChip;
		pdev->post_cmd_to_drive	= IsPostCommandToDriveSuccessICH7R;
#ifdef _SOFT_BBM_
		pdev->bIsSBBMSupported	= 1;
#endif /* _SOFT_BBM_ */
		break;
		
	default:
		printk_Rlsi("Unknown device tid=%x !!!\n", tid);
		return (AR_FALSE);
	}
	
	return (AR_TRUE);
}

/* vim: set ts=8 sw=8 tw=78 ai si: */

