/*
 * Linux MegaRAID IDEal Software RAID Stack
 * Copyright (c) 2001 LSI Logic Corporation, All Rights Reserved.
 *-----------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *-----------------------------------------------------------------------------
 * File Name	: megaide_char.c
 */

static INT giMajorNumber;

static VOID
mide_char_ioctl_over(Scsi_Cmnd * scmd)
{
	HW_DEVICE_EXTENSION	*pdev = (HW_DEVICE_EXTENSION *)SCP2HOST(scmd);
	adapter_t		*adapter = pdev->adapter;
	
	AR_DEBUG(AR_DBG_NIT,("megaide: mide_char_ioctl_over\n"));

	wake_up_interruptible(&adapter->wait_queue);
	scmd->state = MIDE_CHAR_IOCTL_FINISHED;
}

UCHAR
mide_abort_passthu_IOCTL_command(HW_DEVICE_EXTENSION *pdev, Scsi_Cmnd *scmd)
{
	FAST_IDE_SCB_EXTENSION	*scb = NULL;

	AR_DEBUG(AR_DBG_NIT,("megaide: mide_abort_passthu_IOCTL_command\n"));

	scb	= (FAST_IDE_SCB_EXTENSION *) scmd->host_scribble;
	megaide_abort_ioctl (pdev, scb);
	return 0;
}


INT
mide_handle_ioctl(UINT cmd, ULONG arg,
	UINT inlen, UINT outlen, INT adp_no)
{
	INT			i;
	INT			iRetVal = -EFAULT;
	INT			data_size = 0;
	Scsi_Cmnd		*scmd;
	unsigned long		flags=0;
	pchar_megaide_ioctl	p_ioctl;
	HW_DEVICE_EXTENSION	*pdev = NULL;
	FAST_IDE_SCB_EXTENSION	*scb;
	adapter_t		*adapter;

	if (cmd != MEGAIDE_DEV_IOCTL) {
		return -EINVAL;
	}

	data_size	= (inlen > outlen) ? inlen : outlen;
	adapter		= gastCardInfo[adp_no-1];
	pdev		= &adapter->DeviceExtension;

	if (pdev == NULL) {
		printk_Rlsi("megaide: dev ptr is null. adp_no = %d\n", adp_no);
		return -EINVAL;
	}

	/* grab memory for ioctl, it is allocated during initialization */
	scmd = (Scsi_Cmnd *) adapter->nit_scsi_cmd;
	if (scmd == NULL) {
		printk_Rlsi("megaide: driver dont have ioctl memory!\n");
		return -ENOMEM;
	}

	memset_Rlsi(scmd, 0, sizeof(Scsi_Cmnd));

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	scmd->device = adapter->nit_scsi_device;
	if (scmd->device == NULL) {
		printk_Rlsi("megaide: no proper memory for ioctl!\n");
		return -ENOMEM;
	}

	memset_Rlsi(scmd->device, 0, sizeof(struct scsi_device));
#endif

	scmd->host_scribble = adapter->host_scribble;
	scb = (FAST_IDE_SCB_EXTENSION *) scmd->host_scribble;
	if (scb == NULL) {
		printk_Rlsi("megaide: driver dont have ioctl memory 2!\n");
		return -ENOMEM;
	}

	memset_Rlsi(scb, 0, get_scb_extension_size());

	p_ioctl = (pchar_megaide_ioctl) adapter->ioctl_data;
	if (p_ioctl == NULL) {
		printk_Rlsi("megaide: driver dont have ioctl memory 3!\n");
		return -ENOMEM;
	}
	memset_Rlsi(p_ioctl, 0, data_size);

	if (copy_from_user(p_ioctl,(CHAR *)arg, (inlen + CHAR_HEADER_SIZE))) {
		printk_Rlsi("megaide: copy ioctl data from user failed!\n");
		return -EFAULT;
	}
	
	scmd->request_buffer	= &p_ioctl->wdata;
	scb->request_buffer	= (VOID *)scmd->request_buffer;
	scb->pstScsiCmd		= (VOID *)scmd;
	scmd->scsi_done		= mide_char_ioctl_over;
	SCP2HOST(scmd)		= (struct Scsi_Host *) pdev;
	scmd->state		= MIDE_CHAR_IOCTL_CALLED;
	
	init_waitqueue_head(&adapter->wait_queue);

	spin_lock_irqsave(adapter->host_lock, flags);
	
	switch (HandleIOCTLCall(pdev, scb)) {
	
	case IOCTL_OK:
		
		AR_DEBUG(AR_DBG_NIT,("megaide: ioctl completed success\n"));
		
		megaide_complete_command(pdev, scb,(DID_OK << 16));
		iRetVal = 0;
		break;

	case IOCTL_ERROR:
		
		AR_DEBUG(AR_DBG_NIT,("megaide: ioctl error happened\n"));
		
		megaide_complete_command(pdev, scb,(DID_BAD_TARGET << 16));
		break;
	
	case IOCTL_QUEUE:
		
		AR_DEBUG(AR_DBG_NIT,("megaide: ioctl queued and waiting\n"));
		
		spin_unlock_irqrestore(adapter->host_lock, flags);
		
		flags = 0;
		
		for (i=0; i < (MIDE_CHAR_IOCTL_TIMEOUT * HZ); i++) {

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
			interruptible_sleep_on_timeout(&adapter->wait_queue, 1);
#else		
			msleep(1000);
#endif		

			if (scmd->state == MIDE_CHAR_IOCTL_FINISHED) {
				break;
			}
		}

		spin_lock_irqsave(adapter->host_lock, flags);

		if (scmd->state != MIDE_CHAR_IOCTL_FINISHED) {
			AR_DEBUG(AR_DBG_NIT,("megaide:ioctl is aborting\n"));
			mide_abort_passthu_IOCTL_command(pdev, scmd);
		} else {
			iRetVal = 0;
		}
		break;

	default:
		megaide_complete_command(pdev, scb,(DID_BAD_TARGET << 16));
		break;
	}

	spin_unlock_irqrestore(adapter->host_lock, flags);
	
	if (!iRetVal) {
		copy_to_user((CHAR *)arg, p_ioctl, (outlen+CHAR_HEADER_SIZE));
	}
	
	return iRetVal;
}

static INT
megaidedev_ioctl(struct inode *inode, struct file *filep,
		UINT cmd, unsigned long arg)
{
	INT			chk = 0;
	INT			iRetVal = -EFAULT;
	char_megaide_ioctl	ioctl;
	HW_DEVICE_EXTENSION	*pdev = NULL;
	adapter_t		*adapter;
	
	chk = access_ok(VERIFY_READ, arg, sizeof(char_megaide_ioctl));
	if (chk != 1) {
 		printk_Rlsi("megaide: verify ioctl structure(read) failed\n");
		return -EFAULT;
	}

	if (copy_from_user(&ioctl,(CHAR *) arg, sizeof(char_megaide_ioctl))) {
		printk_Rlsi("megaide: copy ioctl data failed\n");
		return -EFAULT;
	}

	chk = access_ok(VERIFY_WRITE, arg, (CHAR_HEADER_SIZE + ioctl.outlen));
	if (chk != 1) {
 		printk_Rlsi("megaide: verify ioctl memory(write) failed\n");
		return -EFAULT;
	}

	chk = access_ok(VERIFY_READ, arg,(CHAR_HEADER_SIZE + ioctl.inlen));
	if (chk != 1) {
 		printk_Rlsi("megaide: verify ioctl memory(read) failed\n");
		return -EFAULT;
	}

	if (memcmp_Rlsi(ioctl.acSignature, "MEGANIT", 7) != 0) {
 		printk_Rlsi("megaide: ioctl with wrong signature\n");
		return -EINVAL;
	}

	if ((ioctl.iAdapNumber > giNumberOfCtrlrs) || (!ioctl.iAdapNumber)) {
 		printk_Rlsi("megaide: ioctl with wrong adapter number\n");
		return -EINVAL;
	}

	/* one ioctl only allowed at a time */
	adapter	= gastCardInfo[ioctl.iAdapNumber-1];
	pdev	= &adapter->DeviceExtension;

	down(&adapter->nit_lock);

	iRetVal = mide_handle_ioctl(cmd, arg, ioctl.inlen,
		ioctl.outlen, ioctl.iAdapNumber);

	up(&adapter->nit_lock);

	return iRetVal;
}

static INT
megaidedev_open(struct inode *inode, struct file *filep)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
#else
	if( !capable(CAP_SYS_ADMIN) ) return -EACCES;
#endif
	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static INT
megaidedev_close(struct inode *inode, struct file *filep)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static struct file_operations megaidedev_fops = {
	ioctl	: megaidedev_ioctl,
	open	: megaidedev_open,
	release	: megaidedev_close,
};
#else
static struct file_operations megaidedev_fops = {
	.owner	= THIS_MODULE,
	.ioctl	= megaidedev_ioctl,
	.open	= megaidedev_open,
};
#endif

VOID
megaidedev_attach(VOID)
{
	giMajorNumber = register_chrdev(0, "megaidedev", &megaidedev_fops);
#ifdef CONFIG_COMPAT 
	/* Register the 32-bit ioctl conversion	 */
	register_ioctl32_conversion(MEGAIDE_DEV_IOCTL, NULL);
#endif
}

VOID
megaidedev_detach(VOID)
{
	unregister_chrdev(giMajorNumber, "megaidedev");
#ifdef CONFIG_COMPAT 
	unregister_ioctl32_conversion(MEGAIDE_DEV_IOCTL);
#endif
}

/* vim: set ts=8 sw=8 tw=78 ai si: */
