/*
 * Linux MegaRAID IDEal Software RAID Stack
 * Copyright (c) 2001 LSI Logic Corporation, All Rights Reserved.
 *-----------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *-----------------------------------------------------------------------------
 * File Name	: megaide_kernel_lib.c
 */

/* library calls */
VOID
ScsiPortStallExecution(UINT usecs)
{
	while (usecs) {
		udelay(1);
		usecs--;
		cpu_relax();
	}
	return;
}

VOID
ScsiPortReadPortBufferUshort(ULONG port, VOID *addr, UINT count)
{
	insw (port,addr,count);
	return;
}

VOID
ScsiPortWritePortBufferUshort(ULONG port, const VOID *addr, UINT count)
{
	outsw (port,addr,count);
	return;
}

UCHAR
ScsiPortReadPortUchar(ULONG port)
{
	UCHAR			return_val;
#ifdef DEBUG_IO_ACCESS
	HW_DEVICE_EXTENSION	*pdev;
	adapter_t		*adapter;
#endif
	return_val = inb(port);
#ifdef DEBUG_IO_ACCESS

	adapter	= (adapter_t *)gastCardInfo[0];
	pdev	= &adapter->DeviceExtension;

	if(pdev->debug_io_head >= (MAX_IO_ACCESS_LOG_ENTRIES-1))
		pdev->debug_io_head=0;
	else
		pdev->debug_io_head++;
		
	pdev->debug_io[pdev->debug_io_head].io_access=IO_ACCESS_READ_IO;
	pdev->debug_io[pdev->debug_io_head].io_register = port;
	pdev->debug_io[pdev->debug_io_head].io_value = return_val;
#endif
	return(return_val);
}

USHORT
ScsiPortReadPortUshort(ULONG port)
{
	USHORT			return_val;
#ifdef DEBUG_IO_ACCESS
	HW_DEVICE_EXTENSION	*pdev;
	adapter_t		*adapter;
#endif

	return_val = inw(port);
#ifdef DEBUG_IO_ACCESS
	adapter	= (adapter_t *)gastCardInfo[0];
	pdev	= &adapter->DeviceExtension;
	
	if(pdev->debug_io_head >= (MAX_IO_ACCESS_LOG_ENTRIES-1))
		pdev->debug_io_head=0;
	else
		pdev->debug_io_head++;
	
	pdev->debug_io[pdev->debug_io_head].io_access=IO_ACCESS_READ_IO;
	pdev->debug_io[pdev->debug_io_head].io_register = port;
	pdev->debug_io[pdev->debug_io_head].io_value = return_val;
#endif
	return(return_val);
}

UINT
ScsiPortReadPortUlong(ULONG port)
{
	UINT			return_val;
#ifdef DEBUG_IO_ACCESS
	HW_DEVICE_EXTENSION	*pdev;
	adapter_t		*adapter;
#endif
	return_val = inl(port);
	
#ifdef DEBUG_IO_ACCESS
	adapter	= (adapter_t *)gastCardInfo[0];
	pdev	= &adapter->DeviceExtension;
	
	if(pdev->debug_io_head >= (MAX_IO_ACCESS_LOG_ENTRIES-1))
		pdev->debug_io_head=0;
	else
		pdev->debug_io_head++;

	pdev->debug_io[pdev->debug_io_head].io_access=IO_ACCESS_READ_IO;
	pdev->debug_io[pdev->debug_io_head].io_register = port;
	pdev->debug_io[pdev->debug_io_head].io_value = return_val;
#endif
	return(return_val);
}

VOID
ScsiPortWritePortUchar(ULONG port, UCHAR value)
{
#ifdef DEBUG_IO_ACCESS
	adapter_t		*adapter;
	HW_DEVICE_EXTENSION	*pdev;

	adapter	= (adapter_t *)gastCardInfo[0];
	pdev	= &adapter->DeviceExtension;

	if(pdev->debug_io_head >= (MAX_IO_ACCESS_LOG_ENTRIES-1))
		pdev->debug_io_head=0;
	else
		pdev->debug_io_head++;

	pdev->debug_io[pdev->debug_io_head].io_access=IO_ACCESS_WRITE_IO;
	pdev->debug_io[pdev->debug_io_head].io_register = port;
	pdev->debug_io[pdev->debug_io_head].io_value = value;
#endif
	outb(value,port);
	return;
}


VOID
ScsiPortWritePortUlong(ULONG port, UINT value)
{
#ifdef DEBUG_IO_ACCESS
	adapter_t		*adapter;
	HW_DEVICE_EXTENSION	*pdev;

	adapter	= (adapter_t *)gastCardInfo[0];
	pdev	= &adapter->DeviceExtension;

	if(pdev->debug_io_head >= (MAX_IO_ACCESS_LOG_ENTRIES-1))
		pdev->debug_io_head=0;
	else
		pdev->debug_io_head++;

	pdev->debug_io[pdev->debug_io_head].io_access=IO_ACCESS_WRITE_IO;
	pdev->debug_io[pdev->debug_io_head].io_register = port;
	pdev->debug_io[pdev->debug_io_head].io_value = value;
#endif
	outl(value,port);
	return;
}

UINT
ScsiPortReadRegisterUlong(ULONG addr)
{
	return readl ((UCHAR *)addr);
}

VOID
ScsiPortWriteRegisterUlong(ULONG addr, UINT value)
{
	writel(value, (UCHAR *)addr);
}

UINT
ScsiPortReadRegisterUchar(ULONG addr)
{
	return readb ((UCHAR *)addr);
}

VOID
ScsiPortWriteRegisterUchar(ULONG addr, UINT value)
{
	writeb(value, (UCHAR *)addr);
}

INT
memcmp_Rlsi(const VOID *cs,const VOID *ct,size_t count)
{
	return memcmp(cs, ct, count); 
}

VOID *
memcpy_Rlsi(VOID *dest,const VOID *src,size_t count)
{
	return memcpy(dest,src,count); 
}

VOID *
memmove_Rlsi(VOID *dest,const VOID *src,size_t count)
{
	return memmove(dest,src,count); 
}

VOID *
memset_Rlsi(VOID *s,INT c,size_t count)
{
	return memset(s,c,count); 
}

CHAR *
strncpy_Rlsi(CHAR * dest,const CHAR *src,size_t count)
{
	return strncpy(dest,src,count); 
}

INT
strncmp_Rlsi(const CHAR * cs,const CHAR * ct,size_t count)
{
	return strncmp(cs,ct,count); 
}

INT
pci_read_config_word_Rlsi(HW_DEVICE_EXTENSION *pdev, INT where, USHORT *val)
{
	INT		return_val;
	adapter_t	*adapter = (adapter_t *)pdev->adapter;
	
	return_val	= pci_read_config_word(adapter->pstDev,where,val);
	
#ifdef DEBUG_IO_ACCESS
	if(pdev->debug_io_head >= (MAX_IO_ACCESS_LOG_ENTRIES-1))
		pdev->debug_io_head=0;
	else
		pdev->debug_io_head++;
	
	pdev->debug_io[pdev->debug_io_head].io_access=IO_ACCESS_READ_PCI;
	pdev->debug_io[pdev->debug_io_head].io_register = where;
	pdev->debug_io[pdev->debug_io_head].io_value = *val;
#endif
	return(return_val);
}

INT
pci_read_config_byte_Rlsi(HW_DEVICE_EXTENSION *pdev, INT where, UCHAR *val)
{
	INT		return_val;
	adapter_t	*adapter = (adapter_t *)pdev->adapter;
	
	return_val = pci_read_config_byte(adapter->pstDev,where,val);

#ifdef DEBUG_IO_ACCESS
	if(pdev->debug_io_head == (MAX_IO_ACCESS_LOG_ENTRIES-1))
		pdev->debug_io_head=0;
	else
		pdev->debug_io_head++;
	
	pdev->debug_io[pdev->debug_io_head].io_access=IO_ACCESS_READ_PCI;
	pdev->debug_io[pdev->debug_io_head].io_register = where;
	pdev->debug_io[pdev->debug_io_head].io_value = *val;
#endif
	return(return_val);
}

INT
pci_write_config_word_Rlsi(HW_DEVICE_EXTENSION *pdev,
	INT where, USHORT val)
{
	adapter_t	*adapter = (adapter_t *)pdev->adapter;

#ifdef DEBUG_IO_ACCESS
	if(pdev->debug_io_head == (MAX_IO_ACCESS_LOG_ENTRIES-1))
		pdev->debug_io_head=0;
	else
		pdev->debug_io_head++;
		
	pdev->debug_io[pdev->debug_io_head].io_access=IO_ACCESS_WRITE_PCI;
	pdev->debug_io[pdev->debug_io_head].io_register = where;
	pdev->debug_io[pdev->debug_io_head].io_value = val;
#endif

	return pci_write_config_word(adapter->pstDev,where,val);
}

INT
pci_write_config_byte_Rlsi(HW_DEVICE_EXTENSION *pdev,
	INT where, UCHAR val)
{
	adapter_t	*adapter = (adapter_t *)pdev->adapter;

#ifdef DEBUG_IO_ACCESS
	if(pdev->debug_io_head == (MAX_IO_ACCESS_LOG_ENTRIES-1))
		pdev->debug_io_head=0;
	else
		pdev->debug_io_head++;
	
	pdev->debug_io[pdev->debug_io_head].io_access=IO_ACCESS_WRITE_PCI;
	pdev->debug_io[pdev->debug_io_head].io_register = where;
	pdev->debug_io[pdev->debug_io_head].io_value = val;
#endif
	return pci_write_config_byte(adapter->pstDev,where,val);
}

VOID
wake_up_Rlsi(HW_DEVICE_EXTENSION *pdev)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	adapter_t	*adapter = (adapter_t *)pdev->adapter;
	
	wake_up(&adapter->internal_wait);
#endif	
	/* else internal_wait is not used */
}

VOID
sleep_on_timeout_Rlsi(HW_DEVICE_EXTENSION *pdev)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	adapter_t	*adapter = (adapter_t *)pdev->adapter;

	sleep_on_timeout(&adapter->internal_wait, 5 * HZ);
#else
	/* 1 sec sleep is enough to abort a command safely */
	msleep(1000);
#endif	
}

asmlinkage INT
printk_Rlsi(const CHAR *fmt, ...)
{
	va_list		ap;
	CHAR		buf[256];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	return printk(buf);
}

ULONG
pci_map_single_Rlsi(HW_DEVICE_EXTENSION *pdev, VOID *buff , UINT len)
{
	return((ULONG) pci_map_single(((adapter_t *)pdev->adapter)->pstDev,
		(VOID *) buff, len, PCI_DMA_BIDIRECTIONAL));
}

VOID
pci_unmap_single_Rlsi(HW_DEVICE_EXTENSION *pdev, VOID *buff , UINT len)
{
	pci_unmap_single(((adapter_t *)pdev->adapter)->pstDev,
		(dma_addr_t) (ULONG) buff, len, PCI_DMA_BIDIRECTIONAL);
}

USHORT
pci_map_sg_Rlsi(HW_DEVICE_EXTENSION *pdev, FAST_IDE_SCB_EXTENSION *scb)
{
	return(pci_map_sg(((adapter_t *)pdev->adapter)->pstDev,
		(struct scatterlist *) scb->request_buffer,
		scb->sglem, PCI_DMA_BIDIRECTIONAL));
}

VOID
pci_unmap_sg_Rlsi(HW_DEVICE_EXTENSION *pdev, FAST_IDE_SCB_EXTENSION *scb)
{
	pci_unmap_sg(((adapter_t *)pdev->adapter)->pstDev,
		(struct scatterlist *) scb->request_buffer,
		scb->sglem, PCI_DMA_BIDIRECTIONAL);
}

UINT
CURRENT_TIME_RLSI(VOID)
{
	return (MRAID_CURRENT_TIME_IN_SEC);
}

UINT get_jiffies_count(VOID)
{
	return jiffies;
}

UINT mega_get_hz_value(VOID)
{
	return HZ;
}

VOID *
request_region_Rlsi(USHORT port, UINT count)
{
	return ((VOID *)request_region(port, count, "megaide"));
}


VOID
release_region_Rlsi(USHORT port, UINT count)
{
	release_region (port , count);
}

VOID * 
kmalloc_Rlsi(INT size)
{
	return (kmalloc(size, GFP_KERNEL));
}

VOID 
kfree_Rlsi(VOID *objp)
{
	kfree(objp);
}
/* vim: set ts=8 sw=8 tw=78 ai si: */

