/*
 * Linux MegaRAID IDEal Software RAID Stack
 * Copyright (c) 2001 LSI Logic Corporation, All Rights Reserved.
 *-----------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *-----------------------------------------------------------------------------
 * File Name	: megaide_pci.c
 */

/**
 * Description	: 
 * This function will remove all allocated memory using kmalloc/pci_malloc.
 * Before removing memory all the memory pointers are checked for validity.
 * This is done to make use of this function from any part of the driver.
 * Command pool memory and scsi host memory are not removed in this function
 *
 * Parameters
 * @param pdev		: Device extention pointer
 * @param adapter	: Adapter structure pointer
 *
 * Return		: None
 **/
VOID
mide_remove_device_ext_mem(HW_DEVICE_EXTENSION *pdev, adapter_t *adapter)
{
	UINT		i;
	
	if (pdev->GateKeeper) {
		kfree(pdev->GateKeeper);
	}
	
	if (pdev->GateScbExtension) {
		kfree (pdev->GateScbExtension);
	}
	
	if (pdev->real_gk_buf) {
		pci_free_consistent(adapter->pstDev, 
			adapter->gate_keeper_buff_length,
			pdev->real_gk_buf, adapter->h_gate_keeper_buff);
	}
	
	for (i = 0; i < MAX_CHANNELS_PER_CONTROLLER; i++)
	{
		if (pdev->Channel[i].pSglEntry) {
			pci_free_consistent(adapter->pstDev,
				(MAX_SGL_ENTRIES_PER_PHYSICAL_DRIVE *
					sizeof(SGL_ENTRY)),
				pdev->Channel[i].pSglEntry,
				adapter->Channel[i].h_sgl_entry);
		}
		
		if (pdev->Channel[i].DiskConfiguration) {
			pci_free_consistent(adapter->pstDev, IDE_SECTOR_SIZE,
				pdev->Channel[i].DiskConfiguration,
				adapter->Channel[i].h_disc_conf);
		}
		
		if (pdev->Channel[i].DiskEnquiry) {
			pci_free_consistent(adapter->pstDev, IDE_SECTOR_SIZE,
				pdev->Channel[i].DiskEnquiry,
				adapter->Channel[i].h_enquiry_conf);
		}
	}

	for (i = 0; i < MAX_LOGICAL_DRIVES; i++)
	{
		if (pdev->LogicalDrive[i].InternalScbExtension) {
			kfree(pdev->LogicalDrive[i].InternalScbExtension);
		}
		
		if (pdev->LogicalDrive[i].GateKeeper) {
			kfree(pdev->LogicalDrive[i].GateKeeper);
		}
	}

	for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++)
	{
		if (pdev->PhysicalDrive[i].ErrorLogSector) {
			pci_free_consistent(adapter->pstDev, IDE_SECTOR_SIZE,
				pdev->PhysicalDrive[i].ErrorLogSector,
				adapter->PhysicalDrive[i].h_err_log_sec);
		}
		
#ifdef _SOFT_BBM_
		if (pdev->PhysicalDrive[i].SBBMChanger.SbbmScb)
			kfree(pdev->PhysicalDrive[i].SBBMChanger.SbbmScb);
			
		if (pdev->PhysicalDrive[i].SBBMChanger.pucSBBMData) {
			pci_free_consistent(adapter->pstDev, 
				(SBBM_TABLE_SIZE_IN_BYTES + REGION_HW_BOUNDARY),
				pdev->PhysicalDrive[i].SBBMChanger.pucSBBMData,
				adapter->PhysicalDrive[i].h_sbbm_sec);
		}
#endif /*_SOFT_BBM_*/
	}
	
	if (adapter->nit_scsi_cmd) {
		pci_free_consistent(adapter->pstDev, sizeof(Scsi_Cmnd),
			adapter->nit_scsi_cmd, adapter->h_nit_scsi_cmd);
	}
	
	if (adapter->host_scribble) {
		pci_free_consistent(adapter->pstDev, get_scb_extension_size(),
			adapter->host_scribble, adapter->h_host_scribble);
	}
	
	if (adapter->ioctl_data) {
		pci_free_consistent(adapter->pstDev,
			(NIT_MAX_DATA_TRANSFER + CHAR_HEADER_SIZE),
			adapter->ioctl_data, adapter->h_ioctl_data);
	}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	if (adapter->nit_scsi_device) {
		pci_free_consistent(adapter->pstDev, 
			sizeof(struct scsi_device),
			adapter->nit_scsi_device, adapter->h_nit_scsi_device);
	}
#endif	
}

/**
 * This function will allocate device memories required in device extension
 * memory allocation failure is handled properly
 */
BOOLEAN
mide_allocate_device_ext_mem(HW_DEVICE_EXTENSION *pdev,
		struct pci_dev *dev, USHORT adpno)
{
	UINT		length = (MAX_SECTORS_PER_IDE_TRANSFER*IDE_SECTOR_SIZE*
			MAX_CHANNELS_PER_CONTROLLER);
	UINT		reduceToAlign;
	UINT		i;
	adapter_t	*adapter = (adapter_t *)pdev->adapter;

	/* allocating memory GateKeeper, GateScbExtension */
	pdev->GateKeeper = kmalloc(get_gatekeeper_size(), GFP_KERNEL);
	pdev->GateScbExtension = (FAST_IDE_SCB_EXTENSION *)
		kmalloc(get_scb_extension_size(), GFP_KERNEL);
	
	length = (length > KMALLOC_MAXSIZE) ? KMALLOC_MAXSIZE : length;
	adapter->gate_keeper_buff_length = length;
	pdev->real_gk_buf = pci_alloc_consistent(dev, length, 
		&adapter->h_gate_keeper_buff);
	
	if (!pdev->GateKeeper || !pdev->GateScbExtension ||
		!pdev->real_gk_buf) {
		printk_Rlsi(KERN_WARNING 
			"megaide: gate keeper memory allocation failed\n");
		goto go_err_free_mem;
	}
	
	memset_Rlsi(pdev->real_gk_buf, 0, length);
	
	pdev->GateKeeperBuffer		= (UCHAR *)pdev->real_gk_buf;
	pdev->GateKeeperPhysicalBuffer	=
		(ULONG)adapter->h_gate_keeper_buff;
	pdev->GateKeeperBufferLength	= length;
	reduceToAlign	= (REGION_HW_BOUNDARY -
		(pdev->GateKeeperPhysicalBuffer % REGION_HW_BOUNDARY));
	pdev->GateKeeperBuffer		+= reduceToAlign;
	pdev->GateKeeperPhysicalBuffer	+= reduceToAlign;
	pdev->GateKeeperBufferLength	-= reduceToAlign;

	if (pdev->GateKeeperBufferLength <= 2*IDE_SECTOR_SIZE) {
		printk_Rlsi(KERN_WARNING 
			"megaide: gate mem is not enough!\n");
		goto go_err_free_mem;
	}

	/* allocating memory for channel */
	for (i = 0; i < MAX_CHANNELS_PER_CONTROLLER; i++)
	{

		pdev->Channel[i].DiskConfiguration =
			pci_alloc_consistent(dev, IDE_SECTOR_SIZE,
				&adapter->Channel[i].h_disc_conf);

		pdev->Channel[i].DiskEnquiry =
			pci_alloc_consistent(dev, IDE_SECTOR_SIZE,
				&adapter->Channel[i].h_enquiry_conf);

		if (!pdev->Channel[i].DiskEnquiry ||
			!pdev->Channel[i].DiskConfiguration) {
			AR_DEBUG(AR_DEBUG_WARNING, (KERN_DEBUG 
				"megaide: channel mem allocation failed!\n"));
			goto go_err_free_mem;
		}

		memset_Rlsi(pdev->Channel[i].DiskConfiguration, 0,
			IDE_SECTOR_SIZE);
		
		memset_Rlsi(pdev->Channel[i].DiskEnquiry, 0,
			IDE_SECTOR_SIZE);

		pdev->Channel[i].DiskConfigurationSgl.Physical.Address =
			(UINT)adapter->Channel[i].h_disc_conf;
		pdev->Channel[i].DiskConfigurationSgl.Physical.Length =
			IDE_SECTOR_SIZE;
		pdev->Channel[i].DiskConfigurationSgl.Physical.EndOfListFlag
			 = 1;
		pdev->Channel[i].DiskEnquirySgl.Physical.Address =
			(UINT)adapter->Channel[i].h_enquiry_conf;
		pdev->Channel[i].DiskEnquirySgl.Physical.Length = 
			IDE_SECTOR_SIZE;
		pdev->Channel[i].DiskEnquirySgl.Physical.EndOfListFlag = 1;
	}

	/* allocating memory for logical drives */
	for (i = 0; i<MAX_LOGICAL_DRIVES; i++)
	{
		pdev->LogicalDrive[i].InternalScbExtension =
			kmalloc(get_scb_extension_size(), GFP_KERNEL);


		pdev->LogicalDrive[i].GateKeeper = (FAST_IDE_SCB_EXTENSION *)
			kmalloc(get_gatekeeper_size(), GFP_KERNEL);

		if (!pdev->LogicalDrive[i].InternalScbExtension ||
			!pdev->LogicalDrive[i].GateKeeper) {
			printk_Rlsi(KERN_WARNING 
				"megaide: log drv memory allocation failed\n");
			goto go_err_free_mem;
		}
	}

	/* allocating memory for physical drive */
	for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++)
	{
#ifdef _SOFT_BBM_
		PHYSICAL_DRIVE_INFO	*pd1 = &pdev->PhysicalDrive[i];
#endif /*_SOFT_BBM_*/
		
		pdev->PhysicalDrive[i].ErrorLogSector = 
			pci_alloc_consistent(dev, IDE_SECTOR_SIZE,
				&adapter->PhysicalDrive[i].h_err_log_sec);
				
		if (!pdev->PhysicalDrive[i].ErrorLogSector) {
			printk_Rlsi(KERN_WARNING 
				"megaide: phy drv memory allocation failed\n");
			goto go_err_free_mem;
		}

		memset_Rlsi(pdev->PhysicalDrive[i].ErrorLogSector, 0,
			IDE_SECTOR_SIZE);
		
		pdev->PhysicalDrive[i].DiskErrorLogSgl.Physical.Address =
		(UINT)adapter->PhysicalDrive[i].h_err_log_sec;
		pdev->PhysicalDrive[i].DiskErrorLogSgl.Physical.Length =
			IDE_SECTOR_SIZE;
		pdev->PhysicalDrive[i].DiskErrorLogSgl.Physical.EndOfListFlag =
			1;
		
#ifdef _SOFT_BBM_
		pd1->SBBMChanger.SbbmScb = (FAST_IDE_SCB_EXTENSION *)
			kmalloc(get_scb_extension_size(), GFP_KERNEL);
			
		pd1->SBBMChanger.pucSBBMData = 
			pci_alloc_consistent(dev, (SBBM_TABLE_SIZE_IN_BYTES +
				REGION_HW_BOUNDARY),
				&adapter->PhysicalDrive[i].h_sbbm_sec);
		/* memory allocated is alread REGION_HW_BOUNDARY aligned
		 * to make sure we allocated extra memory 
		 */
		if (adapter->PhysicalDrive[i].h_sbbm_sec % REGION_HW_BOUNDARY) {
			printk_Rlsi(KERN_WARNING 
				"megaide: sbbm memory allocation incorrect %d\n",
				(adapter->PhysicalDrive[i].h_sbbm_sec % REGION_HW_BOUNDARY));
		}
		
		if (!pd1->SBBMChanger.SbbmScb || 
			!pd1->SBBMChanger.pucSBBMData) {
			printk_Rlsi(KERN_WARNING 
				"megaide: sbbm memory allocation failed\n");
			goto go_err_free_mem;
		}

		memset_Rlsi(pd1->SBBMChanger.pucSBBMData, 0,
			SBBM_TABLE_SIZE_IN_BYTES);
		
		pd1->SBBMChanger.sbbm_sgl.Physical.Address =
			(UINT)adapter->PhysicalDrive[i].h_sbbm_sec;
		pd1->SBBMChanger.sbbm_sgl.Physical.Length =
			SBBM_TABLE_SIZE_IN_BYTES;
		pd1->SBBMChanger.sbbm_sgl.Physical.EndOfListFlag = 1;
#endif /*_SOFT_BBM_*/
	}

	/* IOCTL memory allocation */
	adapter->nit_scsi_cmd	= pci_alloc_consistent(dev, sizeof(Scsi_Cmnd),
		&adapter->h_nit_scsi_cmd);
	adapter->host_scribble	= pci_alloc_consistent(dev, 
		get_scb_extension_size(), &adapter->h_host_scribble);
	adapter->ioctl_data	= pci_alloc_consistent(dev,
		(NIT_MAX_DATA_TRANSFER + CHAR_HEADER_SIZE),
		&adapter->h_ioctl_data);
	
	if (!adapter->nit_scsi_cmd || !adapter->host_scribble ||
		!adapter->ioctl_data) {
		printk_Rlsi(KERN_WARNING 
			"megaide: ioctl memory allocation error\n");
		goto go_err_free_mem;
	}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	adapter->nit_scsi_device= pci_alloc_consistent(dev, 
		sizeof(struct scsi_device), &adapter->h_nit_scsi_device);
	if (!adapter->nit_scsi_device) {
		printk_Rlsi(KERN_WARNING 
			"megaide: ioctl memory allocation error\n");
		goto go_err_free_mem;
	}
#endif
	/* if memory allocation succeed return here */
	return TRUE;
	
/* memory allocation failed, remove the allocated memory */
go_err_free_mem:
	mide_remove_device_ext_mem(pdev, adapter);
	return FALSE;
}

/* store controller's base address in device extension */
BOOLEAN
mide_init_base_address(HW_DEVICE_EXTENSION *pdev, struct pci_dev *dev)
{
	INT	base5_reg_depth = 0;
	
	pdev->Channel[0].BaseIoAddress1 =
		(IDE_REGISTERS_1 *)pci_resource_start(dev, 0);
	pdev->Channel[0].BaseIoAddress2 =
		(IDE_REGISTERS_2 *)pci_resource_start(dev, 1);
	pdev->Channel[1].BaseIoAddress1 =
		(IDE_REGISTERS_1 *)pci_resource_start(dev, 2);
	pdev->Channel[1].BaseIoAddress2 =
		(IDE_REGISTERS_2 *)pci_resource_start(dev, 3);
	pdev->Channel[0].BaseBmAddress =
		(BM_REGISTERS *)pci_resource_start(dev, 4);
	pdev->Channel[1].BaseBmAddress =
		(BM_REGISTERS *)(pci_resource_start(dev, 4) + 8);

	switch (GET_DEVICE_ID(pdev)) {

	case PCI_DEVICE_ID_CMD_3112:
		base5_reg_depth = SIL_3112_BASE_REG_5_DEPTH;
		break;

	case PCI_DEVICE_ID_CMD_680:
	case PCI_DEVICE_ID_DELL_680:
		base5_reg_depth = SIL_680_BASE_REG_5_DEPTH;
		break;

	case PCI_DEVICE_ID_INTEL_ICH6R:
		base5_reg_depth = INTEL_ICH6R_BASE_REG_5_DEPTH;
		break;
	
	case PCI_DEVICE_ID_INTEL_ICH7R:
		base5_reg_depth = INTEL_ICH7R_BASE_REG_5_DEPTH;
		break;

	default:
		break;
	}

	/* there is no base5 register return here */
	if (!base5_reg_depth)
		return TRUE;

	pdev->baseAddress5 = pci_resource_start(dev, 5);
	pdev->virtBaseAddress5 = ioremap(pdev->baseAddress5, base5_reg_depth); 		if (!pdev->virtBaseAddress5) {
		printk_Rlsi(KERN_WARNING "BaseAddress5 iomap failed\n");
		return FALSE;
	}
	return TRUE;
}

VOID
initialize_device_extension(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		i;
	UCHAR		j;

	for (j = 0; j < MAX_LOGICAL_DRIVES; j++)
	{
		pdev->LogicalDrive[j].Status = LDS_NotPresent;
		
		for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++)
		{
			pdev->LogicalDrive[j].PhysicalDriveTid[i] = 
				INVALID_DRIVE_ID;
			pdev->LogicalDrive[j].SlotPhysicalDriveTid[i] = 
				INVALID_DRIVE_ID;
		}
	}

	for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++)
	{
		if (pdev->ucIRCDType == IRCD_3X) {
			pdev->PhysicalDrive[i].Status = PDS_Undefined;
		} else {
			pdev->PhysicalDrive[i].Status = PDS_OffLine;
		}

		pdev->PhysicalDrive[i].ucMirrorDriveId = INVALID_DRIVE_ID;
		pdev->PhysicalDrive[i].LogicalDriveId = INVALID_DRIVE_ID;
		pdev->PhysicalDrive[i].pArrayInfo = NULL;
		/*to correct proc display*/
		pdev->PhysicalDrive[i].PhysicalDriveId = i;
		ENABLE_IRCD_UPDATE_ON_DRIVE(pdev,i);
	}

	for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++)
	{
		/* changed this for supporting hetrogenious array */
		pdev->ArrayDrive[i].ucArrayDriveNumber = 0;
		pdev->ArrayDrive[i].ucLogicalDriveCount = 0;
		pdev->ArrayDrive[i].HeadLogDrive = NULL;
		pdev->ArrayDrive[i].TailLogDrive =NULL;
	}
}

/*
 * This function parse information from the option rom
 */
INT
mide_scan_option_rom(UCHAR * pcRom, UINT uiRomSize, HW_DEVICE_EXTENSION *pDE)
{
	CHAR		acTempStr[80];
	UINT		i, j, k;
	UINT		*puiTempPtr;
	USHORT		*pusROMSig;
	USHORT		bFoundVersion = 0;
	UINT		scan_depth;

	AR_DEBUG(AR_DBG_ROM, (KERN_DEBUG "megaide: mide_scan_option_rom:\n"));

	/*look every 2k*/
	for (i = 0; i < uiRomSize; i += 0x100)
	{
		pusROMSig = (USHORT *)&pcRom[i];
		if (*pusROMSig != 0xAA55) {
			continue;
		}

		/* look for our signature at offset 6 bytes */
		memmove(acTempStr, &pcRom[i+6], 8);
		acTempStr[8] = (CHAR)0;

		/* if both strings are not matching */
		if ((!strcmp(acTempStr, "HyperDis")) &&
			(!strcmp(acTempStr, "MegaRAID"))) {
			continue;
		}

		/*get firmware revision*/
		scan_depth = ((pcRom[i+2]*512) > uiRomSize) ?
			uiRomSize : (pcRom[i+2]*512);

		for (j=14; j < scan_depth ; j++)
		{
			puiTempPtr = (UINT *) &pcRom[i+j];

			/* looking for 'Hype'
			* looking for 'Mega'
			* looking for 'IDE '(this is for Compaq Bios)
			* looking for ''CERC(this is for Dell Bios)
			* looking for $XIDE$
			*/
			if (!(*puiTempPtr == 0x65707948) &&
				!(*puiTempPtr == 0x6167654D) &&
				!(*puiTempPtr == 0x20454449) &&
				!(*puiTempPtr == 0x43524543) &&
				!(*puiTempPtr == 0x44495824)) {
				continue;
			}

			memmove(acTempStr, &pcRom[i + j], 14);
			acTempStr[14] = (CHAR)0;

			/* heres $XIDE$ */
			if (!strncmp(acTempStr, "$XIDE$",6)) {
				j += 6;
				pDE->stFirmwareVersion.usMajorVer =
					pcRom[i + j] - '0';
				
				j ++;
				pDE->stFirmwareVersion.usMinorVer =
					pcRom[i + j] - '0';
				
				j ++;
				pDE->stFirmwareVersion.uiBuild =
					mide_array_to_int(&pcRom[i + j]);

				return(AR_TRUE);
			}

			/* the strcmp string should be 14 bytes.
			 * the 'j' counter should increment pointer to the
			 * location of 'V' begining at Version
			 * look at offset 26 for firmware version(12 bytes),
			 */
			else if (!strcmp(acTempStr, "MegaRAID IDE B")) {
				j += 26;
				bFoundVersion = 1;
			}

			/*
			 * heres old hypderdisk string 
			 * look at offset 23 for firmware version(12 bytes)
			 */
			else if (!strcmp(acTempStr, "HyperDisk BIOS")) {
				j += 23;
				bFoundVersion = 1;
			}

			/* 
			 * heres CPQ bios string
			 * look at offset 22 for firmware version(12 bytes) 
			 */
			else if (!strcmp(acTempStr, "IDE RAID BIOS ")) {
				j += 22;
				bFoundVersion = 1;
			}

			/*
			 * heres DELL bios string
			 * look at offset 22 for firmware version(12 bytes) 
			 */
			else if (!strcmp(acTempStr, "CERC ATA/2S BI")) {
				j += 25;
				bFoundVersion = 1;
			}

			if (bFoundVersion) {

				/* get usMajorVer */
				k = 0;
				/* look for '.'*/
				while (pcRom[i + j + k] != '.') k++;

				pDE->stFirmwareVersion.usMajorVer = 
					mide_array_to_int(&pcRom[i + j]);

				/* get usMinorVer */
				j += (k + 1);
				k = 0;
				/* look for '.'*/
				while (pcRom[i + j + k] != '.') k++;
				pDE->stFirmwareVersion.usMinorVer = 
					mide_array_to_int(&pcRom[i + j]);

				/*get uiBuild*/
				j += (k+1);
				pDE->stFirmwareVersion.uiBuild =
					mide_array_to_int(&pcRom[i + j]);

				/* yes, we want this */
				return(AR_TRUE);
			}
		}/*inner for loop*/
	} /*outer for loop*/
	
	return(AR_FALSE);
}


/**
 * function to get information from the option rom
 **/
INT
mide_get_option_rom_info(HW_DEVICE_EXTENSION *pdev)
{
	ULONG		ulAddr;
	UCHAR		*pucVirAddr;
	INT		iRetVal = AR_FALSE;

	AR_DEBUG(AR_DBG_ROM, (KERN_DEBUG "megaide: get option rom called\n"));

	for (ulAddr = 0xC0000, iRetVal = 0;
		(ulAddr < 0xF0000) && (iRetVal == 0); ulAddr += (PAGE_SIZE*8))
	{
		pucVirAddr = (UCHAR *)ioremap(ulAddr, (PAGE_SIZE*8));
		if (!pucVirAddr) {
			break;
		}

		if (AR_TRUE == mide_scan_option_rom(pucVirAddr, 
			(PAGE_SIZE*8), pdev)) {
			iRetVal = AR_TRUE;
		}
		
		iounmap(pucVirAddr);
	}
	return(iRetVal);
}

VOID
mega_set_switch_drive_on_chan(HW_DEVICE_EXTENSION *pdev, UCHAR chan)
{
	UINT	i;
	UCHAR phy = (chan<<1);

	/* If 2 drives found, set SwitchDrive to 1 */
	if ((pdev->PhysicalDrive[phy].DeviceFlags & DFLAGS_DEVICE_PRESENT)&&
		(pdev->PhysicalDrive[phy+1].DeviceFlags &
			DFLAGS_DEVICE_PRESENT)) {
		pdev->Channel[chan].SwitchDrive = 1;
	} else {
		pdev->Channel[chan].SwitchDrive = 0;
	}
	
	/* setting last drive fed is required if a drive inserted 
	 * at the slave slot 
	 */	
	pdev->Channel[chan].LastDriveFed = 0;
	for (i = 0; i < MAX_DRIVES_PER_CHANNEL; i++)
	{
		phy = (chan << 1) + i;
		if (pdev->PhysicalDrive[phy].DeviceFlags &
			DFLAGS_DEVICE_PRESENT) {
			pdev->Channel[chan].LastDriveFed = i;
		}
	}
}

/* This function is to gather information about the adapter's configuration.
 * In particular, it determines whether a RAID is present.
 */
BOOLEAN
InitIdeRaidController(HW_DEVICE_EXTENSION *pdev)
{
	BOOLEAN		foundDevices = AR_FALSE;
	UCHAR		channel;
	UCHAR		log;
	UCHAR		phy;
	CHAR		raid_type[20];
	CHAR		raid_status[20];
	UINT		sz;
	
	for (channel = 0; channel < MAX_CHANNELS_PER_CONTROLLER; channel++)
	{
		/* Search for devices on this controller.*/
		if (FindDevices(pdev, channel)) {
			AR_DEBUG(AR_DBG_INIT, (KERN_DEBUG 
				"megaide: found device in chan%d\n", channel));
			
			foundDevices = AR_TRUE;
			mega_set_switch_drive_on_chan(pdev, channel);
		}
	}

	if (foundDevices) {
		for (phy = 0; phy < MAX_DRIVES_PER_CONTROLLER; phy++)
		{
			if (!IS_IDE_DRIVE(pdev, phy)) {
				continue;
			}
			if (!GetIoMode(pdev, phy)) {
				/*Disable device.*/
				pdev->PhysicalDrive[phy].DeviceFlags &=
					~DFLAGS_DEVICE_PRESENT;
			}
			pdev->PhysicalDrive[phy].ucIsPowerConnected = 1;
		}
	}

	/* initialize boot log drv mapping */
	for (phy = 0; phy < MAX_LOGICAL_DRIVES; phy++)
	{
		pdev->BootLogDrvMapping[phy] = phy;
	}

	initialize_device_extension(pdev);
	
	if (GetConfigInfoAndErrorLogSectorInfo(pdev)) {

		ConfigureRaidDrives(pdev, pdev->ConfigurationVirtualBuffer);

		if (pdev->ucIRCDType == IRCD_3X) {
			EnquireProperties(pdev, TRUE);
		}
	}

	/* programming drives after Inquiry data collection */
	for (phy = 0; phy < MAX_DRIVES_PER_CONTROLLER; phy++)
	{
		if (pdev->PhysicalDrive[phy].ucIsPowerConnected) {
			SetMaxTransferMode(pdev, phy);
		}
	}

	if (pdev->uiLogicalDriveCount) {

		for (log = 0; log < pdev->uiLogicalDriveCount; log ++)
		{
			switch (pdev->LogicalDrive[log].RaidLevel) {
			
			case None:
				sprintf(raid_type, "%s", "NON-RAID");
				break;
			
			case Raid0:
				sprintf(raid_type, "%s", "RAID0");
				break;
			
			case Raid1:
				sprintf(raid_type, "%s", "RAID1");
				break;
			
			case Raid10:
				sprintf(raid_type, "%s", "RAID10");
				break;
			
			case SpareDrivePool:
				sprintf(raid_type, "%s", "SPAREPOOL");
				break;
			
			default:
				sprintf(raid_type, "%s", "UNKNOWN");
				break;
			}

			switch (pdev->LogicalDrive[log].Status) {
			
			case LDS_Online:
				sprintf(raid_status, "%s", "ONLINE");
				break;
			
			case LDS_Critical:
				sprintf(raid_status, "%s", "CRITICAL");
				break;
			
			case LDS_Degraded:
				sprintf(raid_status, "%s", "DEGRADED");
				break;
				
			case LDS_Deleted:
				sprintf(raid_status, "%s", "DELETED");
				break;
			
			case LDS_OffLine:
			default:
				sprintf(raid_status, "%s", "OFFLINE");
				break;
			}
			
			sz = (pdev->LogicalDrive[log].Sectors/2) * 
				(IDE_SECTOR_SIZE/256);
			
#ifdef _ATAPI_SUPPORT_
			if (pdev->LogicalDrive[log].RaidLevel != None) {
#endif /* _ATAPI_SUPPORT_ */
			printk_Rlsi(KERN_INFO
				"megaide: LD %d %s status=%s"
				" %d 512-byte sectors (%d MB) drives=%d\n",
				log, raid_type, raid_status,
				pdev->LogicalDrive[log].Sectors,
				(sz - sz/625 + 974)/1950,
				pdev->LogicalDrive[log].PhysicalDriveCount);
#ifdef _ATAPI_SUPPORT_
			}
#endif /* _ATAPI_SUPPORT_ */
		}
		
		InitDriveFeatures(pdev);
	}
	return AR_TRUE;
}

/* scan for all the devices attached to the IDE Raid controller and initialize
 * the Device Extetion structure
 */
INT
HyperDiskScanForDevices(HW_DEVICE_EXTENSION *pdev)
{
	if (!InitIdeRaidController(pdev)) {
		return (AR_ERROR);
	}
	
	AtapiHwInitialize(pdev);
	return(AR_SUCCESSS);
}


/* to identify devices attached to an IDE controller*/
BOOLEAN
FindDevices(HW_DEVICE_EXTENSION *pdev, UCHAR Channel)
{
	IDE_REGISTERS_1	*base1 = pdev->Channel[Channel].BaseIoAddress1;
	BOOLEAN		deviceResponded = AR_FALSE;
	UCHAR		deviceNumber;
	UCHAR		i;
	UCHAR		phy;

	AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
		"megaide: find devices Called\n"));

	/* Clear expecting interrupt flags.*/
	pdev->Channel[Channel].ExpectingInterrupt = 0;

	/*Search for devices*/
	for (deviceNumber = 0; deviceNumber < 2; deviceNumber++)
	{
		phy = (Channel << 1) + deviceNumber;
		
		/* serial ATA has only one drive per channel*/
		if (((pdev->ucCardType == SIL_3112) ||
			(pdev->ucCardType == INTEL_ICH5R)) && 
				(deviceNumber > 0)) {
			break;
		}
		
		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG
			"megaide: Searching Channel %d device%d\n",
			Channel, deviceNumber));

		if (!IsDrivePresent(pdev, phy)) {
			continue;
		}


		/*
		 * Issue IDE Identify. If an Atapi device is actually present,
		 * the signature will be asserted, and the drive will
		 * be recognized as such.
		 */

		if (IssueIdentify(pdev, deviceNumber, Channel,
			IDE_COMMAND_IDENTIFY)) {
			/*IDE drive found*/
			AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
				"megaide: Find TID %d is IDE\n", phy));

			deviceResponded = AR_TRUE;
		} else {
			/*Apparently no device*/
			continue;
		}
	}

	for (i = 0; i < MAX_CHANNELS_PER_CONTROLLER; i++)
	{
		phy = (Channel << 1) + i;

		if (pdev->PhysicalDrive[phy].DeviceFlags & 
			DFLAGS_DEVICE_PRESENT) {
			/*Initialize LastDriveFed.*/
			pdev->Channel[Channel].LastDriveFed = i;
/*<<*/
if (!(pdev->PhysicalDrive[phy].DeviceFlags & DFLAGS_ATAPI_DEVICE) &&
	deviceResponded) {
	/*
	* This hideous hack is to deal with ESDI devices
	* that return garbage geometry in the IDENTIFY data.
	* This is ONLY for the crashdump environment as
	* these are ESDI devices.
	*/

	if (pdev->PhysicalDrive[phy].IdentifyData.SectorsPerTrack==0x35&&
		pdev->PhysicalDrive[phy].IdentifyData.NumberOfHeads==0x07) {

		/*Change these values to something reasonable.*/
		pdev->PhysicalDrive[phy].IdentifyData.SectorsPerTrack = 0x34;
		pdev->PhysicalDrive[phy].IdentifyData.NumberOfHeads = 0x0E;
	}

	if (pdev->PhysicalDrive[phy].IdentifyData.SectorsPerTrack==0x35&&
		pdev->PhysicalDrive[phy].IdentifyData.NumberOfHeads==0x0F) {

		/*Change these values to something reasonable.*/
		pdev->PhysicalDrive[phy].IdentifyData.SectorsPerTrack = 0x34;
		pdev->PhysicalDrive[phy].IdentifyData.NumberOfHeads = 0x0F;
	}

	if (pdev->PhysicalDrive[phy].IdentifyData.SectorsPerTrack==0x36&&
		pdev->PhysicalDrive[phy].IdentifyData.NumberOfHeads==0x07) {

		/*Change these values to something reasonable.*/
		pdev->PhysicalDrive[phy].IdentifyData.SectorsPerTrack = 0x3F;
		pdev->PhysicalDrive[phy].IdentifyData.NumberOfHeads = 0x10;
	}
}
/*<<*/
		}
	}

	for (i = 0; i < 2; i++)
	{
		phy = (Channel << 1) + i;
		if (pdev->PhysicalDrive[phy].DeviceFlags & 
			DFLAGS_DEVICE_PRESENT) {
			/*select first available drive*/
			SELECT_DEVICE(base1, phy);
			break;
		}
	}

#ifdef DEBUG_ON
	if (!deviceResponded) {
		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG
			"megaide: No devices Found in Chan %d\n", Channel));
	}
#endif

	return deviceResponded;
}

BOOLEAN 
GetIoMode(HW_DEVICE_EXTENSION *pdev, UCHAR TargetId)
{
	IDENTIFY_DATA	*idenData;
	UINT		sectors;

#ifdef _ATAPI_SUPPORT_
	if (!IS_ATAPI_DRIVE(pdev, TargetId)) {
#endif /* _ATAPI_SUPPORT_ */
	idenData = &(pdev->PhysicalDrive[TargetId].IdentifyData);

	/* Calculate the total addressable sectors using CHS convention. */
	sectors = idenData->NumberOfCylinders * idenData->NumberOfHeads *
		idenData->SectorsPerTrack;

	AR_DEBUG(AR_DBG_INIT, (KERN_DEBUG "megaide: sectors%d \n", sectors));

	if (idenData->TranslationFieldsValid &
		IDENTIFY_FAST_TRANSFERS_SUPPORTED) {

		if ((idenData->Capabilities &
			IDENTIFY_CAPABILITIES_LBA_SUPPORTED)) {
			pdev->PhysicalDrive[TargetId].DeviceFlags |= 
				DFLAGS_LBA;
			/* Get the total number of addressable sectors. */
			sectors = idenData->UserAddressableSectors;
		}
	}

	pdev->PhysicalDrive[TargetId].NumberOfBlocks = sectors;

	if (IS_48_BIT_ENABLED(pdev, TargetId) &&
		idenData->AddressableSectors48bit[0]) {
		pdev->PhysicalDrive[TargetId].NumberOfBlocks48bit =
			idenData->AddressableSectors48bit[0];
	}
	
#ifdef _ATAPI_SUPPORT_
	}
#endif /* _ATAPI_SUPPORT_ */
	GetTransferMode(pdev, TargetId);

	return(AR_TRUE);
}


BOOLEAN 
GetTransferMode(HW_DEVICE_EXTENSION *pdev, UCHAR TargetId)
{
	IDENTIFY_DATA		*idenData;
	PHYSICAL_DRIVE_INFO	*pd;
	
	idenData	= &(pdev->PhysicalDrive[TargetId].IdentifyData);
	pd		= &(pdev->PhysicalDrive[TargetId]);
	
	/*By default assume this drive is PioMode 0*/
	pd->TransferMode = MP_PioMode0;

	if (idenData->AdvancedPioModes != 0) {
		if (idenData->AdvancedPioModes & IDENTIFY_PIO_MODE4) {
			pd->TransferMode = MP_PioMode4;
		} else {
			if (idenData->AdvancedPioModes & IDENTIFY_PIO_MODE3) {
				pd->TransferMode = MP_PioMode3;
			} else {
				pd->TransferMode = MP_PioMode0;
			}
		}
	}

	if (idenData->MultiWordDmaActive) {
		switch (idenData->MultiWordDmaActive) {
		case 1:
			pd->TransferMode = MP_DmaMode0;
			pd->DeviceFlags |= DFLAGS_USE_DMA;
			break;

		case 2:
			pd->TransferMode = MP_DmaMode1;
			pd->DeviceFlags |= DFLAGS_USE_DMA;
			break;

		case 4:
			pd->TransferMode = MP_DmaMode2;
			pd->DeviceFlags |= DFLAGS_USE_DMA;
			break;
		}
	}

	if (idenData->UltraDmaActive) {
		switch (idenData->UltraDmaActive)
		{
		case 1:
			pd->TransferMode = MP_UdmaMode0;
			pd->DeviceFlags |= DFLAGS_USE_UDMA;
			break;

		case 2:
			pd->TransferMode = MP_UdmaMode1;
			pd->DeviceFlags |= DFLAGS_USE_UDMA;
			break;

		case 4:
			pd->TransferMode = MP_UdmaMode2;
			pd->DeviceFlags |= DFLAGS_USE_UDMA;
			break;

		case 8:
			pd->TransferMode = MP_UdmaMode3;
			pd->DeviceFlags |= DFLAGS_USE_UDMA;
			break;

		case 0x10:
			pd->TransferMode = MP_UdmaMode4;
			pd->DeviceFlags |= DFLAGS_USE_UDMA;
			break;

		case 0x20:
			pd->DeviceFlags |= DFLAGS_USE_UDMA;

			/*to take care of higher Udma speed*/
			if (Udma100 >= pdev->ControllerSpeed) {
				pd->TransferMode = MP_UdmaMode5;
			} else {
				pd->TransferMode = MP_UdmaMode4;
			}
			break;

		/* Ultra DMA 6 support*/
		case 0x40:
			pd->DeviceFlags |= DFLAGS_USE_UDMA;
			
			if (Udma133 == pdev->ControllerSpeed) {
				pd->TransferMode = MP_UdmaMode6;
			} else if (Udma100 == pdev->ControllerSpeed) {
				pd->TransferMode = MP_UdmaMode5;
			} else {
				pd->TransferMode = MP_UdmaMode4;
			}
			break;
		}
	}
	return AR_TRUE;
}

BOOLEAN
AssignSglPtrsForChannels(HW_DEVICE_EXTENSION *pdev)
{
	CHANNEL		*pChannel;
	SGL_ENTRY	*sgl;
	INT		lengthInNextBoundary;
	INT		lengthLeftInBoundary;
	UINT		maxSglEntries, chan;
	ULONG		sglBasePhysicalAddress;
	UINT		length;
	ULONG		ulStartAddress;
	ULONG		ulEndAddress;
	UINT		uiStartPage, uiEndPage;
	UINT		uiLengthReduced = 0;
	UCHAR		ucControllerId;
	adapter_t	*adapter = (adapter_t *)pdev->adapter;

	ucControllerId = pdev->ucControllerId;

	pChannel = &(pdev->Channel[0]);

	for (chan = 0; chan < MAX_CHANNELS_PER_CONTROLLER; chan++)
	{
		/*Dynamic DMA Implementation*/
		pChannel[chan].pSglEntry = 
			pci_alloc_consistent(adapter->pstDev,
		(MAX_SGL_ENTRIES_PER_PHYSICAL_DRIVE * sizeof(SGL_ENTRY)),
			&adapter->Channel[chan].h_sgl_entry);

		if (pChannel[chan].pSglEntry == NULL) {
			printk_Rlsi(KERN_ERR 
				"megaide: pci_alloc failed SglEntry\n");
			return FALSE;
		}

		memset_Rlsi(pChannel[chan].pSglEntry, 0,
		MAX_SGL_ENTRIES_PER_PHYSICAL_DRIVE * sizeof(SGL_ENTRY));

		sglBasePhysicalAddress = 
			adapter->Channel[chan].h_sgl_entry;

		sgl = pChannel[chan].pSglEntry;

		if (sglBasePhysicalAddress == 0) {
			printk_Rlsi(KERN_ERR 
				"megaide: Physical Address Came as Zero.\n");
			return(FALSE);
		}

		if (sglBasePhysicalAddress % SGL_HW_BOUNDARY) {
			length = 
			(((sglBasePhysicalAddress / SGL_HW_BOUNDARY) + 1) *
				SGL_HW_BOUNDARY) - sglBasePhysicalAddress;
		} else {
			length = SGL_HW_BOUNDARY;
		}

		if (length >= 
		(MAX_SGL_ENTRIES_PER_PHYSICAL_DRIVE * sizeof(SGL_ENTRY))) {
			length = 
			MAX_SGL_ENTRIES_PER_PHYSICAL_DRIVE * sizeof(SGL_ENTRY);
		}

		ulStartAddress	= sglBasePhysicalAddress;
		ulEndAddress	= ulStartAddress + length;


		/* Check to see if SGL Pointer is DWORD Aligned and make it so.
		 * Remember we allocated an extra SGL entry for this purpose.
		 */

		if (sglBasePhysicalAddress & 3) {

			sglBasePhysicalAddress = (( 
				sglBasePhysicalAddress & 0xfffffffcL) + 4L);
			uiLengthReduced = 
				sglBasePhysicalAddress - ulStartAddress;
			ulStartAddress = sglBasePhysicalAddress;
			length -= uiLengthReduced;
			sgl += uiLengthReduced;
		}

		uiStartPage	= (UINT)(ulStartAddress / SGL_HW_BOUNDARY);
		uiEndPage	= (UINT)(ulEndAddress / SGL_HW_BOUNDARY);

		if (uiStartPage == uiEndPage) {
			/*all entries are in the same page*/
			maxSglEntries = (length / sizeof(SGL_ENTRY));
		} else {
			/* Make sure the SGL does NOT cross a 4K boundary.*/
			lengthLeftInBoundary = SGL_HW_BOUNDARY - 
			((UINT) sglBasePhysicalAddress &(SGL_HW_BOUNDARY - 1));
			lengthInNextBoundary = length - lengthLeftInBoundary;

			if (lengthInNextBoundary > lengthLeftInBoundary) {
			/* SGL crosses 4KB boundary and top portion is larger
			* than bottom. Use top portion.
			*/
				sglBasePhysicalAddress =
				sglBasePhysicalAddress + lengthLeftInBoundary;
				
				sgl = (SGL_ENTRY *)
					((ULONG)sgl + lengthLeftInBoundary);
				
				maxSglEntries = 
					lengthInNextBoundary/sizeof(SGL_ENTRY);
			} else {
				maxSglEntries = 
					lengthLeftInBoundary/sizeof(SGL_ENTRY);
			}
		}

		/*Save SGL info for next call.*/
		pChannel[chan].SglBaseVirtualAddress = (SGL_ENTRY *) sgl;
		pChannel[chan].SglBasePhysicalAddress = sglBasePhysicalAddress;
		pChannel[chan].MaxSglEntries = maxSglEntries;
	}

	return TRUE;
}



/* Removes the build in IDE driver if present*/
VOID
RemoveBuildInIDESupport(struct pci_device_id *id)
{
	struct pci_dev		*pstPciDev;
	INT			iCount;
	USHORT			subvendor;
	USHORT			subdevice;

	AR_DEBUG(AR_DBG_INIT, (KERN_DEBUG 
		"megaide: remove native driver support\n"));

	for (iCount = 0; iCount < MAX_HWIFS; ++iCount)
	{
		pstPciDev = ide_hwifs[iCount].pci_dev;

		if (!pstPciDev || ide_hwifs[iCount].present == 0) {
			continue;
		}

		subvendor = ide_hwifs[iCount].pci_dev->subsystem_vendor;
		subdevice = ide_hwifs[iCount].pci_dev->subsystem_device;
		
		/* Remove Only requested Device */
		if ((id->device == ide_hwifs[iCount].pci_dev->device) &&
			(id->vendor == ide_hwifs[iCount].pci_dev->vendor) &&
			((id->subvendor & subvendor) == subvendor) &&
			((subdevice & id->subdevice) == subdevice)) {
			
			AR_DEBUG(AR_DBG_INIT, (KERN_DEBUG 
				"megaide: ide_hwifs[%d].present=%d\n",
				iCount, ide_hwifs[iCount].present));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
			spin_unlock_irq(&io_request_lock);
			ide_unregister(iCount);
			spin_lock_irq(&io_request_lock);
#endif			
		}
	}
}

/*This function check for the io region availability and claim
 *those io regions for megaide driver
 */
	BOOLEAN
ClaimOurIOPorts(HW_DEVICE_EXTENSION *pdev)
{

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) 
	if (check_region((ULONG)pdev->Channel[0].BaseIoAddress1, 8) ||
			check_region((ULONG)pdev->Channel[0].BaseIoAddress2, 4) ||
			check_region((ULONG)pdev->Channel[1].BaseIoAddress1, 8) ||
			check_region((ULONG)pdev->Channel[1].BaseIoAddress2, 4) ||
			check_region((ULONG)pdev->Channel[0].BaseBmAddress, 8) ||
			check_region((ULONG)pdev->Channel[1].BaseBmAddress, 8)) {

		printk_Rlsi(KERN_WARNING 
				"megaide: check resource failed!\n");
		return (AR_FALSE);
	}
#endif	
	if (!request_region((ULONG)pdev->Channel[0].BaseIoAddress1,
				8, "megaide"))
		goto release_step1;

	if (!request_region((ULONG)pdev->Channel[0].BaseIoAddress2,
				4, "megaide"))
		goto release_step2;

	if (!request_region((ULONG)pdev->Channel[1].BaseIoAddress1,
				8, "megaide"))
		goto release_step3;

	if (!request_region((ULONG)pdev->Channel[1].BaseIoAddress2,
				4, "megaide"))
		goto release_step4;

	/* checking success */
	if(request_region((ULONG)pdev->Channel[0].BaseBmAddress,
				16, "megaide"))
		return (AR_TRUE);
	
	release_region((ULONG)pdev->Channel[1].BaseIoAddress2, 4);
	
release_step4:
	release_region((ULONG)pdev->Channel[1].BaseIoAddress1, 8);
	
release_step3:
	release_region((ULONG)pdev->Channel[0].BaseIoAddress2, 4);
	
release_step2: 
	release_region((ULONG)pdev->Channel[0].BaseIoAddress1, 8);

release_step1:
	printk_Rlsi(KERN_WARNING "megaide: request resource failed!\n");
	return (AR_FALSE);
}


BOOLEAN
ReleaseOurIOPorts(HW_DEVICE_EXTENSION *pdev)
{

	AR_DEBUG(AR_DEBUG_RELEASE, (KERN_DEBUG 
		"megaide: ReleaseOur IOPorts\n"));
	
	release_region((ULONG)pdev->Channel[0].BaseIoAddress1, 8);
	release_region((ULONG)pdev->Channel[0].BaseIoAddress2, 4);
	release_region((ULONG)pdev->Channel[1].BaseIoAddress1, 8);
	release_region((ULONG)pdev->Channel[1].BaseIoAddress2, 4);
	release_region((ULONG)pdev->Channel[0].BaseBmAddress, 16);
	return AR_TRUE;
}

/*
 * This function will check for the presence of the device with the
 * given target ID
 */
BOOLEAN 
IsDrivePresent(HW_DEVICE_EXTENSION *pdev, UCHAR tid)
{
	IDE_REGISTERS_1	*base1 = pdev->Channel[tid>>1].BaseIoAddress1;
	UCHAR		ucStatus;

	AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
		"megaide: is_drive _present chan %d \n", (tid>>1)));
	
	if(GET_DEVICE_ID(pdev)==PCI_DEVICE_ID_CMD_3112) {
	
		/*One drive per channel*/
		if (tid & 0x01) return AR_TRUE;
		
		if (0x03/*DET_DEVICE_DETECTED_PHY_GOOD*/ != 
			AthenaDetectPhysicalChange(pdev, (tid>>1))) {
			AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
				"megaide: drv not responding at %d \n", tid));
			return AR_FALSE;
		} else {
			AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
				"megaide: drv responding at %d \n", tid));
			return AR_TRUE;
		}
	}
	
	SELECT_DEVICE(base1, tid);
	GET_STATUS(base1, ucStatus);
	
	if (0xff == ucStatus || 0x7f == ucStatus) {
		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
			"megaide: No divice Status returned as 0xff\n"));
		return AR_FALSE;
	}
	
	SELECT_DEVICE(base1, tid);
	ScsiPortWritePortUchar((ULONG)&base1->SectorCount, 0x5a);
	ScsiPortStallExecution(60);
	
	if (0x5a == ScsiPortReadPortUchar((ULONG)&base1->SectorCount)) {
		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
			"megaide: Drv %d is present\n", tid));
		return AR_TRUE;
	} else {
		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
			"megaide: Drv %d is not present\n", tid));
		return AR_FALSE;
	}
}



/*Issue IDENTIFY command to a device*/
BOOLEAN 
IssueIdentify(HW_DEVICE_EXTENSION *pdev, UCHAR DeviceNo, 
	UCHAR Channel, UCHAR Command)
{
	IDE_REGISTERS_1	*base1 = pdev->Channel[Channel].BaseIoAddress1;
	UINT		i;
	UCHAR		status;
	UCHAR		phy;
	USHORT		*puIdentifyData;
	BM_REGISTERS	*baseBm = pdev->Channel[Channel].BaseBmAddress;

	AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG "megaide: Issue Identify:\n"));

	/* Select device 0 or 1*/
	SELECT_DEVICE(base1, DeviceNo);

#ifdef _ATAPI_SUPPORT_
	/* Send Packet IDENTIFY command first */
	ScsiPortWritePortUchar((ULONG)&base1->Command, 0xA1);

	WAIT_ON_BUSY(base1, status);
	
	if ((!(status & IDE_STATUS_BUSY)) && (!(status & IDE_STATUS_DRQ))) {
		ScsiPortWritePortUchar((ULONG)&base1->CylinderHigh,
			(0x200 >> 8));
		ScsiPortWritePortUchar((ULONG)&base1->CylinderLow,
			(0x200 & 0xFF));
		ScsiPortWritePortUchar((ULONG)&base1->Command,
			Command);

		WAIT_ON_BUSY(base1, status);
		if ((!(status & IDE_STATUS_BUSY)) &&
			(!(status & IDE_STATUS_DRQ))) {

			AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG
				"megaide: Identtify Failed here\n"));
			//this is an error... so let us not try any more
			if (IS_ICHxR_CONTROLLER(pdev)) {
				UCHAR status1;
				CLEAR_BM_INT(baseBm, status1);
			}
			return AR_FALSE;
		}
	}
#else
	GET_BASE_STATUS(base1, status);

	/* Load CylinderHigh and CylinderLow with number bytes to transfer*/
	ScsiPortWritePortUchar((ULONG)&base1->CylinderHigh, (0x200 >> 8));
	ScsiPortWritePortUchar((ULONG)&base1->CylinderLow, (0x200 & 0xFF));

	/*Send IDENTIFY command.*/
	ScsiPortWritePortUchar((ULONG)&base1->Command, Command);

	WAIT_ON_BUSY(base1, status);

	if ((!(status & IDE_STATUS_BUSY)) && (!(status & IDE_STATUS_DRQ))) {
		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
			"megaide: Identtify Failed\n"));
		/* this is an error... so let us not try any more.*/
		if (IS_ICHxR_CONTROLLER(pdev)) {
			UCHAR status1;
			CLEAR_BM_INT(baseBm, status1);
		}
		return AR_FALSE;
	}
#endif /* _ATAPI_SUPPORT_ */
	WAIT_ON_BUSY(base1, status);

	/* Wait for DRQ. */
	WAIT_FOR_DRQ(base1, status);

	/* Read status to acknowledge any interrupts generated. */
	GET_BASE_STATUS(base1, status);

	/*
	 * Check for error on really stupid master devices
	 * that assert random patterns of bits in the status
	 * register at the slave address.
	 */

	if ((Command == IDE_COMMAND_IDENTIFY) && (status & IDE_STATUS_ERROR)) {
		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
			"megaide: Identify Failed 3\n"));
		if (IS_ICHxR_CONTROLLER(pdev)) {
			UCHAR status1;
			CLEAR_BM_INT(baseBm, status1);
		}
		return AR_FALSE;
	}

	/*
	 * Suck out 256 words. After waiting for one model
	 * that asserts busy after receiving the Packet Identify command.
	 */

	WAIT_ON_BUSY(base1, status);

	if (!(status & IDE_STATUS_DRQ)) {
		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
			"megaide: Failed to access\n"));
		if (IS_ICHxR_CONTROLLER(pdev)) {
			UCHAR status1;
			CLEAR_BM_INT(baseBm, status1);
		}
		return AR_FALSE;
	}

	phy = (Channel << 1) + DeviceNo;

	memset_Rlsi(&pdev->PhysicalDrive[phy].IdentifyData,
		0, sizeof(IDENTIFY_DATA));

	AR_READ_BUFFER(base1,
		(USHORT *)&pdev->PhysicalDrive[phy].IdentifyData, 256);

	/* Check out a few capabilities / limitations of the device.*/
	if (pdev->PhysicalDrive[phy].IdentifyData.MediaStatusNotification &1) {
		/* Determine if this drive supports the MSN functions.*/

		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
			"megaide:Identify:Marking drive %d as removable\n",
			(Channel * 2 + DeviceNo)));

		pdev->PhysicalDrive[(Channel * 2) + DeviceNo].DeviceFlags |=
			DFLAGS_REMOVABLE_DRIVE;
	}

	if (pdev->PhysicalDrive[phy].IdentifyData.MaximumBlockTransfer) {
		/* Determine max. block transfer for this device*/
		pdev->PhysicalDrive[phy].MaximumBlockXfer = (UCHAR)
	(pdev->PhysicalDrive[phy].IdentifyData.MaximumBlockTransfer & 0xFF);
	}

	puIdentifyData = (USHORT *) &(pdev->PhysicalDrive[phy].IdentifyData);

	if (pdev->PhysicalDrive[phy].IdentifyData.GeneralConfiguration&0x20 &&
		Command != IDE_COMMAND_IDENTIFY) {

		/* This device interrupts with the assertion of DRQ after
		* receiving Atapi Packet Command
		*/
		pdev->PhysicalDrive[phy].DeviceFlags |= DFLAGS_INT_DRQ;

		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
			"megaide: dev will interrupts on DRQassertion\n"));
	}
#ifdef AR_DEBUG
	else {
		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
			"megaide: dev will not interrupt on DRQassertion\n"));
	}
#endif
	if (((pdev->PhysicalDrive[phy].IdentifyData.GeneralConfiguration &
		0xF00) == 0x100) && Command != IDE_COMMAND_IDENTIFY) {
		/*This is a tape.*/
		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
			"megaide: Identify: Device is a tape drive.\n"));
		pdev->PhysicalDrive[phy].DeviceFlags |= DFLAGS_TAPE_DEVICE;
	} else {
		AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
			"megaide: Identify:Device is not a tape drive.\n"));
	}

	/*
	 * Work around for some IDE and one model Atapi that
	 * will present more than 256 bytes for the Identify data.
	 */

	WAIT_ON_BUSY(base1, status);

	for (i = 0; i < 0x10000; i++)
	{
		GET_STATUS(base1, status);

		if (status & IDE_STATUS_DRQ) {
			/*Suck out any remaining bytes and throw away.*/
			ScsiPortReadPortUshort((ULONG)&base1->Data);
		} else {
			break;
		}
	}

	AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG
		"megaide: Status after read words(%x) \n", status));

	GET_BASE_STATUS(base1, status);

	if (IS_ICHxR_CONTROLLER(pdev)) {
		CLEAR_BM_INT(baseBm, status);
	}
	
	pdev->PhysicalDrive[phy].DeviceFlags |= DFLAGS_DEVICE_PRESENT;

#ifdef _ATAPI_SUPPORT_
	if (pdev->PhysicalDrive[phy].IdentifyData.GeneralConfiguration & 0x8000) {
		pdev->PhysicalDrive[phy].DeviceFlags |= DFLAGS_ATAPI_DEVICE;
	} else {
		pdev->PhysicalDrive[phy].DeviceFlags &= ~DFLAGS_ATAPI_DEVICE;
	}
#else			
	pdev->PhysicalDrive[phy].DeviceFlags &= ~DFLAGS_ATAPI_DEVICE;
#endif /* _ATAPI_SUPPORT_ */

	return AR_TRUE;
}


BOOLEAN 
InitDriveFeatures(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		i;

	for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++)
	{
		mega_set_rw_cache_mode(pdev, i);
		set_multi_block_xfers(pdev, i);
	}
	
	return AR_TRUE;
}

BOOLEAN
set_multi_block_xfers(HW_DEVICE_EXTENSION *pdev, UCHAR tid)
{
	IDE_REGISTERS_1		*base1;
	UCHAR			statusByte = 0;
	UCHAR			err;

	base1 = pdev->Channel[tid>>1].BaseIoAddress1;

	AR_DEBUG(AR_DEBUG_FINDDEV, (KERN_DEBUG 
		"megaide: set multiblock transfer called for drv%d \n", tid));

	/*If supported, setup Multi-block transfers.*/
	if (!pdev->PhysicalDrive[tid].MaximumBlockXfer) {
		return(AR_FALSE);
	}
	/*Select the device.*/
	SELECT_DEVICE(base1, tid);

	/* Setup sector count to reflect the # of blocks.*/
	ScsiPortWritePortUchar((ULONG)&base1->SectorCount,
		pdev->PhysicalDrive[tid].MaximumBlockXfer);

	/* Issue the command.*/
	ScsiPortWritePortUchar((ULONG)&base1->Command,
		IDE_COMMAND_SET_MULTIPLE);

	/*Wait for busy to drop.*/
	WAIT_ON_BASE_BUSY(base1, statusByte);

	/*
	* Check for errors. Reset the value to 0(disable MultiBlock)
	* if the command was aborted.
	*/

	if (statusByte & IDE_STATUS_ERROR) {

		/*Read the error register.*/
		err = ScsiPortReadPortUchar((ULONG)((UCHAR *) base1 + 1));

		AR_DEBUG(AR_DBG_INIT, (KERN_DEBUG 
			"megaide: multmode setting status=%x, err=%x\n",
			statusByte, err));

		/* Adjust the devExt. value, if necessary.*/
		pdev->PhysicalDrive[tid].MaximumBlockXfer = 0;
	} 
	
	if (IS_ICHxR_CONTROLLER(pdev)) {
		BM_REGISTERS	*baseBm = pdev->Channel[tid>>1].BaseBmAddress;
		CLEAR_BM_INT(baseBm, statusByte);
	}
	return AR_TRUE;
}

BOOLEAN mega_set_rw_cache_mode(HW_DEVICE_EXTENSION *pdev, UCHAR ucDrvInd)
{
	IDE_REGISTERS_1		*base1;
	UCHAR			statusByte;

	if (!IS_IDE_DRIVE(pdev, ucDrvInd))
		return AR_FALSE;

	base1 = pdev->Channel[ucDrvInd >> 1].BaseIoAddress1;

	SELECT_DEVICE(base1, ucDrvInd);
	
	if (pdev->PhysicalDrive[ucDrvInd].bEnableRwCache) {
		AR_DEBUG(AR_DEV1, (
			"megaide: w-cache enabled on drv %d\n",	ucDrvInd));
		ScsiPortWritePortUchar((ULONG)((UCHAR *) base1 + 1),
			FEATURE_ENABLE_WRITE_CACHE);
	} else {
		AR_DEBUG(AR_DEV1, (
			"megaide: w-cache disabled on drv %d\n", ucDrvInd));
		ScsiPortWritePortUchar((ULONG)((UCHAR *) base1 + 1),
			FEATURE_DISABLE_WRITE_CACHE);
	}

	ScsiPortWritePortUchar((ULONG)&(base1->Command),
		IDE_COMMAND_SET_FEATURES);
	WAIT_ON_BASE_BUSY(base1, statusByte);

	if (pdev->PhysicalDrive[ucDrvInd].bEnableReadAheadCache) {
		AR_DEBUG(AR_DEV1, (
			"megaide: r-ahead enabled on drv %d\n",	ucDrvInd));
		ScsiPortWritePortUchar((ULONG)((UCHAR *) base1 + 1),
			FEATURE_ENABLE_READ_CACHE);
	} else {
		AR_DEBUG(AR_DEV1, (
			"megaide: r-ahead disabled on drv %d\n", ucDrvInd));
		ScsiPortWritePortUchar((ULONG)((UCHAR *) base1 + 1),
			FEATURE_DISABLE_READ_CACHE);
	}

	ScsiPortWritePortUchar((ULONG)&(base1->Command),
		IDE_COMMAND_SET_FEATURES);
	
	WAIT_ON_BASE_BUSY(base1, statusByte);
	
	GET_BASE_STATUS(base1, statusByte);
	
	if (IS_ICHxR_CONTROLLER(pdev)) {
		BM_REGISTERS *baseBm = pdev->Channel[ucDrvInd>>1].BaseBmAddress;
		CLEAR_BM_INT(baseBm, statusByte);
	}
	
	return AR_TRUE;
}

/**
 * Set all drive parameters
 **/
VOID
SetDriveParameters(HW_DEVICE_EXTENSION *pdev)
{
	UCHAR		phy;
	
	for (phy = 0; phy < MAX_DRIVES_PER_CONTROLLER; phy++) 
	{
		SetDriveParams(pdev, phy);
	}

}


/**
 * Set single drive parameters
 **/
VOID
SetDriveParams(HW_DEVICE_EXTENSION *pdev, UCHAR phy)
{
	UINT		i;
	UCHAR		statusByte = 0;
	UCHAR		Channel;
	IDE_REGISTERS_1	*base1;
	IDE_REGISTERS_2	*base2;
	IDENTIFY_DATA	*idenData = NULL;
	BM_REGISTERS	*baseBm;

	if (!IS_IDE_DRIVE(pdev, phy)) {
		return;
	}

	Channel	= phy >> 1;
	base1	= (IDE_REGISTERS_1 *) pdev->Channel[Channel].BaseIoAddress1;
	base2	= (IDE_REGISTERS_2 *) pdev->Channel[Channel].BaseIoAddress2;
	idenData= &pdev->PhysicalDrive[phy].IdentifyData;


	/*Set up registers for SET PARAMETER command.*/
	ScsiPortWritePortUchar((ULONG)&base1->DriveSelect,
		(UCHAR)(((phy << 4) | 0xA0) |(idenData->NumberOfHeads - 1)));

	ScsiPortWritePortUchar((ULONG)&base1->SectorCount,
		(UCHAR) idenData->SectorsPerTrack);

	/* Send SET PARAMETER command*/
	ScsiPortWritePortUchar((ULONG)&base1->Command,
		IDE_COMMAND_SET_DRIVE_PARAMETERS);

	ScsiPortStallExecution(1000);

	/*Wait for up to 30 milliseconds for ERROR or command complete*/
	for (i = 0; i < 30 * 1000; i++)
	{
		UCHAR		err;

		statusByte =
			ScsiPortReadPortUchar((ULONG)&base2->AlternateStatus);

		if (statusByte & IDE_STATUS_ERROR) {
			err = ScsiPortReadPortUchar(
				(ULONG)((UCHAR *) base1 + 1));

			AR_DEBUG(AR_DBG_INIT, (KERN_DEBUG
				"megaide: Error bit set Status %x, error %x\n",
				err, statusByte));
			break;
		} else if ((statusByte & ~IDE_STATUS_INDEX) == IDE_STATUS_IDLE) {
			break;
		} else {
			ScsiPortStallExecution(100);
		}
		
		if (IS_ICHxR_CONTROLLER(pdev)) {
			baseBm = pdev->Channel[phy>>1].BaseBmAddress;
			CLEAR_BM_INT(baseBm, statusByte);
		}
	}
}

VOID 
CleanUpAdapter(HW_DEVICE_EXTENSION *pdev)
{
	adapter_t	*adapter = (adapter_t *)pdev->adapter;

	AR_DEBUG(AR_DEBUG_RELEASE, (KERN_DEBUG 
		"megaide: clean up adapter called\n"));
	
	if (!pdev || !adapter) {
		return;
	}
	
	megaide_kill_backgroundtask_and_wait(pdev);

	mide_remove_device_ext_mem(pdev, adapter);
	
	if (pdev->virtBaseAddress5) {
		iounmap ((VOID *)pdev->virtBaseAddress5);
	}
}

BOOLEAN
PostCommandToDrive(HW_DEVICE_EXTENSION *pdev, UINT uiTargetId,
	UINT uiStartBlock, UINT uiBlockCount, UCHAR ucCmd,
	ULONG ulSglPhysicalAddress)
{
	UINT		chan = uiTargetId >> 1;
	CHANNEL		*pChannel = &(pdev->Channel[chan]);
	UCHAR		statusByte, ucSectorNumber,
			ucCylinderLow, ucCylinderHigh, bmStatus;
	UCHAR		bmCommand = 0;
	IDE_REGISTERS_1	*base1 = pChannel->BaseIoAddress1;
	BM_REGISTERS	*bmBase = pChannel->BaseBmAddress;
	UCHAR		ucSectorNumberExt, b48bitFlag = FALSE;

	ucSectorNumber	= (UCHAR)(uiStartBlock & 0x000000ff);
	ucCylinderLow	= (UCHAR)((uiStartBlock & 0x0000ff00) >> 8);
	ucCylinderHigh	= (UCHAR)((uiStartBlock & 0x00ff0000) >> 16);

	if (IS_48_BIT_ADDESSABLE(pdev, uiTargetId) &&
		(uiStartBlock > 0xFFFFFFF)) {

		b48bitFlag		= TRUE;
		ucSectorNumberExt	= (uiStartBlock & 0xff000000) >> 24;

		SELECT_LBA_DEVICE_48BIT(base1, uiTargetId);

		/* Previous contents */
		ScsiPortWritePortUchar((ULONG)&base1->SectorCount,
			((uiBlockCount & 0xFF00)>>8));
		ScsiPortWritePortUchar((ULONG)&base1->SectorNumber,
			ucSectorNumberExt);
		ScsiPortWritePortUchar((ULONG)&base1->CylinderLow, 0);
		ScsiPortWritePortUchar((ULONG)&base1->CylinderHigh, 0);

		/* New contents */
		ScsiPortWritePortUchar((ULONG)&base1->SectorCount,
			uiBlockCount & 0xFF);
		ScsiPortWritePortUchar((ULONG)&base1->SectorNumber,
			ucSectorNumber);
		ScsiPortWritePortUchar((ULONG)&base1->CylinderLow,
			ucCylinderLow);
		ScsiPortWritePortUchar((ULONG)&base1->CylinderHigh,
			ucCylinderHigh);
	} else {
		SELECT_LBA_DEVICE(base1, uiTargetId, uiStartBlock);

		ScsiPortWritePortUchar((ULONG)&base1->SectorCount,
			(UCHAR) uiBlockCount);
		ScsiPortWritePortUchar((ULONG)&base1->SectorNumber,
			ucSectorNumber);
		ScsiPortWritePortUchar((ULONG)&base1->CylinderLow,
			ucCylinderLow);
		ScsiPortWritePortUchar((ULONG)&base1->CylinderHigh,
			ucCylinderHigh);
	}

	switch (ucCmd) {
	case READ:
	case READ_6:
	case READ_10:
		/* Here no need to handle WriteBack commands cause
		* this function is only used for BBW
		*/
		if (b48bitFlag==TRUE) {
			ScsiPortWritePortUchar((ULONG)&base1->Command,
				IDE_COMMAND_READ_DMA_EXT);
		} else {
			ScsiPortWritePortUchar((ULONG)&base1->Command,
				IDE_COMMAND_READ_DMA);
		}
		bmCommand = READ_TRANSFER;
		break;
	case WRITE:
	case WRITE_6:
	case WRITE_10:
		if (b48bitFlag==TRUE) {
			ScsiPortWritePortUchar((ULONG)&base1->Command,
				IDE_COMMAND_WRITE_DMA_EXT);
		} else {
			ScsiPortWritePortUchar((ULONG)&base1->Command,
				IDE_COMMAND_WRITE_DMA);
		}
		bmCommand = WRITE_TRANSFER;
		break;

	case VERIFY:
		/*verify command doesn't have any buffer transfers*/
		ScsiPortWritePortUchar((ULONG)&base1->Command,
				IDE_COMMAND_VERIFY);
		pChannel->ExpectingInterrupt = IDE_PIO_INTERRUPT;
		return TRUE;
		break;
	}

	bmStatus = ScsiPortReadPortUchar((ULONG)&(bmBase->Status.AsUchar));

	/*Clear interrupt and error bits(if set)*/
	CLEAR_BM_INT(bmBase, statusByte);

	ScsiPortWritePortUlong((ULONG)&(bmBase->SglAddress),
		ulSglPhysicalAddress);

	/*Start transfer*/
	ScsiPortWritePortUchar((ULONG)&(bmBase->Command.AsUchar),
		(UCHAR)(bmCommand | START_TRANSFER));

	return TRUE;
}

HW_DEVICE_EXTENSION *
GetDeviceExtension(UCHAR ucIndex)
{
	if (ucIndex < giNumberOfCtrlrs) {
		return (&gastCardInfo[ucIndex]->DeviceExtension);
	} else {
		return NULL;
	}
}

/**
 * Description
 * 	This funtion returns the integer equvalent value of the array
 *
 * Arguments
 * 	pcArray-> Array holds the integer number
 *
 * Return Value
 * 	Signed integer equivalent value
 */

INT
mide_array_to_int(const CHAR * pcArray)
{
	INT		iCurrent;	/* current CHAR */
	INT		iTotal;	/* current total */
	INT		iSign;		/* if '-', then -ve, otherwise +ve */

	/* skip whitespace */
	while (isspace((INT)(UCHAR) *pcArray)) {
		++pcArray;
	}

	iCurrent = (INT)(UCHAR) *pcArray++;
	iSign = iCurrent;
	if (iCurrent == '-' || iCurrent == '+') {
		/* skip sign */
		iCurrent = (INT)(UCHAR)*pcArray++;
	}

	iTotal = 0;

	while (isdigit(iCurrent)) {
		/*accumulate digit*/
		iTotal = 10 * iTotal +(iCurrent - '0');
		/*get next CHAR*/
		iCurrent = (INT)(UCHAR)*pcArray++;
	}

	if (iSign == '-') {
		return -iTotal;
	} else {
		return iTotal;
	}
}


/**
 * mega_start_ICHxR_poll_thread: Start polling thread for ICHxR
 * This function starts the polling thread that is used for detecting the
 * drive insertion in a ICHxR controller. If the thread is already running
 * function will not try to create the thread.
 *
 * Parameters
 * @pdev	: Driver data structure pointer
 * @return	: VOID
 **/
VOID mega_start_ICHxR_poll_thread(HW_DEVICE_EXTENSION *pdev)
{

	adapter_t	* adapter = (adapter_t *)pdev->adapter;

	/* check if the thread is already running*/
	if (pdev->ICHxR_thread_flag) {
		return;
	}

	/*Initializing completion handler. to detect the thread completion*/
	init_completion(&adapter->ICHxR_completion);

#if LINUX_VERSION_CODE == KERNEL_VERSION(2,4,21)
	spin_unlock_irq(&io_request_lock);
#endif
	/* start monitoring thread */
	switch (pdev->ucCardType) {

	case INTEL_ICH5R:
		pdev->ICHxR_thread_pid = kernel_thread(
			(INT (*)(VOID *))mega_ICH5R_poll_thread, pdev, 0);
		break;
	
	case INTEL_ICH6R:
		pdev->ICHxR_thread_pid = kernel_thread(
			(INT (*)(VOID *))mega_ICH6R_poll_thread, pdev, 0);
		break;
	
	case INTEL_ICH7R:
		pdev->ICHxR_thread_pid = kernel_thread(
			(INT (*)(VOID *))mega_ICH7R_poll_thread, pdev, 0);
		break;
	
	default:
		printk_Rlsi("megaide: unknown ichxr controller!!\n");
		break;
	
	}
#if LINUX_VERSION_CODE == KERNEL_VERSION(2,4,21)
	spin_lock_irq(&io_request_lock);
#endif
}

/**
 * mega_kill_ICHxR_poll_thread: Terminate ICHxR polling thread.
 * This function terminate the polling thread that is used for detecting the
 * drive insertion in a ICHxR controller. Function will actually wait till the
 * thread stop
 * Parameters
 * @pdev	: Driver data structure pointer
 * @return	: VOID
 **/
VOID mega_kill_ICHxR_poll_thread(HW_DEVICE_EXTENSION *pdev)
{
	adapter_t	* adapter = (adapter_t *)pdev->adapter;

	if (pdev->ICHxR_thread_flag) {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
		pdev->ICHxR_thread_flag = 0;
#else
		kill_proc(pdev->ICHxR_thread_pid, SIGKILL, 0);
#endif
		wait_for_completion(&adapter->ICHxR_completion);
	}
}

/* vim: set ts=8 sw=8 tw=78 ai si: */

