/*
 * Linux MegaRAID IDEal Software RAID Stack
 * Copyright (c) 2001 LSI Logic Corporation, All Rights Reserved.
 *-----------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *-----------------------------------------------------------------------------
 * File Name	: megaide_proc.c
 */

#define DRIVER_PROC_NAME	"megaide"
#define ADP_CONFIG_PROC_NAME	"config"
#define ADP_STATUS_PROC_NAME	"status"
#define ADP_REGISTER_PROC_NAME	"registers"
#define PD_PROC_DIR_NAME	"physicaldrives"
#define LD_PROC_DIR_NAME	"logicaldrives"

#define AR_PDRV_SET_FAILED	"set status failed"
#define AR_PDRV_SET_HOTSPARE	"set status hotspare"
#define AR_PDRV_INIT_ERRLOG	"init errorlog"
#define AR_PDRV_GEN_ERROR	"generate error"

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
INT 
megaide_proc_info(struct Scsi_Host *host, CHAR *pcBuffer, CHAR **ppcStart,
	off_t offset, INT iLength, INT iHostNo)
{
	AR_DEBUG(AR_DBG_ENTRY, (KERN_DEBUG "megaide_proc _info Called\n"));
	
	*ppcStart = pcBuffer;
	return 0;
}
#else
INT 
megaide_proc_info(CHAR *pcBuffer, CHAR **ppcStart, off_t offset,
	INT iLength, INT iHostNo, INT iInOut)
{
	AR_DEBUG(AR_DBG_ENTRY, (KERN_DEBUG "megaide_proc _info Called\n"));
	
	*ppcStart = pcBuffer;
	return 0;
}
#endif

#ifdef CONFIG_PROC_FS
extern struct proc_dir_entry proc_root;
static struct proc_dir_entry *megaide_proc_dir_g;

static CHAR *Ldrv_Raid_Str[5] = 
{
	"NONE   ",
	"RAID 0 ",
	"RAID 1 ",
	"RAID 10",
	"NOT PRESENT" 
};

static CHAR *Ldrv_Status_Str[5] = 
{
	"ONLINE  ",
	"CRITICAL",
	"DEGRADED",
	"OFFLINE ",
	"NOT PRESENT" 
};

static CHAR *Yes_No_Str[2] = { "NO ", "YES" };

static CHAR *Pdrv_Id_Str[5] = 
{
	"PRIMARY MASTER  ",
	"PRIMARY SLAVE   ",
	"SECONDARY MASTER",
	"SECONDARY SLAVE ",
	"NOT PRESENT     " 
};

/*
 * This will write the debug informations in to the proc buffer
 * in device extention structure
 */
INT
mega_proc_printf(HW_DEVICE_EXTENSION *pdev, const CHAR *pcFmt, ...)
{
	va_list		args;
	INT		iCount;

	if (pdev->iProcIdx > PROCBUFSIZE){
		return 0;
	}
	
	va_start(args, pcFmt);
	iCount = vsprintf((pdev->achProcBuf + pdev->iProcIdx), pcFmt, args);
	va_end(args);
	pdev->iProcIdx += iCount;

	return iCount;
}


static INT
mega_proc_read_config(CHAR *page, CHAR **start, off_t offset,
	INT iCount, INT *eof, VOID *data)
{
	INT			i;
	INT			count = 0;
	HW_DEVICE_EXTENSION	*pdev = (HW_DEVICE_EXTENSION *)data;
	
	*start = page;

	mega_proc_printf(pdev, "Driver Version         : %d.%02d%c \t\n",
		HYPERDISK_MAJOR_VERSION,
		HYPERDISK_MINOR_VERSION,
		HYPERDISK_BUILD_VERSION + 'a' - 1);
	
	if ((pdev->stFirmwareVersion.usMajorVer == 0) &&
		(pdev->stFirmwareVersion.usMinorVer == 0) &&
		(pdev->stFirmwareVersion.uiBuild == 0)) {
		mega_proc_printf(pdev, "BIOS                   : Disabled\n");
	} else {
		mega_proc_printf(pdev, "BIOS Revision          : %d.%d.%8.8d\n",
			pdev->stFirmwareVersion.usMajorVer,
			pdev->stFirmwareVersion.usMinorVer,
			pdev->stFirmwareVersion.uiBuild);
	}
	
	if(!pdev->stIRCDVersion.usMajorVer) {
		mega_proc_printf(pdev, "IRCD Version           : N/A\n");
	} else {
		mega_proc_printf(pdev, "IRCD Version           : %c.%c \t\n",
			pdev->stIRCDVersion.usMajorVer,
			pdev->stIRCDVersion.usMinorVer);
	}
	
	mega_proc_printf(pdev,
	"Primary Base Address   : %8xH\tPrimary Control Address   : %8xH\n",
		pdev->Channel[0].BaseIoAddress1,
		pdev->Channel[0].BaseIoAddress2);
	
	mega_proc_printf(pdev,
	"Secondary Base Address : %8xH\tSecondary Control Address : %8xH\n",
		pdev->Channel[1].BaseIoAddress1,
		pdev->Channel[1].BaseIoAddress2);
	
	mega_proc_printf(pdev,
	"BusMaster Base Address : %8xH\tIRQ                       : %8d\n",
		pdev->Channel[0].BaseBmAddress, pdev->uiIntLine);
	
	mega_proc_printf(pdev,
	"PCI VendorId           : %8xH\tPCI DeviceId              : %8xH\n",
		pdev->uiVendorId, pdev->uiDeviceId);
	
	mega_proc_printf(pdev,
	"PCI SubSytem VendorId  : %8xH\tPCI SubSytem DeviceId     : %8xH\n",
		pdev->usSubsysVendorId,
		pdev->usSubsysDeviceId);
	
	mega_proc_printf(pdev,
	"PCI Bus Number         : %8d \tPCI Device Number         : %8d\n",
		pdev->ucPCIBus, pdev->ucPCIDev);
	
	mega_proc_printf(pdev, "PCI Function           : %8d \n",
		pdev->ucPCIFun);
	
	mega_proc_printf(pdev, "# Logical Drives       : %8d \t",
		pdev->uiLogicalDriveCount);

	/* Changed the logic to find Number of drives */
	for(i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++)
	{
		if (pdev->PhysicalDrive[i].DeviceFlags&DFLAGS_DEVICE_PRESENT) {
			count++;
		}
	}

	mega_proc_printf(pdev,"# Physical Drives         : %8d\n\n", count);
	
	mega_proc_printf(pdev, "Driver Options:\n");
	
	mega_proc_printf(pdev,
	"Polling Interval       : %8u \tError Threshold          : %8u\n",
		pdev->Options.uiErrorPolling, pdev->Options.uiErrorThreshold);
	
	mega_proc_printf(pdev,
	"Fail Drive             : %8d \tAuto Rebuild             : %8d\n",
		pdev->Options.uiFailDrive, pdev->Options.uiAutoRebuild);
	
	mega_proc_printf(pdev,
	"Check Array            : %8d \tStop Rebuild             : %8d\n",
		pdev->Options.uiCheckArray, pdev->Options.uiStopRebuild);
	
	mega_proc_printf(pdev,
	"Identical              : %8d \tRebuild Rate             : %8d\n",
		pdev->Options.uiIdentical, pdev->Options.uiRebuildRate*10);

	COPY_BACK(pdev, offset, iCount);
	return iCount;
}

static VOID 
pdrv_status_to_string(MP_PHYSICAL_DRIVE_STATUS eStatus, CHAR *cString)
{
	switch (eStatus) {
	
	case PDS_Online :
		sprintf(cString, "ONLINE    ");
		break;
	
	case PDS_Critical :
		sprintf(cString, "CRITICAL  ");
		break;
	
	case PDS_Failed :
		sprintf(cString, "FAILED    ");
		break;
	
	case PDS_HotSpare :
		sprintf(cString, "HOT SPARE ");
		break;
	
	case PDS_Rebuilding :
		sprintf(cString, "REBUILDING");
		break;
	
	case PDS_OffLine :
		sprintf(cString, "OFFLINE   ");
		break;
	
	default :
		sprintf(cString, "UNKNOWN   ");
		break;
	}
}

static INT
mega_proc_read_pdrv_info(CHAR *page, CHAR **start, off_t offset,
	INT iCount, INT *eof, VOID *data)
{
	PHYSICAL_DRIVE_INFO	*pPdrv = (PHYSICAL_DRIVE_INFO *) data;
	HW_DEVICE_EXTENSION	*pdev;
	CHAR		tmode[8];
	CHAR		status[12];
	CHAR		cModelNo[PHYSICAL_DRIVE_MODEL_LENGTH + 1];
	CHAR		cSerialNo[PHYSICAL_DRIVE_SERIAL_NO_LENGTH + 1];
	CHAR		cFirmwareRev[PHYSICAL_DRIVE_FIRMWARE_REV_LENGTH + 1];
	UCHAR		ucIs80PinCable;
	UCHAR		i;
	adapter_t	*adapter;

	*start = page;

	adapter	= (adapter_t *)gastCardInfo[pPdrv->ucControllerId];
	pdev	= &adapter->DeviceExtension;

	if (!(pPdrv->DeviceFlags & DFLAGS_DEVICE_PRESENT)) {
		mega_proc_printf(pdev, "Drive is Not Present\n");
		COPY_BACK(pdev, offset, iCount);
		return iCount;
	}

	ucIs80PinCable = Is80PinCable(pdev, pPdrv->PhysicalDriveId>>1);
	
	pdrv_status_to_string(pPdrv->Status, status);

	switch (pPdrv->TransferMode) {
	case MP_PioMode0 :
	case MP_PioMode1 :
	case MP_PioMode2 :
	case MP_PioMode3 :
	case MP_PioMode4 :
		sprintf(tmode, "PIO(%d)", pPdrv->TransferMode);
		break;

	case MP_DmaMode0 :
	case MP_DmaMode1 :
	case MP_DmaMode2 :
		sprintf(tmode, "DMA(%d)", pPdrv->TransferMode - 5);
		break;

	case MP_UdmaMode0 :
	case MP_UdmaMode1 :
	case MP_UdmaMode2 :
	case MP_UdmaMode3 :
	case MP_UdmaMode4 :
	case MP_UdmaMode5 :
		sprintf(tmode, "UDMA(%d)", pPdrv->TransferMode - 8);
		break;

	default :
		sprintf(tmode, "UNKNOWN");
		break;

	}

	for(i = 0; i < 40; i += 2)
	{
		cModelNo[i] = pPdrv->IdentifyData.ModelNumber[i + 1];
		cModelNo[i + 1] = pPdrv->IdentifyData.ModelNumber[i];
	}
	cModelNo[40] = '\0';

	for(i = 0; i < 20; i += 2)
	{
		cSerialNo[i] = pPdrv->IdentifyData.SerialNumber[i + 1];
		cSerialNo[i + 1] = pPdrv->IdentifyData.SerialNumber[i];
	}
	cSerialNo[20] = '\0';

	for(i = 0; i < 8; i += 2)
	{
		cFirmwareRev[i] = pPdrv->IdentifyData.FirmwareRevision[i + 1];
		cFirmwareRev[i + 1] = pPdrv->IdentifyData.FirmwareRevision[i];
	}
	cFirmwareRev[8] = '\0';

	mega_proc_printf(pdev, "Model No : %s\n", cModelNo);
	mega_proc_printf(pdev, "Serial No : %s\n", cSerialNo);
	mega_proc_printf(pdev, "Firmware Revision : %s\n", cFirmwareRev);
	mega_proc_printf(pdev, "Drive Id : %s\tTransfer Mode : %s\n",
			Pdrv_Id_Str[pPdrv->PhysicalDriveId], tmode);
	mega_proc_printf(pdev, "Status : %s\t\tNo Of Drive Errors : %d\n",
			status, pPdrv->ErrorFound);
	mega_proc_printf(pdev, "Capacity : %d\t\t80 pin Cable : %s\n",
			pPdrv->BiosExposedSectors,
			Yes_No_Str[ucIs80PinCable]);
	mega_proc_printf(pdev, "Smart Compliant : %s\t\tSmart Enabled : %s\n",
			Yes_No_Str[pPdrv->IdentifyData.CmdSupported1 & 1],
			Yes_No_Str[pPdrv->IdentifyData.CmdEnabled1 & 1]);

	COPY_BACK(pdev, offset, iCount);
	return iCount;
}

static INT 
mega_proc_read_registers(CHAR *page, CHAR **start, off_t offset,
	INT iCount, INT *eof, VOID *data)
{
	HW_DEVICE_EXTENSION	*pdev = (HW_DEVICE_EXTENSION *) data;
	UCHAR	ucCfr, ucArttim0, ucArttim1, ucDrwtim0, ucDrwtim1;
	UCHAR	ucArttim2, ucDrwtim2, ucDrwtim3, ucMrdmode;
	UCHAR	ucBmidesr0, ucBmidesr1, ucUdidetcr0, ucUdidetcr1;

	*start = page;

	pci_read_config_byte_Rlsi(pdev, 0x50, &ucCfr);
	pci_read_config_byte_Rlsi(pdev, 0x53, &ucArttim0);
	pci_read_config_byte_Rlsi(pdev, 0x54, &ucDrwtim0);
	pci_read_config_byte_Rlsi(pdev, 0x55, &ucArttim1);
	pci_read_config_byte_Rlsi(pdev, 0x56, &ucDrwtim1);
	pci_read_config_byte_Rlsi(pdev, 0x57, &ucArttim2);
	pci_read_config_byte_Rlsi(pdev, 0x58, &ucDrwtim2);
	pci_read_config_byte_Rlsi(pdev, 0x5b, &ucDrwtim3);
	pci_read_config_byte_Rlsi(pdev, 0x71, &ucMrdmode);
	pci_read_config_byte_Rlsi(pdev, 0x72, &ucBmidesr0);
	pci_read_config_byte_Rlsi(pdev, 0x73, &ucUdidetcr0);
	pci_read_config_byte_Rlsi(pdev, 0x7a, &ucBmidesr1);
	pci_read_config_byte_Rlsi(pdev, 0x7b, &ucUdidetcr1);

	mega_proc_printf(pdev, "PCI Register Information\n");

	mega_proc_printf(pdev,
		"CFR      : %x\tARTTIM0   : %x\tDRWTIM0  : %x\n",
		ucCfr, ucArttim0, ucDrwtim0);
	mega_proc_printf(pdev,
		"ARTTIM1  : %x\tDRWTIM1   : %x\tARTTIM2  : %x\n",
		ucArttim1, ucDrwtim1, ucArttim2);
	mega_proc_printf(pdev,
		"DRWTIM2  : %x\tDRWTIM3   : %x\tMRDMODE  : %x\n",
		ucDrwtim2, ucDrwtim3, ucMrdmode);
	mega_proc_printf(pdev,
		"BMIDESR0 : %x\tUDIDETCR0 : %x\tBMIDESR1 : %x\n",
		ucBmidesr0, ucUdidetcr0, ucBmidesr1);
	mega_proc_printf(pdev, "UDIDETCR1 : %x\n", ucUdidetcr1);

	COPY_BACK(pdev, offset, iCount);
	return iCount;
}

static INT 
mega_proc_read_status(CHAR *page, CHAR **start, off_t offset,
	INT iCount, INT *eof, VOID *data)
{
	HW_DEVICE_EXTENSION	*pdev = (HW_DEVICE_EXTENSION *) data;
	LOGICAL_DRIVE_INFO	*pLdrv = NULL;
	INT		i;
	CHAR		cStatus1[12];
	CHAR		cStatus2[12];
	CHAR		cStatus3[12];
	CHAR		cStatus4[12];

	*start = page;

	mega_proc_printf(pdev, "No Of Pending SCB's in Controller : %d\n",
		pdev->NumberOfPendingScb);
	mega_proc_printf(pdev, "No Of Pending Commands in Controller : %d\n\n",
		pdev->NumberOfPendingCmd);

	mega_proc_printf(pdev, "Logical Drive Information\n");

	for(i = 0; i < pdev->uiLogicalDriveCount; i++)
	{
		pLdrv = &pdev->LogicalDrive[i];
		mega_proc_printf(pdev, "\tStatus : %s\t\t",
		Ldrv_Status_Str [pLdrv->Status]);
		mega_proc_printf(pdev, "Logical Drive No: %d\n", i);
	}

	mega_proc_printf(pdev, "\nPhysical Drive Information\n");
	mega_proc_printf(pdev,
	       "\t\t\t------  PRIMARY  ------\t\t-----  SECONDARY  -----\n");
	mega_proc_printf(pdev, "\t\t\tMASTER\t\tSLAVE\t\tMASTER\t\tSLAVE\n");
	pdrv_status_to_string(pdev->PhysicalDrive[0].Status, cStatus1);
	pdrv_status_to_string(pdev->PhysicalDrive[1].Status, cStatus2);
	pdrv_status_to_string(pdev->PhysicalDrive[2].Status, cStatus3);
	pdrv_status_to_string(pdev->PhysicalDrive[3].Status, cStatus4);
	mega_proc_printf(pdev, "STATUS : \t\t%s\t%s\t%s\t%s\n",
			cStatus1, cStatus2, cStatus3, cStatus4);
	mega_proc_printf(pdev, "No Of Errors : \t\t %8d\t%8d\t%8d\t%8d\n",
			pdev->PhysicalDrive[0].ErrorFound,
			pdev->PhysicalDrive[1].ErrorFound,
			pdev->PhysicalDrive[2].ErrorFound,
			pdev->PhysicalDrive[3].ErrorFound);
	mega_proc_printf(pdev, "No Of Pending Commands : %8d\t%8d\t%8d\t%8d\n",
			pdev->PhysicalDrive[0].CurrentQueueLength,
			pdev->PhysicalDrive[1].CurrentQueueLength,
			pdev->PhysicalDrive[2].CurrentQueueLength,
			pdev->PhysicalDrive[3].CurrentQueueLength);

#ifdef PERFORMANCE_NUMBER
	{
		INT i, len=DBG_TIME_INTERVAL, len1 =1, len2= DBG_TIME_INTERVAL;
	UINT io_2_cmd0=0, io_2_cmd1=0, cmd_2_isr0=0, cmd_2_isr1=0, io_2_isr=0;

	mega_proc_printf(pdev, "Total Commands   = %8d\n", pdev->io_count);
	mega_proc_printf(pdev,
		 "Drive1 Total[%5d]                 Drive2 Total[%5d]\n",
		 pdev->int_count[0], pdev->int_count[1]);
	mega_proc_printf(pdev, "io-2-cmd          cmd-2-isr           \
		io-2-cmd          cmd-2-isr         IO [queue to ISR]\n");

	for (i=0; i<26; i++)
	{
		mega_proc_printf(pdev, "[%3d-%3d] =%5d  [%3d-%3d] =%5d    \
			[%3d-%3d] =%5d  [%3d-%3d] =%5d  [%3d-%3d] =%5d\n",
			(i*len), ((i+1)*len), pdev->io_2_pcmd[0][i], (i*len1),
			((i+1)*len1), pdev->pcmd_2_isr[0][i],
			(i*len), ((i+1)*len), pdev->io_2_pcmd[1][i], (i*len1),
			((i+1)*len1), pdev->pcmd_2_isr[1][i],
			(i*len2), ((i+1)*len2), pdev->scbduration[i]);
		
		io_2_cmd0 += (pdev->io_2_pcmd[0][i] *
			((i*len)+((i+1)*len))) / 2;
		
		io_2_cmd1 += (pdev->io_2_pcmd[1][i] *
			((i*len)+((i+1)*len))) / 2;
		
		cmd_2_isr0 += (pdev->pcmd_2_isr[0][i] *
			((i*len1)+((i+1)*len1))) /2;
		
		cmd_2_isr1 += (pdev->pcmd_2_isr[1][i] *
			((i*len1)+((i+1)*len1))) /2;
		
		io_2_isr += (pdev->scbduration[i]*
			((i*len2)+((i+1)*len2)))/2;
	}

	io_2_cmd0	= (io_2_cmd0/(pdev->int_count[0]+1));
	cmd_2_isr0	= (cmd_2_isr0 /(pdev->int_count[0]+1));
	io_2_cmd1	= (io_2_cmd1 /(pdev->int_count[1]+1));
	cmd_2_isr1	= (cmd_2_isr1 /(pdev->int_count[1]+1));
	io_2_isr	= (io_2_isr / (pdev->io_count+1));

	mega_proc_printf(pdev, "Average-->   %3d               \
		%3d                 %3d               %3d               %3d\n",
		io_2_cmd0, cmd_2_isr0, io_2_cmd1, cmd_2_isr1, io_2_isr);
}
#endif /* PERFORMANCE_NUMBER */
	
	COPY_BACK(pdev, offset, iCount);
	return iCount;
}

static INT
mega_proc_read_ldrv_info(CHAR *page, CHAR **start, off_t offset,
	INT iCount, INT *eof, VOID *data)
{
	LOGICAL_DRIVE_INFO	*pLdrv = (LOGICAL_DRIVE_INFO *) data;
	HW_DEVICE_EXTENSION	*pdev;
	adapter_t		*adapter;

	*start = page;

	adapter	= (adapter_t *)gastCardInfo[pLdrv->ucControllerId];
	pdev	= &adapter->DeviceExtension;
	
	if (pLdrv->Status == LDS_NotPresent) {
		mega_proc_printf(pdev, "This logical drive is not present\n");
	} else {
		mega_proc_printf(pdev, "RAID Level : %s\t\tStatus : %s\n",
			Ldrv_Raid_Str[pLdrv->RaidLevel],
			Ldrv_Status_Str[pLdrv->Status]);
		if (pLdrv->RaidLevel == Raid1) {
			mega_proc_printf(pdev, 
				"Sectors : %d\t\tStripe Size : N/A\n",
				pLdrv->Sectors);
		} else {
			mega_proc_printf(pdev, 
				"Sectors : %d\t\tStripe Size : %dKB\n",
				pLdrv->Sectors, pLdrv->StripeSize/2);
		}
		mega_proc_printf(pdev, 
			"Stripes Per Row : %d\t\tPhysical Drive Count : %d\n",
			pLdrv->StripesPerRow, pLdrv->PhysicalDriveCount);
	}
	COPY_BACK(pdev, offset, iCount);
	return iCount;
}

static VOID
mraid_create_ld_proc(adapter_t *adapter)
{
	UCHAR			name_str[30] = { 0 };
	INT			i;
	HW_DEVICE_EXTENSION	*pdev = 
		(HW_DEVICE_EXTENSION *) &adapter->DeviceExtension;
	
	AR_DEBUG(AR_DBG_PROC, (KERN_DEBUG 
		"megaide: creating pd root proc directory\n"));
	
	adapter->ld_dir_proc_entry =     
		proc_mkdir(LD_PROC_DIR_NAME, adapter->adp_proc_entry);
	
	if (!adapter->ld_dir_proc_entry){
		printk_Rlsi(KERN_WARNING 
			"megaide: ld root proc dir creation failed\n");
		return;
	}
	
	for(i = 0; i < MAX_LOGICAL_DRIVES; i++)
	{
		sprintf(name_str, "log_drv_%d_info", i);
		create_proc_read_entry (name_str, S_IRUSR|S_IFREG, 
			adapter->ld_dir_proc_entry, mega_proc_read_ldrv_info, 
			&pdev->LogicalDrive[i]);
	} 
}

static VOID
mraid_remove_ld_proc(adapter_t *adapter)
{
	UCHAR			name_str[30] = { 0 };
	INT			i;
	
	for(i = 0; i < MAX_LOGICAL_DRIVES; i++)
	{
		sprintf(name_str, "log_drv_%d_info", i);
		
		remove_proc_entry(name_str, adapter->ld_dir_proc_entry);
	}
	
	if (adapter->ld_dir_proc_entry) {
		
		AR_DEBUG(AR_DBG_PROC, (KERN_DEBUG 
			"megaide: removing ld root proc directory\n"));
		
		remove_proc_entry(LD_PROC_DIR_NAME, adapter->adp_proc_entry);
	}
}

static VOID
mraid_create_pd_proc(adapter_t *adapter)
{
	UCHAR			name_str[30] = { 0 };
	INT			i;
	HW_DEVICE_EXTENSION	*pdev = 
		(HW_DEVICE_EXTENSION *) &adapter->DeviceExtension;
	
	AR_DEBUG(AR_DBG_PROC, (KERN_DEBUG 
		"megaide: creating root pd proc directory\n"));
	
	adapter->pd_dir_proc_entry =  
		proc_mkdir(PD_PROC_DIR_NAME, adapter->adp_proc_entry);
		
	if (!adapter->pd_dir_proc_entry) {
		printk_Rlsi(KERN_WARNING 
			"megaide: pd root proc dir creation failed\n");
		return;
	} 

	for(i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++)
	{
		sprintf(name_str, "phy_drv_%d_info", i);
		create_proc_read_entry (name_str, S_IRUSR|S_IFREG, 
			adapter->pd_dir_proc_entry, mega_proc_read_pdrv_info, 
			&pdev->PhysicalDrive[i]);
	}
}

static VOID
mraid_remove_pd_proc(adapter_t *adapter)
{
	UCHAR			name_str[30] = { 0 };
	INT			i;
	
	for(i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++)
	{
		AR_DEBUG(AR_DBG_PROC, (KERN_DEBUG 
			"megaide: removing pd [%d] proc dir\n", i));
		sprintf(name_str, "phy_drv_%d_info", i);
		remove_proc_entry(name_str, adapter->pd_dir_proc_entry);
	}
	
	if (adapter->pd_dir_proc_entry) {
		
		AR_DEBUG(AR_DBG_PROC, (KERN_DEBUG 
			"megaide: removing root pd proc directory\n"));
		
		remove_proc_entry(PD_PROC_DIR_NAME,	
			adapter->adp_proc_entry);	
	}
}


static VOID 
mraid_create_adp_proc(adapter_t *adapter)
{
	UCHAR			name_str[12] = { 0 };

	sprintf(name_str, "%d", adapter->DeviceExtension.ucControllerId);
	
	AR_DEBUG(AR_DBG_PROC, (KERN_DEBUG 
		"megaide: creating adapter[%d] proc\n",
		adapter->DeviceExtension.ucControllerId));

	adapter->adp_proc_entry = proc_mkdir(name_str, megaide_proc_dir_g);

	if (!adapter->adp_proc_entry) {
		printk_Rlsi(KERN_WARNING "megaide: proc_mkdir failed\n");
	} else {
		create_proc_read_entry (ADP_CONFIG_PROC_NAME,
			(S_IRUSR|S_IFREG), adapter->adp_proc_entry, 
			mega_proc_read_config, &adapter->DeviceExtension);
		
		create_proc_read_entry (ADP_STATUS_PROC_NAME,
			(S_IRUSR|S_IFREG), adapter->adp_proc_entry, 
			mega_proc_read_status, &adapter->DeviceExtension);
		
		create_proc_read_entry (ADP_REGISTER_PROC_NAME,
			(S_IRUSR|S_IFREG), adapter->adp_proc_entry, 
			mega_proc_read_registers, &adapter->DeviceExtension);
		
		mraid_create_pd_proc(adapter);
		mraid_create_ld_proc(adapter);
	}

}

static VOID 
mraid_remove_adp_proc(adapter_t *adapter)
{
	CHAR		name_str[12] = { 0 };
	
	sprintf(name_str, "%d", adapter->DeviceExtension.ucControllerId);
	
	AR_DEBUG(AR_DBG_PROC, (KERN_DEBUG 
		"megaide: removing adapter[%d] proc\n",
		adapter->DeviceExtension.ucControllerId));
	
	mraid_remove_ld_proc(adapter);
	mraid_remove_pd_proc(adapter);
	
	/* removing all adapter related proc informations */
	remove_proc_entry (ADP_CONFIG_PROC_NAME, adapter->adp_proc_entry);
	remove_proc_entry (ADP_STATUS_PROC_NAME, adapter->adp_proc_entry);
	remove_proc_entry (ADP_REGISTER_PROC_NAME, adapter->adp_proc_entry);
	
	remove_proc_entry(name_str, megaide_proc_dir_g);
	
	adapter->adp_proc_entry = NULL;
}

VOID
megaideproc_attach(adapter_t *adapter)
{
	if (!megaide_proc_dir_g) {
		AR_DEBUG(AR_DBG_PROC, (KERN_DEBUG 
			"megaide: creating driver proc\n"));
		megaide_proc_dir_g = proc_mkdir(DRIVER_PROC_NAME, &proc_root);
	}

	if (megaide_proc_dir_g) {
		mraid_create_adp_proc(adapter);
	}

}

VOID
megaideproc_detach(adapter_t *adapter)
{
	if(adapter->adp_proc_entry ) {
		mraid_remove_adp_proc(adapter);
	}
	
	/* remove driver proc entry if all adapters removed */
	if (giNumberOfCtrlrs == 1) {
		AR_DEBUG(AR_DBG_PROC, (KERN_DEBUG 
			"megaide: removing driver proc\n"));
		remove_proc_entry(DRIVER_PROC_NAME, &proc_root);
	}
}

#endif	/* CONFIG_PROC_FS */

/* vim: set ts=8 sw=8 tw=78 ai si: */

