/*
 * Linux MegaRAID IDEal Software RAID Stack
 * Copyright (c) 2001 LSI Logic Corporation, All Rights Reserved.
 *-----------------------------------------------------------------------------
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *-----------------------------------------------------------------------------
 *  File Name	: mp_hd_ioctls.h
 */

#ifndef __HD_IOCTLS_H__
#define __HD_IOCTLS_H__

/* macros Definition */
#define MAX_LOGICAL_DRIVES        8
#define MAX_DRIVES_PER_CHANNEL    2

#define INVALID_CONNECTION_ID  0xff
#define INVALID_DRIVE_ID       0xff
#define INVALID_ARRAY_ID       0xff
#define INVALID_CHANNEL_ID     0xff

#define PHYSICAL_DRIVE_MODEL_LENGTH        80
#define PHYSICAL_DRIVE_SERIAL_NO_LENGTH    40
#define PHYSICAL_DRIVE_FIRMWARE_REV_LENGTH 10

/**
 * IRCD3.x organization
 *
 * Organization of Reserved 2 MB Space at the end of the drive in IRCD3x
 *
 * -----------------------4096	: 2MB
 * Configuration Back UP	: 64 Sectors (32KB)
 * -----------------------4032
 * Diagnostic Block		: 512 Sectors (256KB)
 * -----------------------3520
 * Reserved			: 2944 Sectors (1472KB)
 * ------------------------576
 * Bad Block Management Block	: 512 sectors (256kb)
 * -------------------------64
 * Reserved for future use
 * of config information	: 64 Sectors (32kb)
 * --------------------------4
 * Enquiry			: 2 Sectors (1kb)
 * --------------------------2
 * Error Log			: 1 Sector
 * --------------------------1
 * Actual configuration		: 1 Sector
 * --------------------------0
 * END OF DRIVE
 */

#define IDE_RESERVED_SECTORS      4096
#define DIAG_TOTAL_SECTORS        0xFF	/* 256 */
#define DIAG_OFFSET_FROM_END    (3936)	/* 3 MB */

#define MEGAIDE_IOCTL          0x80
#define MEGAIDE_IOCTL_NEW      0xE0

/* opcodes */
#define AR_PROBE_AMIRAID               1
#define AR_RAID_INFO                   2
#define AR_SET_OPTIONS                 3
#define AR_GET_ERROR_LOG               4
#define AR_READ_BLK_CMD                5
#define AR_WRITE_BLK_CMD               6
#define AR_GET_EVENT                   7
#define AR_SEND_ACTIVITY               8
#define AR_GET_ACTIVITY                9
#define AR_GET_ACTIVITY_EX            10
#define AR_SEND_ACTIVITY_EX           11
#define AR_RAID_STATUS_INFO_EX        14
#define AR_GET_PHYSICAL_DRIVE_INFO    15
#define AR_GET_VERSION_EX             16
#define AR_GET_CONTROLLER_INFO        18
#define AR_GET_SPAREPOOL_INFO         19
#define AR_GET_FAILDRIVE_INFO         20
#define AR_GET_VERSION                21
#define AR_EXECUTE_SMART_COMMAND      22
#define AR_CLEAR_ERROR_LOG            40
#define AR_RAID_STATUS_INFO           41
#define AR_FAIL_DRIVE                 42
#define AR_RESET_CONTROLLER           43
#define AR_INIT_ERRORLOG              44
#define AR_LOG_ERROR                  45
#define AR_MAKE_HOTSPARE              46
#define AR_EXECUTE_PASS_THRU_COMMAND  47
#define AR_EXECUTE_IDENTIFY_COMMAND   48
#define AR_EXECUTE_FEATURES_COMMAND   49
#define AR_GET_GEOMETRY_COMMAND       50
#define AR_INTERNAL_PASS_THRU_COMMAND 51
#define AR_EXECUTE_IRCD_COMMAND       52
#define AR_ERROR_RECOVERY             53
#define AR_DIAG_PASSTHRU              54
#define AR_SET_DRIVE_STATUS           55
#define AR_INJECT_CLEAR_ERROR         56

#define AR_GET_DRIVER_CAPABILITY      57

#ifdef _SOFT_BBM_
#define AR_GET_SBBM_LIST              58
/* do not use only for driver testing */
#define AR_SET_SBBM_LIST              59
#endif /*_SOFT_BBM_*/

/* ioctls for backplane */
#define AR_BACKPLANE_ACTIVITY        100
/* 101- 120 used for LED activity IOCTLS */
#define AR_DETECT_DRIVE              120
#define AR_REMOVE_DRIVE              121

#define AR_READ_WRITE_IRCD_INFO      125
#define AR_ENQUIRY_INFO              126
/* for flash ioctl support */
#define AR_READ_WRITE_FLASH_MEM      127

/* these IOCTLS are for driver testing purpose */
/* Applications should not use it. */
#define AR_CHECK_FOR_CONFIG_UPDATE      150
#define AR_DBG_RESET_PERFORMANCE_COUNTS 151

#ifndef MAX_COMMAND_SIZE
 #define  MAX_COMMAND_SIZE       12
#endif

/* return_code */
#define  MP_IOCTL_SUCCESS             0
#define  MP_IOCTL_FAIL                1
#define  MP_IOCTL_TIMEOUT             2
#define  MP_IOCTL_PENDING             4
#define  MP_IOCTL_DATALEN_MISMATCH    5
#define  MP_IOCTL_BAD_OPCODE          6
#define  MP_IOCTL_PARAM_MISMATCH      7
#define  MP_IOCTL_SMART_STATUS_OK     8
#define  MP_IOCTL_SMART_STATUS_BAD    9
#define  MP_IOCTL_DATA_LOCKED        10
#define  MP_IOCTL_POWER_CONNECTED    11
#define  MP_IOCTL_NO_POWER_CONNECTED 12
#define  MP_IOCTL_NO_48_BIT_SUPPORT  13

/* even more return codes */
#define IOCTL_OK     0
#define IOCTL_ERROR  1
#define IOCTL_QUEUE  2

#define NO_HOLD          0
#define UPDATE_IRCD      1
#define UPDATE_ERRORLOG  2
#define UPDATE_ENQUIRY   3
#define UPDATE_ENQ_IRCD  4

#define  BOTH_CHANNELS  0x11
#define FIRST_CHANNEL   0x10
#define SECOND_CHANNEL  0x01

typedef struct _ioctl_header {
	UINT  header_len;  /* length of this header */
	UINT  timeout;     /* timeout for this command (1 - no timeout) */
	UINT  opcode;      /* requested opcode */
	UINT  return_code; /* return code */
	UINT  data_len;    /* length of data area */
} ioctl_header;

/* Driver will handle copying the data_buffer[] to and from user space */
typedef struct _ioctl_data {
	ioctl_header  sic; /* header */
	UCHAR data_buffer[(4*1024) - sizeof(ioctl_header)];
	/* ptr to data body */
}
__attribute__ ((packed))
 ioctl_data, *pioctl_data;

/* this is the structure passed to scsi mid layer */
typedef struct _megaide_ioctl_in_new{
	UINT  inlen;		/* length of data to be written to device */
	UINT  outlen;		/* length of data to be read from device */
	UCHAR  cmnd[10];	/* SCSI Command (copied to scsi_cmnd->cmnd) */
	ioctl_data  wdata;	/* Data written to device starts here */
				/* (coped to scsi_cmnd->request_buffer) */
}
__attribute__ ((packed))
megaide_ioctl_in_new , *pmegaide_ioctl_in_new ;

typedef struct _megaide_ioctl_in {
	UINT  inlen;		/* length of data to be written to device */
	UINT  outlen;		/* length of data to be read from device */
	UCHAR  cmnd[12];	/* SCSI Command (copied to scsi_cmnd->cmnd) */
	ioctl_data  wdata;	/* Data written to device starts here */
				/* (coped to scsi_cmnd->request_buffer) */
}
__attribute__ ((packed))
megaide_ioctl_in , *pmegaide_ioctl_in ;

/* this is the structure returned from scsi mid layer */
typedef struct _megaide_ioctl_out {
	UINT  inlen;		/* length of data to be written to device */
	UINT  outlen;		/* length of data to be read from device */
	ioctl_data  wdata;	/* Data written to device starts here */
				/* (coped to scsi_cmnd->request_buffer) */
}
__attribute__ ((packed))
megaide_ioctl_out, *pmegaide_ioctl_out;

/* Diver will handle copying the data_buffer[] to and from user space */
typedef struct _char_ioctl_data {
	ioctl_header   sic;      /* IOCTL header */
	UCHAR   data_buffer[1]; /* ptr to data body , driver will */
}
__attribute__ ((packed))
char_ioctl_data, *pchar_ioctl_data;

/* New IOCTL Implementaion it does not requre cmnd field */
typedef struct _char_megaide_ioctl {
	UINT  inlen;         /* length of data to be written to device */
	UINT  outlen;        /* length of data to be read from device */
	INT    iAdapNumber;    /* Adapter number starts From 1 */
	CHAR    acSignature[7];/* Must contain "MEGANIT" without ending '\0' */
	CHAR    acReserved[23];
	char_ioctl_data  wdata;/* Data written to device starts here */
				/* (coped to scsi_cmnd->request_buffer) */
}
__attribute__ ((packed))
char_megaide_ioctl, *pchar_megaide_ioctl;

#define MEGAIDE_IOC_MAGIC  'm'
/* hard coding the value of following macro to suppor 32+64 bit ioclts */
/* _IOW(MEGAIDE_IOC_MAGIC, 1, sizeof(char_megaide_ioctl)) */
#define MEGAIDE_DEV_IOCTL 0x40046D01

#define NIT_MAX_DATA_TRANSFER  (128*1024)
#define CHAR_HEADER_SIZE \
	(sizeof(char_megaide_ioctl) - sizeof(char_ioctl_data))

/* Drive ID */
typedef enum {
	PrimaryMaster = 0,
	PrimarySlave = 1,
	SecondaryMaster = 2,
	SecondarySlave = 3,
	InvalidDrive = INVALID_DRIVE_ID
}
MP_DRIVE_ID;

/* RAID levels. */
typedef enum {
	None = 0x00,
	Raid0 = 0x01,	/* stripe */
	Raid1 = 0x02,	/* mirror */
	Raid10 = 0x03,	/* stripe+mirror */
	SpareDrivePool = 0x08
} MP_RAID_LEVEL;

/* RAID Array status */
typedef enum {
	LDS_Online = 0,
	LDS_Critical = 1,
	LDS_Degraded = 2,
	LDS_OffLine = 3,
	LDS_Deleted = 4,
	LDS_NotPresent = 0xff
} MP_LOGICAL_DRIVE_STATUS;

/* RAID Physical Drive status. */
typedef enum {
	PDS_Online = 0,
	PDS_Critical = 1,
	PDS_Failed = 2,
	PDS_HotSpare = 0xd,
	PDS_Rebuilding = 0xe,
	PDS_OffLine = 0xf,
	PDS_Undefined = 0xff
} MP_PHYSICAL_DRIVE_STATUS;

/* Define transfer modes */
typedef enum {
	MP_PioMode0 = 0,
	MP_PioMode1 = 1,
	MP_PioMode2 = 2,
	MP_PioMode3 = 3,
	MP_PioMode4 = 4,
	MP_DmaMode0 = 5,
	MP_DmaMode1 = 6,
	MP_DmaMode2 = 7,
	MP_UdmaMode0 = 8,
	MP_UdmaMode1 = 9,
	MP_UdmaMode2 = 10,
	MP_UdmaMode3 = 11,
	MP_UdmaMode4 = 12,
	MP_UdmaMode5 = 13,
	MP_UdmaMode6 = 14
} MP_TRANSFER_MODE, *PMP_TRANSFER_MODE;

/* Physical Drive Information Structure */
typedef struct _MP_RAID_INFO_PD {
	MP_DRIVE_ID  DriveID;
	UCHAR  caModelInfo[PHYSICAL_DRIVE_MODEL_LENGTH + 1];
	UCHAR  caSerialNumber[PHYSICAL_DRIVE_SERIAL_NO_LENGTH + 1];
	UCHAR  caFirmwareRevision[PHYSICAL_DRIVE_FIRMWARE_REV_LENGTH + 1];
	MP_TRANSFER_MODE  TransferMode;
	MP_PHYSICAL_DRIVE_STATUS  PhyStatus;
	UINT  uiPhySize;
	UCHAR  ucIsPhyDrvPresent;
	UCHAR  ucIsPowerConnected;
	UCHAR  ucIs80PinCable;
	UCHAR  ucSMART_Compliant;
	UCHAR  ucSMART_Enabled;
}
__attribute__ ((packed))
MP_RAID_INFO_PD, *PMP_RAID_INFO_PD;

/* Yet Another Physical Drive Info IOCTL */
typedef struct _IDE_PHYSICAL_DRIVE_INFO {
	UCHAR  ucNumPhysicalDrives;
	MP_RAID_INFO_PD  PhysicalDrives[MAX_DRIVES_PER_CONTROLLER];
}
__attribute__ ((packed))
IDE_PHYSICAL_DRIVE_INFO, * PIDE_PHYSICAL_DRIVE_INFO;

/* Logical Drive Information */
typedef struct  _MP_RAID_INFO_LD {
	MP_RAID_LEVEL  Mode;
	MP_LOGICAL_DRIVE_STATUS  Status;
	UINT  uiStripeSize;	/* Stripe Size in Sector Count */
	UINT  uiStripesPerRow;
	UINT  uiTotDriveCnt;
	UINT  uiCapacity;	/* Drive Size */
	UINT  uiArrayId;
	MP_RAID_INFO_PD   phyDrives[MAX_DRIVES_PER_CONTROLLER];
}
__attribute__ ((packed))
MP_RAID_INFO_LD, *PMP_RAID_INFO_LD;

/* Raid Information */
typedef struct _MP_IDE_RAID_INFO {
	UCHAR  ucNumLogicalDrives;
	UCHAR  ucMajorVersion;
	UCHAR  ucMinorVersion;
	MP_RAID_INFO_LD  LogDrives[MAX_DRIVES_PER_CONTROLLER];
}
__attribute__ ((packed))
MP_IDE_RAID_INFO, *PMP_IDE_RAID_INFO;

typedef struct _MP_SPAREPOOL_DATA {
	UINT  uiNumberOfDrives;
	MP_RAID_INFO_PD  Info[MAX_DRIVES_PER_CONTROLLER];
}
__attribute__ ((packed))
MP_SPAREPOOL_DATA, *PMP_SPAREPOOL_DATA;

typedef struct _MP_CHIPSET_INFO{
	USHORT  usVendorID;
	USHORT  usDeviceID;
	UCHAR  ucPciBus;
	UCHAR  ucPciDevice;
	UCHAR  ucPciFunction;
}
__attribute__ ((packed))
MP_CHIPSET_INFO;

typedef struct _MP_IDE_VERSION {
	USHORT  usMajorVer;
	USHORT  usMinorVer;
	UINT  uiBuild;
}
__attribute__ ((packed))
MP_IDE_VERSION, *PMP_IDE_VERSION;

typedef enum {
	VERSION_IRCD = 0,
	VERSION_BIOS = 1,
	VERSION_DRIVER = 2
}VERSION_INFO;

typedef struct _MP_IDE_VERSION_EX {
	VERSION_INFO  version;
	USHORT  usMajorVer;
	USHORT  usMinorVer;
	UINT  uiBuild;
}
__attribute__ ((packed))
MP_IDE_VERSION_EX, *PMP_IDE_VERSION_EX;

typedef struct _MP_PROBE_AMIRAID {
	UCHAR  acData[8];
	USHORT  usControllerId;	/* ZERO based-based on pci detection */
}
__attribute__ ((packed))
MP_PROBE_AMIRAID, *PMP_PROBE_AMIRAID;

typedef struct _MP_IDE_CONTROLLER_INFO{
	USHORT  usControllerId;		/* ZERO based-based on pci detection */
	UINT  uiPrimaryBaseAddress;	/* 1st range primary */
	UINT  uiPrimaryControlAddress;	/* 2nd range */
	UINT  uiSecondaryBaseAddress;	/* 1st range primary */
	UINT  uiSecondaryControllAddress;/* 2nd range */
	UINT  uiBusMasterBaseAddress;
	USHORT  usIRQ;			/* IRQ */
	MP_IDE_VERSION  FwVersion;	/* version of hyperdsk card */
	MP_CHIPSET_INFO ChipsetInfo;	/* controller chipset */
}
__attribute__ ((packed))
MP_IDE_CONTROLLER_INFO, *PMP_IDE_CONTROLLER_INFO;

typedef struct{
	UINT  uiControllerCount;
	MP_IDE_CONTROLLER_INFO  Info[MAX_CONTROLLERS];
}
__attribute__ ((packed))
MP_CONTROLLER_DATA,*PMP_CONTROLLER_DATA;

typedef struct _MP_RAID_OPTIONS {
	UINT  uiErrorPolling;
	UINT  uiErrorThreshold;
	UINT  uiFailDrive;
	UINT  uiAutoRebuild;
	UINT  uiCheckArray;
	UINT  uiStopRebuild;
	UINT  uiIdentical;
	UINT  uiRebuildRate;
}
__attribute__ ((packed))
MP_RAID_OPTIONS, * PMP_RAID_OPTIONS;

typedef struct _MP_RAID_OPTIONS_EX {
	UINT  uiErrorPolling;
	UINT  uiErrorThreshold;
	UINT  uiFailDrive;
	UINT  uiAutoRebuild;
	UINT  uiCheckArray;
	UINT  uiStopRebuild;
	UINT  uiIdentical;
	UINT  uiRebuildRate;
	UINT  uiDiskCoercion;
	UINT  uiAutoResume;
	UINT  uiLastActiveOperation[MAX_LOGICAL_DRIVES];
	UINT  uiLastSector[MAX_LOGICAL_DRIVES];
	UINT  uiReserved[20];
}
__attribute__ ((packed))
MP_RAID_OPTIONS_EX, * PMP_RAID_OPTIONS_EX;

typedef enum {
	MP_OP_NOP  = 0,
	MP_OP_REBUILD  = 1,
	MP_OP_CHKCON  = 2,
	MP_OP_MKCON  = 3
}
MP_BACKGROUND_ACTIVITY;

typedef struct _MP_RAID_STATUS_INFO {
	MP_LOGICAL_DRIVE_STATUS  ArrayStatus[MAX_DRIVES_PER_CONTROLLER];
	MP_PHYSICAL_DRIVE_STATUS  PhyStatus[MAX_DRIVES_PER_CONTROLLER];
	BOOLEAN  CheckErrorLog[MAX_DRIVES_PER_CONTROLLER];
	MP_BACKGROUND_ACTIVITY CurrentArrayActivity[MAX_DRIVES_PER_CONTROLLER];
}
__attribute__ ((packed))
MP_RAID_STATUS_INFO, * PMP_RAID_STATUS_INFO;

/* if BBM List is changed in the driver it shuld be notified 
 * to the application with a new uiStatus IDERAID_STATUS_FLAG_NEW_BBM_LIST
 */
#define IDERAID_STATUS_FLAG_NEW_BBM_LIST       0x10

typedef struct _MP_RAID_STATUS_INFO_EX {
	USHORT    usSize;
	MP_LOGICAL_DRIVE_STATUS  ArrayStatus[MAX_LOGICAL_DRIVES];
	MP_PHYSICAL_DRIVE_STATUS  PhyStatus[MAX_DRIVES_PER_CONTROLLER];
	BOOLEAN  CheckErrorLog [MAX_DRIVES_PER_CONTROLLER];
	MP_BACKGROUND_ACTIVITY  CurrentArrayActivity[MAX_LOGICAL_DRIVES];
	UCHAR  PowerConnected;
#ifdef _SOFT_BBM_
        UINT uiStatus;
        UCHAR ulReserved[15];
#else	
	UCHAR  ucReserved[19];
#endif /*_SOFT_BBM_*/
}
__attribute__ ((packed))
MP_RAID_STATUS_INFO_EX, * PMP_RAID_STATUS_INFO_EX;

typedef enum {
	START_REBUILDING = 0,
	CONSISTENCY_CHECK = 1,
	MAKE_CONSISTENT = 2,
	ABORT_COMMAND = 3,
	DO_NEXT_BLOCK = 4
} MP_ACTION_TYPES;

typedef enum {
	ACTIVITY_NOP = 0,
	ACTIVITY_SUCCEED = 1,
	ACTIVITY_ABORTED = 2
} MP_ACTIVITY_STATUS;

typedef struct _MP_SEND_ACTIVITY {
	MP_ACTION_TYPES    Action;
	UINT  uiBlockCount;
	UINT  uiErrorCount;
	MP_DRIVE_ID  SourceDriveId;
	MP_DRIVE_ID  TargetDriveId;
}
__attribute__ ((packed))
MP_SEND_ACTIVITY , * PMP_SEND_ACTIVITY ;

typedef struct _MP_SEND_ACTIVITY_EX {
	MP_ACTION_TYPES Action;
	UINT  uiBlockCount;
	UINT  uiErrorCount;
	MP_DRIVE_ID  SourceDriveId;
	MP_DRIVE_ID  TargetDriveId;
	UCHAR  ucLogicalDriveId;
	UCHAR  ucReserved[18];
}
__attribute__ ((packed))
MP_SEND_ACTIVITY_EX , * PMP_SEND_ACTIVITY_EX ;

typedef struct _MP_GET_ACTIVITY {
	UINT  uiLogicalDriveId;
	MP_BACKGROUND_ACTIVITY  Action;
	UINT  uiTotBlockCount;
	UINT  uiDoneBlockCount;
	UINT  uiInconsistencyCount;
	MP_DRIVE_ID  SourceDriveId;
	MP_DRIVE_ID  TargetDriveId;
	UINT  StartTime;
}
__attribute__ ((packed))
MP_GET_ACTIVITY , * PMP_GET_ACTIVITY ;

typedef struct _MP_GET_ACTIVITY_EX {
	UINT  uiLogicalDriveId;
	MP_BACKGROUND_ACTIVITY  Action;
	UINT  uiTotBlockCount;
	UINT  uiDoneBlockCount;
	UINT  uiInconsistencyCount;
	MP_DRIVE_ID  SourceDriveId;
	MP_DRIVE_ID  TargetDriveId;
	UINT  StartTime;
	UINT  uiErrorCount;
	MP_ACTIVITY_STATUS  LastStatus;
	MP_BACKGROUND_ACTIVITY  LastAction;
	UCHAR  ucReserved[4];
}
__attribute__ ((packed))
MP_GET_ACTIVITY_EX , * PMP_GET_ACTIVITY_EX ;

typedef struct _MP_FAIL_DRIVE {
	MP_DRIVE_ID  DriveId;
}
__attribute__ ((packed))
MP_FAIL_DRIVE , * PMP_FAIL_DRIVE;

typedef struct _MP_DETECT_DRIVE {
	MP_DRIVE_ID  DriveId;
	UCHAR  ucVersionNumber; /* not used now */
	UCHAR  ucReserved[32];
}
__attribute__ ((packed))
MP_DETECT_DRIVE, *PMP_DETECT_DRIVE;

typedef struct _MP_SET_PDRV_STATUS{
	MP_DRIVE_ID  DriveId;
	MP_PHYSICAL_DRIVE_STATUS  DriveStatus;
}MP_SET_PDRV_STATUS,*PMP_SET_PDRV_STATUS;

typedef struct _PASS_THRU_DATA {
	UCHAR  uchTargetID; /* PM=0, PS=1, SM=2, SS=3 */
	UINT  uiDataSize;      /* size of data buffer */
	BOOLEAN  bIsPIO;    /* 0 = DMA Transfer , 1 = PIO Transfer */
	BOOLEAN  bIsIn;     /* 1 = reading from dev, 0 = writing to dev */
	UCHAR  aucVersionNumber; /* Version of this IOCTL */
	UCHAR  aucReserved[7];
	UCHAR  aucTaskFile[8];   /* IDE Task File Registers */
	UCHAR  ucTaskFileBitMap; /* bitmask of registers to be written to */
	UCHAR  ucStatusByte;
	UCHAR  ucErrorByte;     /* error register value */
	UCHAR  aucReserved2[32];
	UCHAR  aucBuffer[512];  /* data buffer */
}
__attribute__ ((packed))
PASS_THRU_DATA, * PPASS_THRU_DATA;

typedef struct _PASS_THRU_DATA_EX {
	UCHAR  uchTargetID;          /* PM=0, PS=1, SM=2, SS=3 */
	UINT  uiDataSize;               /* size of data buffer */
	BOOLEAN  bIsPIO;             /* 0 = DMA Transfer , 1 = PIO Transfer */
	BOOLEAN  bIsIn;              /* 1 = reading from, 0 = writing to dev*/
	UCHAR  aucVersionNumber;     /* Version 2 means using this structure*/
	UCHAR  aucReserved[7];
	UCHAR  aucTaskFile[8];       /* IDE Task File Registers */
	UCHAR  ucTaskFileBitMap;     /* bitmask of registers to be written */
	UCHAR  ucStatusByte;
	UCHAR  ucErrorByte;          /* error register value */
	UCHAR  aucTaskFile48bit[8];  /* this is to hold 48 bit address */
	UCHAR  ucTaskFileBitMap48bit;/* bitmask of registers for 48 bit */
	UCHAR  aucReserved2[23];
	UCHAR  aucBuffer[1];         /* buffer User need to allocate memory */
}
__attribute__ ((packed))
PASS_THRU_DATA_EX, * PPASS_THRU_DATA_EX;


/* Diagnostic SubCodes */
#define DIAG_WRITE_DMA              0x00
#define DIAG_WRITE_MULTIPLE_PIO     0x01
#define DIAG_WRITE_SECTOR_PIO       0x02
#define DIAG_READ_DMA               0x03
#define DIAG_READ_MULTIPLE_PIO      0x04
#define DIAG_READ_VERIFY_SECTOR_PIO 0x05
#define DIAG_READ_SECTOR_PIO        0x06

/* Diagnostic Ioctl status */
#define DIAG_SUCCESS             0x00
#define DIAG_MEDIUM_ERROR        0x01
#define DIAG_ICRC_ERROR          0x02
#define DIAG_INVALID_SECTOR      0x03
#define DIAG_INVALID_PARAMETERS  0x04



typedef struct _DIAG_PASS_THRU_DATA {
	MP_DRIVE_ID DriveID;       /* drive Id */
	UCHAR diag_sub_code;       /* See Sub opcode definition */
	UINT sector;              /* Offset into diagnostic region */
	UINT data_transfer_length;/* in bytes */
	UCHAR  status;             /* see return status */
	UCHAR  reserved[32];
	PASS_THRU_DATA_EX passthru_data;
}
__attribute__ ((packed))
DIAG_PASS_THRU_DATA, * PDIAG_PASS_THRU_DATA;


/* ERROR LOG */
#define  AR_ERRLOG_ERR_TYPE_UNKNOWN      0
#define  AR_ERRLOG_ERR_TYPE_READ         1
#define  AR_ERRLOG_ERR_TYPE_WRITE        2
#define  AR_ERRLOG_ERR_TYPE_VERIFY       3
#define  AR_ERRLOG_ERR_TYPE_DRIVE        4
 /* this is the MSB, decides time format */
#define  AR_ERRLOG_ERR_TYPE_TIMESTAMP 0x80


/* When the error type MSB is enabled, use the following time format */
typedef struct _WIN9X_METHOD {
	UCHAR  Hour;
	UCHAR  Minute;
	UCHAR  Second;
	UCHAR  Century;
	UCHAR  Year;
	UCHAR  Month;
	UCHAR  Day;
}
__attribute__ ((packed))
WIN9X_METHOD;

typedef struct _MP_ERROR_LOG_HEADER {
	UCHAR Signature[6];      /* $XERR$ */
	UCHAR Size;              /* Size of this structure in bytes */
	UCHAR SizeErrorLogStruct;/* Size of struct ERROR_LOG */
	UCHAR ErrorLogSectors;   /* Number of sectors to use for error log */
	USHORT MaxErrorCount;    /* Maximum error log records */
	USHORT TotalErrorCount;  /* Total error log records */
	USHORT Head;             /* location of the newest entry in errlog */
	USHORT Tail;             /* location of the oldest entry in errLog */
}
__attribute__ ((packed))
MP_ERROR_LOG_HEADER, *PMP_ERROR_LOG_HEADER;

typedef struct _MP_ERROR_LOG_RECORD {
	UCHAR err_type;        /* error types ~ MSB desinates the timestamp */
	UCHAR err_sector[4];   /* sector where error occured */
	UCHAR err_timestamp[8];/* time stamp, based on MSB of "err_type" */
}
__attribute__ ((packed))
MP_ERROR_LOG_RECORD,*PMP_ERROR_LOG_RECORD;

#define AR_MAX_ERROR_LOG		38
typedef struct _MP_ERROR_LOG {
	MP_DRIVE_ID DriveID;
	MP_ERROR_LOG_HEADER errorlog_hdr;
	MP_ERROR_LOG_RECORD errorlog_rec[AR_MAX_ERROR_LOG];
}
__attribute__ ((packed))
MP_ERROR_LOG, *PMP_ERROR_LOG;

/* ioctls for backplane */
typedef enum {
	MP_SILENCE_BEEPER = 0,
	MP_SOUND_BEEPER = 1 ,
	MP_ENABLE_BEEPER = 2,
	MP_DISABLE_BEEPER =  3,
	MP_GET_STATUS = 4,
	MP_IDENTIFY_DRIVE = 5
} MP_BACKPLANE_REQUEST;

typedef enum {
	MP_DRIVE_PRESENT = 0,
	MP_DRIVE_NOT_PRESENT = 1
} MP_DRIVE_BACKPLANE_STATUS;

typedef struct _MP_BACKPLANE_ACTIVITY {
	MP_BACKPLANE_REQUEST Action; /* IN - the request , OUT - not used */
	UINT  ChannelId;    /* IN - MP_IDENTIFY_DRIVE only, OUT - not used */
	/* OUT - MP_GET_STATUS only */
	MP_DRIVE_BACKPLANE_STATUS  ChannelStatus[MAX_CHANNELS_PER_CONTROLLER];
} MP_BACKPLANE_ACTIVITY, *PMP_BACKPLANE_ACTIVITY ;

/* ioctls for backplane - end */
typedef enum {
	MP_SMART_READ_DATA  = 0xD0,
	MP_SMART_ENABLE_DISABLE_ATTRIBUTE_AUTOSAVE = 0xD2,
	MP_SMART_EXECUTE_OFF_LINE_IMMEDIATE = 0xD4,
	MP_SMART_READ_LOG = 0xD5,
	MP_SMART_WRITE_LOG = 0xD6,
	MP_SMART_ENABLE = 0xD8,
	MP_SMART_DISABLE = 0xD9,
	MP_SMART_READ_STATUS = 0xDA,
} MP_SMART_ACTION_TYPES;

typedef struct _MP_SMART_ACTIVITY {
	MP_DRIVE_ID  TargetID;
	MP_SMART_ACTION_TYPES  Action;
	UCHAR  ucData[512];
}
__attribute__((packed))
MP_SMART_ACTIVITY, *PMP_SMART_ACTIVITY;

typedef enum {
	MP_IRCD_READ = 0,
	MP_IRCD_WRITE = 1,
}MP_IRCD_ACTION;

typedef struct _MP_IRCD_ACTIVITY {
	MP_DRIVE_ID  TargetID;
	MP_IRCD_ACTION  Action;
	UCHAR  ucData[512];
}
__attribute__((packed))
MP_IRCD_ACTIVITY, *PMP_IRCD_ACTIVITY;

#define IO_USING_SEQUENCE 0x00
#define SEQ_READ          0x01
#define SEQ_WRITE         0x02
#define SEQ_POLL_BIT      0x03
#define SEQ_POLL_BYTE     0x04

#define FLASH_BIT_25      0x02000000
#define FLASH_BIT_27      0x08000000
#define FLASH_BIT_24      0x01000000
#define FLASH_BIT_00to18  0x0007ffff /* Flash address bitset */

typedef struct _SEQUENCE_HEADER {
	UCHAR  ucNumberOfSequence;
	UINT  uiBufferLength;
	UCHAR  ucStatus;
	UINT  uiNumberOfRepeat;
	UCHAR  ucReserved[6];
}
__attribute__((packed))
SEQUENCE_HEADER;

typedef struct _SEQUENCE {
	UCHAR  ucSeqOperation;
	UINT  uiSeqAddress;
	UCHAR  ucSeqData;
	UINT  uiNumberOfTimes;
	UCHAR  bUseBuffer;
	UCHAR  ucPollMask;
	UCHAR  ucReserved[4];
}
__attribute__((packed))
SEQUENCE;

typedef struct  _ERROR_INJECTOR_DATA {
	UCHAR	flag;	/* 
	 		 * 0 - Inject temporary error, inject err
			 *     only once and clear it after first occurance
	 		 * 1 - Inject permanent error, inject a 
			 *     permanent err, to clear this error user have to 
			 *     either reload the driver or call clear err IOCTL
	 		 * 2 - clear one error pattern populated at given index
			 * 3 - clear all error patten, if this option
			 *     is set, all other fields are invalid.
			 */  
	UINT	index;	/* error pattern index 0-9, Max 10 errors can
			 * be injected at the same time
			 */
	UINT	sector;	/* sector at which the error should inject */
	BOOLEAN	is_write;/* 0 - read error, 1 - write error */
	
	union _type{
		/* to inject read error */
		struct {
			UINT r_blk_err;	 	/* 0-No err, 1-Err */
			UINT r_sec_err;		/* 0-No err, 1-Err */
			UINT r_sec_frm_mirror_err;/* 0-No err, 1-Err */
			UINT w_back_to_mirror_err;/* 0-No err, 1-Err */
			UINT reserved[5];	/* for future use */
		} read;
		
		/* to inject write error */
		struct {
			UINT	w_blk_err;	/* to fail write command, */
						/* 0-No err, 1-Error */
			UINT	w_sec_err;	/* to fail write sector */
						/* 0-No err, 1-Error */
			UINT reserved[7];	/* for future use */
		} write;
	} type;

} ERROR_INJECTOR_DATA;


/* driver capability bit map values */
#define DRIVER_SMART_SUPPORT			0x01
#define DRIVER_RANDOM_DELETION_SUPPORT		0x02
#define DRIVER_HETROGENEOUS_LD_SUPPORT		0x04
#define DRIVER_SBBM_SUPPORT			0x08

typedef struct _DRIVER_CAPABILITY {
	UINT	uiCapability;	
	UINT	uiBBMVerInfo;		// Driver will fill this as 1
	UINT	uiBBMStructureSize;	// size in bytes and will be 2048.
	UCHAR	aucReserved[24];
}
__attribute__((packed))
DRIVER_CAPABILITY;


#ifdef _SOFT_BBM_
/* maximum sbbm error list possible with this design */
#define SBBM_MAX_REASSIGNED_SECTORS		508
/* Maximum One Command can be 256 Sectors  */	
#define SBBM_MAX_HANDLE_BBRW_SECTORS     	0x100	
/* Size of the whole table in bytes [4*514]-4Sectors*/	
#define SBBM_TABLE_SIZE_IN_BYTES		2048

/*
 * Description:
 * This structure holds soft bad block information of a hard drive.
 * The list will be in ascending order and it is maitained by Drivers, BIOS 
 * and Ctrl+M. For more information about this structure and its uses ses 
 * Soft Bad Block Management Specifications.
 *
 * @uiCheckWord	- this is a structure signature to verify the presence
 * 			  of SBBM, it is defined as 0x4D424253 (SBBM)
 * @ucCheckSum		- this structure check sum
 * @ucTableOptions	- 1 - No reassignment, just "Soft Bad" List
 *			- 2 - Reassigned list
 * @usBadSectorCount	- Number of maximum Bad Sectors, max value is 508
 * @aucReserved1	- Reserved Space 1, Not used by SBBM at this time
 * @aucReserved2	- Reserved Space 2, Not used by SBBM at this time
 * @aulLBA		- This is the array containing Sector Number for Bad
 *			  Sectors. This list will be in ascending order.
 */
typedef struct _sbbm_table_t {
	UINT	uiCheckWord;
	UCHAR	ucCheckSum;
	UCHAR	ucTableOptions;
	USHORT	usBadSectorCount;
	UCHAR	aucReserved1[2];
	UCHAR	aucReserved2[6];
	UINT	auiLBA[SBBM_MAX_REASSIGNED_SECTORS];
}
__attribute__ ((packed))
SOFT_BBM_TABLE;

typedef struct _GET_BBM_LIST {
	UINT		uiTargetId;
	UINT		auiReserved[4];
	SOFT_BBM_TABLE	BBMTable; // defined in SBBM design document.
}
__attribute__((packed))
GET_BBM_LIST;

#endif /*_SOFT_BBM_*/

#endif /* __HD_IOCTLS_H__ */

/* vim: set ts=8 sw=8 tw=78 ai si: */


